Package: L0ggm
Title: Smooth L0 Penalty Approximations for Gaussian Graphical Models
Version: 0.1.0
Date: 2026-05-15
Authors@R: c(
    person("Alexander", "Christensen", email = "alexpaulchristensen@gmail.com", role = c("aut", "cre"), comment = c(ORCID = "0000-0002-9798-7037")),
    person("Jeongwon", "Choi", email = "jungwon799@gmail.com", role = "ctb", comment = c(ORCID = "0000-0001-6087-2124")),
    person("John", "Fox", role = c("cph", "ctb"), comment = "Original implementation of polyserial correlations in auto_correlate.R"),
    person("Yves", "Rosseel", role = c("cph", "ctb"), comment = "Original implementation of rmsea_ci in network_fit.R"),
    person("Alexander", "Robitzsch", role = c("cph", "ctb"), comment = "C++ implementation of Drezner-Wesolowsky bivariate normal CDF in polychoric_matrix.c"),
    person("David", "Blackman", role = "ctb", comment = "Original xoshiro.c implementation"),
    person("Sebastiano", "Vigna", role = "ctb", comment = "Original xoshiro.c implementation"),
    person("John", "Burkardt", role = c("cph", "ctb"), comment = "Original ziggurat.c implementation")
   )
Description: Provides smooth approximations to the L0 norm penalty for
    estimating sparse Gaussian graphical models (GGMs). Network estimation
    is performed using the Local Linear Approximation (LLA) framework
    (Fan & Li, 2001 <doi:10.1198/016214501753382273>;
    Zou & Li, 2008 <doi:10.1214/009053607000000802>) with five penalty
    functions: arctangent (Wang & Zhu, 2016 <doi:10.1155/2016/6495417>),
    EXP (Wang, Fan, & Zhu, 2018 <doi:10.1007/s10463-016-0588-3>), Gumbel,
    Log (Candes, Wakin, & Boyd, 2008 <doi:10.1007/s00041-008-9045-x>),
    and Weibull. Adaptive penalty parameters for EXP, Gumbel, and Weibull
    are estimated via maximum likelihood, and model selection uses
    information criteria including AIC, BIC, and EBIC (Extended BIC).
    Simulation functions generate multivariate normal data from GGMs with
    stochastic block model or small-world (Watts-Strogatz) network structures.
Depends: R (>= 3.5.0)
License: AGPL (>= 3.0)
Encoding: UTF-8
LazyData: true
NeedsCompilation: yes
Imports: igraph, glasso, glassoFast, Matrix, methods, psych, stats
Copyright: See inst/COPYRIGHTS for details
BugReports: https://github.com/AlexChristensen/L0ggm/issues
Config/roxygen2/version: 8.0.0
Packaged: 2026-05-15 19:34:34 UTC; alextops
Author: Alexander Christensen [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-9798-7037>),
  Jeongwon Choi [ctb] (ORCID: <https://orcid.org/0000-0001-6087-2124>),
  John Fox [cph, ctb] (Original implementation of polyserial correlations
    in auto_correlate.R),
  Yves Rosseel [cph, ctb] (Original implementation of rmsea_ci in
    network_fit.R),
  Alexander Robitzsch [cph, ctb] (C++ implementation of
    Drezner-Wesolowsky bivariate normal CDF in polychoric_matrix.c),
  David Blackman [ctb] (Original xoshiro.c implementation),
  Sebastiano Vigna [ctb] (Original xoshiro.c implementation),
  John Burkardt [cph, ctb] (Original ziggurat.c implementation)
Maintainer: Alexander Christensen <alexpaulchristensen@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-15 22:20:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-05-15 23:05:47 UTC; unix
Archs: L0ggm.so.dSYM

Package: L0ggm
Title: Smooth L0 Penalty Approximations for Gaussian Graphical Models
Version: 0.1.0
Date: 2026-05-15
Authors@R: c(
    person("Alexander", "Christensen", email = "alexpaulchristensen@gmail.com", role = c("aut", "cre"), comment = c(ORCID = "0000-0002-9798-7037")),
    person("Jeongwon", "Choi", email = "jungwon799@gmail.com", role = "ctb", comment = c(ORCID = "0000-0001-6087-2124")),
    person("John", "Fox", role = c("cph", "ctb"), comment = "Original implementation of polyserial correlations in auto_correlate.R"),
    person("Yves", "Rosseel", role = c("cph", "ctb"), comment = "Original implementation of rmsea_ci in network_fit.R"),
    person("Alexander", "Robitzsch", role = c("cph", "ctb"), comment = "C++ implementation of Drezner-Wesolowsky bivariate normal CDF in polychoric_matrix.c"),
    person("David", "Blackman", role = "ctb", comment = "Original xoshiro.c implementation"),
    person("Sebastiano", "Vigna", role = "ctb", comment = "Original xoshiro.c implementation"),
    person("John", "Burkardt", role = c("cph", "ctb"), comment = "Original ziggurat.c implementation")
   )
Description: Provides smooth approximations to the L0 norm penalty for
    estimating sparse Gaussian graphical models (GGMs). Network estimation
    is performed using the Local Linear Approximation (LLA) framework
    (Fan & Li, 2001 <doi:10.1198/016214501753382273>;
    Zou & Li, 2008 <doi:10.1214/009053607000000802>) with five penalty
    functions: arctangent (Wang & Zhu, 2016 <doi:10.1155/2016/6495417>),
    EXP (Wang, Fan, & Zhu, 2018 <doi:10.1007/s10463-016-0588-3>), Gumbel,
    Log (Candes, Wakin, & Boyd, 2008 <doi:10.1007/s00041-008-9045-x>),
    and Weibull. Adaptive penalty parameters for EXP, Gumbel, and Weibull
    are estimated via maximum likelihood, and model selection uses
    information criteria including AIC, BIC, and EBIC (Extended BIC).
    Simulation functions generate multivariate normal data from GGMs with
    stochastic block model or small-world (Watts-Strogatz) network structures.
Depends: R (>= 3.5.0)
License: AGPL (>= 3.0)
Encoding: UTF-8
LazyData: true
NeedsCompilation: yes
Imports: igraph, glasso, glassoFast, Matrix, methods, psych, stats
Copyright: See inst/COPYRIGHTS for details
BugReports: https://github.com/AlexChristensen/L0ggm/issues
Config/roxygen2/version: 8.0.0
Packaged: 2026-05-15 19:34:34 UTC; alextops
Author: Alexander Christensen [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-9798-7037>),
  Jeongwon Choi [ctb] (ORCID: <https://orcid.org/0000-0001-6087-2124>),
  John Fox [cph, ctb] (Original implementation of polyserial correlations
    in auto_correlate.R),
  Yves Rosseel [cph, ctb] (Original implementation of rmsea_ci in
    network_fit.R),
  Alexander Robitzsch [cph, ctb] (C++ implementation of
    Drezner-Wesolowsky bivariate normal CDF in polychoric_matrix.c),
  David Blackman [ctb] (Original xoshiro.c implementation),
  Sebastiano Vigna [ctb] (Original xoshiro.c implementation),
  John Burkardt [cph, ctb] (Original ziggurat.c implementation)
Maintainer: Alexander Christensen <alexpaulchristensen@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-15 22:20:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-05-15 22:47:20 UTC; unix
Archs: L0ggm.so.dSYM

# Generated by roxygen2: do not edit by hand

export(auto_correlate)
export(categorize)
export(edge_confusion)
export(network_estimation)
export(network_fit)
export(polychoric_matrix)
export(proxswap_lattice)
export(simulate_sbm)
export(simulate_smallworld)
export(smallworldness)
export(weibull_parameters)
importFrom(methods,is)
importFrom(stats,complete.cases)
importFrom(stats,cor)
importFrom(stats,cov2cor)
importFrom(stats,dist)
importFrom(stats,dnorm)
importFrom(stats,na.omit)
importFrom(stats,optimize)
importFrom(stats,pchisq)
importFrom(stats,pnorm)
importFrom(stats,predict)
importFrom(stats,qnorm)
importFrom(stats,rnorm)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,uniroot)
importFrom(utils,capture.output)
importFrom(utils,data)
importFrom(utils,packageVersion)
useDynLib(L0ggm, .registration = TRUE)

#%%%%%%%%%%%%%%%%%#
#### Penalties ####
#%%%%%%%%%%%%%%%%%#

#' @noRd
# Updated 25.07.2025
atan_penalty <- function(x, lambda, gamma = 0.01, ...)
{
  return(lambda * (gamma + 2 / pi) * atan(abs(x) / gamma))
}

#' @noRd
# Updated 10.01.2026
exp_penalty <- function(x, lambda, gamma = 0.01, ...)
{

  # Pre-compute components
  x <- abs(x)

  # Return penalty
  return(lambda * (1 - exp(-(x / gamma))))

}

#' @noRd
# Updated 11.04.2026
gumbel_penalty <- function(x, lambda, gamma = 0.01, ...)
{

  # Pre-compute components
  x <- abs(x)

  # Return penalty
  return(lambda * (exp(-exp(-x / gamma)) - exp(-exp(x / gamma))))

}

#' @noRd
# Updated 08.03.2026
log_penalty <- function(x, lambda, gamma = 0.10, ...)
{
  return(lambda * log(1 + abs(x) / gamma) / log(1 + 1 / gamma))
}

#' @noRd
# Updated 15.03.2026
weibull_penalty <- function(x, lambda, gamma = 0.01, shape, ...)
{

  # Pre-compute
  x <- abs(x)

  # Check for shape greater than one
  if(shape > 1){

    # Mirror derivative
    shape_one  <- shape - 1
    peak       <- gamma * ((shape - 1) / shape)^(1 / shape)
    peak_gamma <- peak / gamma
    peak_value <- (shape / gamma) * peak_gamma^shape_one * exp(-peak_gamma^shape)

    # Value of Weibull penalty at peak
    at_peak <- 1 - exp(-peak_gamma^shape)

    # Shift so pieces meet at peak, anchored at zero
    shift <- peak_value * peak - at_peak

    # Return penalty
    return(lambda * swiftelse(x <= peak, peak_value * x, 1 - exp(-(x / gamma)^shape) + shift))

  }

  # Otherwise, return standard penalty
  return(lambda * (1 - exp(-(x / gamma)^shape)))

}

#' @title L0 Norm Regularized Network Estimation
#'
#' @description A general function to estimate Gaussian graphical models using
#' L0 penalty approximations. All penalties are implemented using
#' either single-pass or full Local Linear Approximation (LLA: Fan & Li, 2001; Zou & Li, 2008)
#'
#' @param data Matrix or data frame.
#' Should consist only of variables to be used in the analysis
#'
#' @param n Numeric (length = 1).
#' Sample size \strong{must} be provided if \code{data} provided is a correlation matrix
#'
#' @param corr Character (length = 1).
#' Method to compute correlations.
#' Defaults to \code{"auto"}.
#' Available options:
#'
#' \itemize{
#'
#' \item \code{"auto"} --- Automatically computes appropriate correlations for
#' the data using Pearson's for continuous, polychoric for ordinal,
#' tetrachoric for binary, and polyserial/biserial for ordinal/binary with
#' continuous. To change the number of categories that are considered
#' ordinal, use \code{ordinal_categories}
#' (see \code{\link[L0ggm]{polychoric_matrix}} for more details)
#'
#' \item \code{"pearson"} --- Pearson's correlation is computed for all
#' variables regardless of categories
#'
#' \item \code{"spearman"} --- Spearman's rank-order correlation is computed
#' for all variables regardless of categories
#'
#' }
#'
#' For other similarity measures, compute them first and input them
#' into \code{data} with the sample size (\code{n})
#'
#' @param na_data Character (length = 1).
#' How should missing data be handled?
#' Defaults to \code{"pairwise"}.
#' Available options:
#'
#' \itemize{
#'
#' \item \code{"pairwise"} --- Computes correlation for all available cases between
#' two variables
#'
#' \item \code{"listwise"} --- Computes correlation for all complete cases in the dataset
#'
#' }
#'
#' @param penalty Character (length = 1).
#' Defaults to \code{"weibull"}.
#' Available options:
#'
#' \itemize{
#'
#' \item \code{"atan"} --- Arctangent (Wang & Zhu, 2016)
#' \deqn{\lambda \cdot (\gamma + \frac{2}{\pi}) \cdot \arctan\left(\frac{|x|}{\gamma}\right)}
#'
#' \item \code{"exp"} --- EXP (Wang, Fan, & Zhu, 2018)
#' \deqn{\lambda \cdot (1 - e^{-\frac{|x|}{\gamma}})}
#'
#' \item \code{"gumbel"} --- Gumbel
#' \deqn{\frac{\lambda}{1 - e^{-1}} \cdot \left(e^{-e^{-\frac{|x|}{\gamma}}} - e^{-1}\right)}
#'
#' \item \code{"log"} --- Log (Candes, Wakin, & Boyd, 2008)
#' \deqn{\frac{\lambda \cdot \log\left(1 + \frac{|x|}{\gamma}\right)}{\log\left(1 + \frac{1}{\gamma}\right)}}
#'
#' \item \code{"weibull"} --- Weibull
#' \deqn{\lambda \cdot \left(1 - e^{-\left(\frac{|x|}{\gamma}\right)^k}\right)}
#'
#' }
#'
#' @param gamma Numeric (length = 1).
#' Adjusts the shape of the penalty.
#' Defaults:
#'
#' \itemize{
#'
#' \item \code{"atan"} = 0.01
#'
#' \item \code{"exp"} = 0.01
#'
#' \item \code{"gumbel"} = 0.01
#'
#' \item \code{"log"} = 0.10
#'
#' \item \code{"weibull"} = 0.01
#'
#' }
#'
#' @param adaptive Boolean (length = 1).
#' Whether data-adaptive (gamma) parameters should be used.
#' Defaults to \code{TRUE}.
#' Set to \code{FALSE} to apply default gamma parameters for
#' adaptive penalties.
#' Available options:
#'
#' \itemize{
#'
#' \item \code{"exp"}
#'
#' \item \code{"gumbel"}
#'
#' \item \code{"weibull"}
#'
#' }
#'
#' When \code{adaptive = TRUE}, gamma is set to the 10th percentile of the
#' distribution fitted to the empirical partial correlations
#'
#' @param nlambda Numeric (length = 1).
#' Number of lambda values to test.
#' Defaults to \code{50}
#'
#' @param lambda_min_ratio Numeric (length = 1).
#' Ratio of lowest lambda value compared to maximal lambda.
#' Defaults to \code{0.01}
#'
#' @param penalize_diagonal Boolean (length = 1).
#' Should the diagonal be penalized?
#' Defaults to \code{TRUE}
#'
#' @param ic Character (length = 1).
#' What information criterion should be used for model selection?
#' Available options include:
#'
#' \itemize{
#'
#' \item \code{"AIC"} --- Akaike's information criterion: \eqn{-2L + 2E}
#'
#' \item \code{"AICc"} --- AIC corrected: \eqn{AIC + \frac{2E^2 + 2E}{n - E - 1}}
#'
#' \item \code{"BIC"} --- Bayesian information criterion: \eqn{-2L + E \cdot \log{(n)}}
#'
#' \item \code{"BIC0"} --- Bayesian information criterion not (Dicker et al., 2013): \eqn{\log{\large(\frac{D}{n - E}\large)} + \large(\frac{\log{(n)}}{n}\large) \cdot E}
#'
#' \item \code{"EBIC"} --- Extended BIC: \eqn{BIC + 4E \cdot \gamma \cdot \log{(E)}}
#'
#' \item \code{"MBIC"} --- Modified Bayesian information criterion (Wang et al., 2018):  \eqn{\log{\large(\frac{D}{n - E}\large)} + \large(\frac{\log{(n)} \cdot E}{n}\large) \cdot \log{(\log{(p)}})}
#'
#' }
#'
#' Term definitions:
#'
#' \itemize{
#'
#' \item \eqn{n} --- sample size
#'
#' \item \eqn{p} --- number of variables
#'
#' \item \eqn{E} --- edges
#'
#' \item \eqn{S} --- empirical correlation matrix
#'
#' \item \eqn{K} --- estimated inverse covariance matrix (network)
#'
#' \item \eqn{L = \frac{n}{2} \cdot \log \text{det} K - \sum_{i=1}^p (SK)_{ii}}
#'
#' \item \eqn{D = n \cdot \sum_{i=1}^p (SK)_{ii} - \log \text{det} K}
#'
#' }
#'
#' Defaults to \code{"BIC"}
#'
#' @param ebic_gamma Numeric (length = 1)
#' Value to set gamma parameter in EBIC (see above).
#' Defaults to \code{0.50}
#'
#' \emph{Only used if \code{ic = "EBIC"}}
#'
#' @param fast Boolean (length = 1).
#' Whether the \code{\link[glassoFast]{glassoFast}} version should be used
#' to estimate the GLASSO.
#' Defaults to \code{TRUE}.
#'
#' The fast results \emph{may} differ by less than floating point of the original
#' GLASSO implemented by \code{\link[glasso]{glasso}} and should not impact reproducibility much (set to \code{FALSE} if concerned)
#'
#' @param LLA Boolean (length = 1).
#' Should Local Linear Approximation be used to find optimal minimum?
#' Defaults to \code{FALSE} or a single-pass approximation, which can be
#' significantly faster (Zou & Li, 2008).
#' Set to \code{TRUE} to find global minimum based on convergence (\code{LLA_threshold})
#'
#' @param LLA_threshold Numeric (length = 1).
#' When performing the Local Linear Approximation, the maximum threshold
#' until convergence is met.
#' Defaults to \code{1e-03}
#'
#' @param LLA_iter Numeric (length = 1).
#' Maximum number of iterations to perform to reach convergence.
#' Defaults to \code{10000}
#'
#' @param network_only Boolean (length = 1).
#' Whether the network only should be output.
#' Defaults to \code{TRUE}.
#' Set to \code{FALSE} to obtain all output for the
#' network estimation method
#'
#' @param verbose Boolean (length = 1).
#' Whether messages and (insignificant) warnings should be output.
#' Defaults to \code{FALSE} (silent calls).
#' Set to \code{TRUE} to see all messages and warnings for every function call
#'
#' @param ... Additional arguments to be passed on to \code{auto.correlate}
#'
#' @return When \code{network_only = TRUE} (default), returns a \eqn{p \times p}
#' numeric matrix of partial correlations representing the estimated Gaussian
#' graphical model. Off-diagonal entry \eqn{[i,j]} is the partial correlation
#' between variables \eqn{i} and \eqn{j} controlling for all other variables,
#' with values in \eqn{[-1, 1]}; a value of zero indicates the absence of an
#' edge. Diagonal entries are zero. Row and column names are inherited from
#' \code{data}.
#'
#' When \code{network_only = FALSE}, returns a named list with the following
#' elements:
#'
#' \describe{
#'
#' \item{\code{network}}{The \eqn{p \times p} partial correlation matrix
#' described above}
#'
#' \item{\code{K}}{The \eqn{p \times p} estimated inverse covariance
#' (precision) matrix at the optimal lambda. Diagonal entries are the
#' conditional precisions; off-diagonal entries are proportional to
#' partial covariances}
#'
#' \item{\code{R}}{The \eqn{p \times p} regularized covariance matrix
#' returned by GLASSO at the optimal lambda (the \code{w} component of the
#' GLASSO solution)}
#'
#' \item{\code{penalty}}{Character string naming the penalty function used
#' (one of \code{"atan"}, \code{"exp"}, \code{"gumbel"}, \code{"log"},
#' \code{"weibull"})}
#'
#' \item{\code{lambda}}{Numeric scalar giving the regularization parameter
#' value selected by the information criterion. Larger values correspond
#' to sparser networks}
#'
#' \item{\code{gamma}}{Numeric scalar giving the shape parameter of the
#' penalty actually used. For adaptive penalties (\code{adaptive = TRUE}),
#' this is the data-derived value; otherwise it is the default or
#' user-supplied value}
#'
#' \item{\code{correlation}}{The \eqn{p \times p} empirical correlation
#' matrix computed from \code{data}, used as input to the GLASSO}
#'
#' \item{\code{criterion}}{Character string naming the information criterion
#' used for model selection (e.g., \code{"bic"})}
#'
#' \item{\code{IC}}{Numeric scalar giving the value of the information
#' criterion at the optimal lambda}
#'
#' }
#'
#' @author Alexander P. Christensen <alexpaulchristensen@gmail.com>
#'
#' @references
#'
#' \strong{Log penalty} \cr
#' Candes, E. J., Wakin, M. B., & Boyd, S. P. (2008).
#' Enhancing sparsity by reweighted l1 minimization.
#' \emph{Journal of Fourier Analysis and Applications}, \emph{14}(5), 877--905.
#'
#' \strong{BIC0} \cr
#' Dicker, L., Huang, B., & Lin, X. (2013).
#' Variable selection and estimation with the seamless-L0 penalty.
#' \emph{Statistica Sinica}, \emph{23}(2), 929--962.
#'
#' \strong{Local Linear Approximation} \cr
#' Fan, J., & Li, R. (2001).
#' Variable selection via nonconcave penalized likelihood and its oracle properties.
#' \emph{Journal of the American Statistical Association}, \emph{96}(456), 1348--1360.
#'
#' \strong{EXP penalty} \cr
#' Wang, Y., Fan, Q., & Zhu, L. (2018).
#' Variable selection and estimation using a continuous approximation to the L0 penalty.
#' \emph{Annals of the Institute of Statistical Mathematics}, \emph{70}(1), 191--214.
#'
#' \strong{Atan penalty} \cr
#' Wang, Y., & Zhu, L. (2016).
#' Variable selection and parameter estimation with the Atan regularization method.
#' \emph{Journal of Probability and Statistics}, \emph{2016}, 1--12.
#'
#' \strong{Seminal simulation in network psychometrics} \cr
#' Williams, D. R. (2020).
#' Beyond lasso: A survey of nonconvex regularization in Gaussian graphical models.
#' \emph{PsyArXiv}.
#'
#' \strong{One-step Local Linear Approximation} \cr
#' Zou, H., & Li, R. (2008).
#' One-step sparse estimates in nonconcave penalized likelihood models.
#' \emph{Annals of Statistics}, \emph{36}(4), 1509--1533.
#'
#' @examples
#' # Obtain default estimator (adaptive Weibull)
#' weibull_network <- network_estimation(
#'   data = basic_smallworld, LLA = TRUE
#' )
#'
#' # Obtain Atan network
#' atan_network <- network_estimation(
#'   data = basic_smallworld, penalty = "atan", LLA = TRUE
#' )
#'
#' # Obtain static EXP network
#' exp_network <- network_estimation(
#'   data = basic_smallworld, penalty = "exp",
#'   adaptive = FALSE, LLA = TRUE
#' )
#'
#' @export
#'
# Apply non-convex regularization ----
# Updated 11.04.2026
network_estimation <- function(
    data, n = NULL,
    corr = c("auto", "pearson", "spearman"),
    na_data = c("pairwise", "listwise"),
    penalty = c("atan", "exp", "gumbel", "log", "weibull"),
    gamma = NULL, adaptive = TRUE,
    nlambda = 50, lambda_min_ratio = 0.01, penalize_diagonal = TRUE,
    ic = c("AIC", "AICc", "BIC", "BIC0", "EBIC", "MBIC"), ebic_gamma = 0.50,
    fast = TRUE, LLA = FALSE, LLA_threshold = 1e-03, LLA_iter = 10000,
    network_only = TRUE, verbose = FALSE, ...
)
{

  # Check for missing arguments (argument, default, function)
  # Uses actual function they will be used in
  # (keeping non-function choices for `cor_auto`)
  corr <- set_default(corr, "auto", network_estimation)
  na_data <- set_default(na_data, "pairwise", network_estimation)
  penalty <- set_default(penalty, "weibull", network_estimation)
  ic <- set_default(ic, "bic", network_estimation)

  # Argument errors (return data in case of tibble)
  data <- network_regularization_errors(
    data, n, gamma, adaptive, nlambda,
    lambda_min_ratio, penalize_diagonal, ebic_gamma,
    fast, LLA, LLA_threshold, network_only, verbose, ...
  )

  # Check whether penalty is adaptive option
  adaptive_option <- c("exp", "gumbel", "weibull")
  adaptive_flag <- adaptive & (penalty %in% adaptive_option)

  # Get necessary inputs
  output <- obtain_sample_correlations(
    data = data, n = n,
    corr = corr, na_data = na_data,
    verbose = verbose, needs_usable = FALSE, # skips usable data check
    ...
  )

  # Get outputs
  data <- output$data; n <- output$n
  S <- output$correlation_matrix

  # Get number of variables
  nodes <- dim(S)[2]

  # Obtain precision matrix
  K <- solve(S)

  # Obtain GLASSO function
  glasso_FUN <- swiftelse(fast, glassoFast::glassoFast, glasso::glasso)

  # Get function arguments
  glasso_ARGS <- obtain_arguments(glasso_FUN, FUN.args = list(...))

  # Supply correlation matrix
  glasso_ARGS[[1]] <- S

  # Get derivative function
  derivative_FUN <- switch(
    penalty,
    "atan" = atan_derivative,
    "exp" = exp_derivative,
    "gumbel" = gumbel_derivative,
    "log" = log_derivative,
    "weibull" = weibull_derivative
  )

  # Set shape (only used for Weibull)
  shape <- 1

  # Set gamma (set ahead of time for messaging on adaptive)
  if(is.null(gamma)){

    # Set defaults
    gamma <- switch(
      penalty,
      "atan" = 0.01,
      "exp" = 0.01,
      "gumbel" = 0.01,
      "log" = 0.10,
      "weibull" = 0.01
    )

  }

  # Check for adaptive and penalty is adaptive option
  if(adaptive_flag){

    # Set lower triangle
    lower_triangle <- lower.tri(S)

    # Set partial correlations
    P <- inv2pcor(K); lower_P <- abs(P[lower_triangle])

    if(penalty == "exp"){

      # Set 10th percentile
      gamma <- -log(0.90) * mean(lower_P)

    }else if(penalty == "gumbel"){

      # Set 10th percentile
      gamma <- qgumbel(0.10, gumbel_mle(lower_P))

    }else if(penalty == "weibull"){

      # Obtain Weibull estimates
      estimates <- weibull_mle(lower_P)

      # Get shape
      shape <- estimates[["shape"]]
      # 'shape' needs to be set! Otherwise, will default to 1 (from above)

      # Set 10th percentile
      gamma <- estimates[["scale"]] * (-log(0.90))^(1 / shape)

    }

  }

  # Initialize lambda matrix
  lambda_matrix <- matrix(0, nrow = nodes, ncol = nodes)

  # Simplify source for fewer computations (minimal improvement)
  S_zero_diagonal <- S - diag(nodes) # makes diagonal zero
  lambda_max <- max(abs(S_zero_diagonal)) # uses absolute rather than inverse
  lambda_min <- lambda_max * lambda_min_ratio
  lambda_sequence <- exp(seq.int(log(lambda_min), log(lambda_max), length.out = nlambda))

  # Get GLASSO output
  glasso_list <- lapply(lambda_sequence, function(lambda){

    # Set lambda matrix
    glasso_ARGS$rho <- lambda

    # Obtain estimate
    estimate <- do.call(what = glasso_FUN, args = glasso_ARGS)

    # Check for LLA
    if(LLA){

      # Obtain new K
      new_K <- estimate$wi

      # Set convergence and iterations
      convergence <- Inf; iterations <- 0

      # Loop over to convergence
      while((convergence > LLA_threshold) & (iterations < LLA_iter)){

        # Set old K
        old_K <- new_K

        # Obtain lambda matrix
        lambda_matrix[] <- derivative_FUN(x = old_K, lambda = lambda, gamma = gamma, shape = shape)

        # Check for diagonal penalization
        if(!penalize_diagonal){
          diag(lambda_matrix) <- 0
        }

        # Set lambda matrix
        glasso_ARGS$rho <- lambda_matrix

        # Obtain estimate
        estimate <- do.call(what = glasso_FUN, args = glasso_ARGS)

        # Obtain new K
        new_K <- estimate$wi

        # Increase iterations
        iterations <- iterations + 1

        # Check for diagonal penalization
        if(!penalize_diagonal){
          diag(new_K) <- diag(old_K) <- 0
        }

        # Compute convergence
        convergence <- max(abs(new_K - old_K))

      }

    }else{

      # Obtain lambda matrix
      lambda_matrix[] <- derivative_FUN(x = estimate$wi, lambda = lambda, gamma = gamma, shape = shape)

      # Check for diagonal penalization
      if(!penalize_diagonal){
        diag(lambda_matrix) <- 0
      }

      # Set lambda matrix
      glasso_ARGS$rho <- lambda_matrix

      # Obtain estimate
      estimate <- do.call(what = glasso_FUN, args = glasso_ARGS)

    }

    # Estimate
    return(estimate)

  })

  # Compute ICs
  ICs <- nvapply(glasso_list, function(element){
    information_criterion(
      S = S, K = element$wi, n = n, nodes = nodes,
      ic = ic, ebic_gamma = ebic_gamma
    )
  })

  # Optimal value
  optimal <- which.min(ICs)

  # Get R
  R <- glasso_list[[optimal]]$w

  # Get W
  W <- inv2pcor(glasso_list[[optimal]]$wi)
  dimnames(R) <- dimnames(W) <- dimnames(S)

  # Return results
  if(network_only){
    return(W)
  }else{
    return(
      list(
        network = W, K = glasso_list[[optimal]]$wi, R = R,
        penalty = penalty, lambda = lambda_sequence[[optimal]], gamma = gamma,
        correlation = S, criterion = ic, IC = ICs[[optimal]]
      )
    )
  }

}

# Bug checking ----
# data = basic_smallworld; n = NULL; corr = "auto"
# na_data = "pairwise"; penalty = "weibull"; adaptive = TRUE
# gamma = NULL; lambda = NULL; nlambda = 50
# lambda_min_ratio = 0.01; penalize_diagonal = TRUE
# optimize.lambda = FALSE; ic = "BIC"; network_only = TRUE
# ebic_gamma = 0.5; fast = TRUE; verbose = FALSE
# LLA = TRUE; LLA_threshold = 1e-04; LLA_iter = 10000

#' @noRd
# Errors ----
# Updated 19.01.2026
network_regularization_errors <- function(
    data, n, gamma, adaptive, nlambda,
    lambda_min_ratio, penalize_diagonal, ebic_gamma,
    fast, LLA, LLA_threshold, network_only, verbose, ...
)
{

  # 'data' errors
  object_error(data, c("matrix", "data.frame", "tibble"), "network_estimation")

  # Check for tibble
  if(get_object_type(data) == "tibble"){
    data <- as.data.frame(data)
  }

  # 'n' errors
  if(!is.null(n)){
    length_error(n, 1, "network_estimation")
    typeof_error(n, "numeric", "network_estimation")
  }

  # 'gamma' errors
  if(!is.null(gamma)){
    length_error(gamma, 1, "network_estimation")
    typeof_error(gamma, "numeric", "network_estimation")
    range_error(gamma, c(0, Inf), "network_estimation")
  }

  # 'adaptive' errors
  length_error(adaptive, 1, "network_estimation")
  typeof_error(adaptive, "logical", "network_estimation")

  # 'nlambda' errors
  length_error(nlambda, 1, "network_estimation")
  typeof_error(nlambda, "numeric", "network_estimation")
  range_error(nlambda, c(1, Inf), "network_estimation")

  # 'lambda_min_ratio' errors
  length_error(lambda_min_ratio, 1, "network_estimation")
  typeof_error(lambda_min_ratio, "numeric", "network_estimation")
  range_error(lambda_min_ratio, c(0, 1), "network_estimation")

  # 'penalize_diagonal' errors
  length_error(penalize_diagonal, 1, "network_estimation")
  typeof_error(penalize_diagonal, "logical", "network_estimation")

  # 'ebic_gamma' errors
  length_error(ebic_gamma, 1, "network_estimation")
  typeof_error(ebic_gamma, "numeric", "network_estimation")
  range_error(ebic_gamma, c(0, Inf), "network_estimation")

  # 'fast' errors
  length_error(fast, 1, "network_estimation")
  typeof_error(fast, "logical", "network_estimation")

  # 'LLA' errors
  length_error(LLA, 1, "network_estimation")
  typeof_error(LLA, "logical", "network_estimation")

  # 'LLA_threshold' errors
  if(LLA){
    length_error(LLA_threshold, 1, "network_estimation")
    typeof_error(LLA_threshold, "numeric", "network_estimation")
    range_error(LLA_threshold, c(-Inf, 0.10), "network_estimation")
  }

  # 'network_only' errors
  length_error(network_only, 1, "network_estimation")
  typeof_error(network_only, "logical", "network_estimation")

  # 'verbose' errors
  length_error(verbose, 1, "network_estimation")
  typeof_error(verbose, "logical", "network_estimation")

  # Check for usable data
  if(needs_usable(list(...))){
    data <- usable_data(data, verbose)
  }

  # Return data in case of tibble
  return(data)

}

#' @noRd
# (Folded) Gumbel quantile----
# Updated 11.04.2026
qgumbel <- function(p, scale)
{
  uniroot(
    f = function(q){
      exp(-exp(-q / scale)) - exp(-exp(q / scale)) - p
    },
    interval = c(0, scale * 100)
  )$root
}

#' @noRd
# Information criterion ----
# Updated 09.03.2026
information_criterion <- function(S, K, n, nodes, ic, ebic_gamma)
{

  # Compute Gaussian likelihood (minus two for convenience)
  trSK <- sum(S * K)
  log_det_K <- log(det(K))
  L <- swiftelse(
    ic %in% c("bic0", "mbic"),
    n * trSK - log_det_K, # use deviance
    -2 * (n / 2) * (log_det_K - trSK) # use log-likelihood
  )

  # Set diagonal of K to zero
  diag(K) <- 0

  # Get parameters (edges)
  E <- edge_count(K, nodes)

  # Ensure that there is enough degrees of freedom
  df <- n - E

  # Return information criterion
  return(
    switch(
      ic,
      "aic" = L + 2 * E,
      "aicc" = swiftelse(n - E - 1 > 0, L + 2 * E + (2 * E^2 + 2 * E) / (n - E - 1), Inf),
      "bic" = L + E * log(n),
      "ebic" = L + E * log(n) + 4 * E * ebic_gamma * log(nodes),
      "bic0" = swiftelse( # see https://www.jstor.org/stable/24310368
        df > 0, log(L / df) + (log(n) * E / n), -Inf
      ),
      "mbic" = swiftelse( # see https://doi.org/10.1007%2Fs10463-016-0588-3
        df > 0, log(L / df) + (log(n) * E / n) * log(log(nodes)), -Inf
      )
    )
  )

}

#' @title Computes Various Small-Worldness Metrics
#'
#' @description Computes the small-worldness of a network using one of five
#' methods. All simulation-based methods generate degree-preserving random
#' graphs as the null baseline. The lattice reference network (where required)
#' is constructed using \code{\link[L0ggm]{proxswap_lattice}}, which produces a
#' degree-preserving ring lattice that maximizes the clustering coefficient
#'
#' @param network Matrix or data frame.
#' A square, symmetric numeric matrix representing a network (e.g., partial
#' correlations). Absolute values are taken internally, so signed weights are
#' handled automatically. Non-zero off-diagonal entries are treated as edges.
#'
#' @param lattice Matrix (optional).
#' A pre-computed lattice adjacency matrix to use as the regular-network
#' reference for the \code{"omega"} and \code{"SWP"} methods.
#' If not provided, a lattice is generated automatically via
#' \code{\link[L0ggm]{proxswap_lattice}}. Ignored by \code{"analytical"}
#' and \code{"S"}
#'
#' @param method Character (length = 1).
#' The method used to compute small-worldness.
#' Defaults to \code{"SWP"}.
#' Available options:
#'
#' \itemize{
#'
#' \item \code{"analytical"} --- Computes the Humphries & Gurney (2008)
#' \eqn{S} metric using closed-form approximations for the Erdős–Rényi
#' random graph baseline:
#' \deqn{S = \frac{C / C_{\text{rand}}}{L / L_{\text{rand}}}}
#' where \eqn{C_{\text{rand}} \approx \langle k \rangle / n} and
#' \eqn{L_{\text{rand}} \approx \ln(n) / \ln(\langle k \rangle)}.
#' No random graphs are generated. The \code{weighted} argument is ignored.
#' Values greater than 1 indicate small-world structure
#'
#' \item \code{"S"} --- Computes the Humphries & Gurney (2008) \eqn{S}
#' metric by simulation. Random graphs are generated under the
#' Erdős–Rényi model (\code{\link[igraph]{sample_gnm}}) with the same
#' number of nodes and edges as \code{network}. The empirical and random
#' clustering coefficients both use global transitivity (\eqn{C^\Delta}).
#' When \code{weighted = TRUE}, edge weights are reassigned to each random
#' graph topology via \code{assign_weights}. Values greater than 1
#' indicate small-world structure
#'
#' \item \code{"omega"} --- Computes the \eqn{\omega} metric of
#' Telesford et al. (2011):
#' \deqn{\omega = \frac{L_{\text{rand}}}{L} - \frac{C}{C_{\text{latt}}}}
#' Random graphs preserve the degree sequence via edge-switching
#' (\code{\link[igraph]{sample_degseq}}, \code{method = "edge.switching.simple"}).
#' The clustering coefficient uses average local transitivity (\eqn{\bar{C}}).
#' Note: Telesford et al. (2011) defined and validated \eqn{\omega} exclusively
#' on binary, unweighted networks. When \code{weighted = TRUE}, edge weights
#' are reassigned to each random graph topology via \code{assign_weights} and
#' the lattice is constructed with \code{weighted = TRUE}, following the
#' approach of Muldoon et al. (2016); this is an extension beyond the original
#' formulation. Values near zero indicate small-world structure; negative
#' values indicate lattice-like structure; positive values indicate
#' random-like structure. Bounded in \eqn{[-1, 1]}
#'
#' \item \code{"SWI"} --- Computes the Small-World Index of
#' Neal (2015):
#' \deqn{\text{SWI} = \frac{L - L_{\text{latt}}}{L_{\text{rand}} - L_{\text{latt}}} \times \frac{C - C_{\text{rand}}}{C_{\text{latt}} - C_{\text{rand}}}}
#' Each term captures where the observed network's path length and clustering
#' coefficient fall within the lattice-to-random range, so that SWI = 1
#' only when \eqn{L = L_{\text{rand}}} and \eqn{C = C_{\text{latt}}}
#' simultaneously. Because this ideal is mathematically unachievable in a
#' finite network, SWI = 1 is a conceptual upper bound rather than a
#' realisable value. Random graphs preserve the degree sequence via
#' edge-switching (\code{\link[igraph]{sample_degseq}},
#' \code{method = "edge.switching.simple"}). The clustering coefficient uses
#' average local transitivity (\eqn{\bar{C}}). When \code{weighted = TRUE},
#' edge weights are reassigned to each random graph topology and the lattice
#' is constructed with \code{weighted = TRUE}. Note: SWI and SWP were
#' developed independently and concurrently; both apply the same
#' double-normalisation framework but differ in how the two deviation terms
#' are combined (product vs. Euclidean distance). Values close to 1 indicate
#' strong small-world structure; values near 0 indicate lattice-like or
#' random-like structure. Bounded in \eqn{[0, 1]}
#'
#' \item \code{"SWP"} --- Computes the Small-World Propensity of
#' Muldoon et al. (2016):
#' \deqn{\phi = 1 - \sqrt{\frac{\Delta_C^2 + \Delta_L^2}{2}}}
#' where \eqn{\Delta_C = (C_{\text{latt}} - C) / (C_{\text{latt}} - C_{\text{rand}})}
#' and \eqn{\Delta_L = (L - L_{\text{rand}}) / (L_{\text{latt}} - L_{\text{rand}})}.
#' Both \eqn{\Delta_C} and \eqn{\Delta_L} are bounded to \eqn{[0, 1]},
#' guaranteeing \eqn{\phi \in [0, 1]}. Random graphs preserve the degree
#' sequence via edge-switching. The clustering coefficient uses average
#' local transitivity (\eqn{\bar{C}}). When \code{weighted = TRUE}, edge
#' weights are reassigned to each random graph topology via
#' \code{assign_weights} and the lattice is constructed with
#' \code{weighted = TRUE}. Values close to 1 indicate strong small-world
#' structure; a pragmatic threshold of \eqn{\phi_T = 0.60} has been suggested
#'
#' }
#'
#' @param weighted Logical (length = 1).
#' Whether to compute small-worldness on the weighted network. When
#' \code{TRUE}, edge weights from \code{network} are preserved for the
#' empirical graph and reassigned to random and lattice graph topologies
#' via \code{assign_weights}, following Muldoon et al. (2016). When
#' \code{FALSE} (default), all graphs are treated as binary. Ignored when
#' \code{method = "analytical"}
#'
#' @param iter Numeric (length = 1).
#' Number of random graphs to generate when estimating the null baseline.
#' Defaults to \code{100}.
#' Ignored when \code{method = "analytical"}.
#' Higher values produce more stable estimates at the cost of computation time
#'
#' @return Numeric (length = 1).
#' The small-worldness value computed using the chosen \code{method}:
#'
#' \itemize{
#'
#' \item \code{"analytical"} and \code{"S"} --- \eqn{S > 1} indicates small-world structure
#'
#' \item \code{"omega"} --- Values near \eqn{|\omega| < 0.50} indicate small-world structure
#' (ranges between -1 and 1)
#'
#' \item \code{"SWI"} --- Values close to 1 indicate strong small-world structure;
#' values near 0 indicate lattice-like or random-like structure
#' (ranges between 0 and 1)
#'
#' \item \code{"SWP"} --- Values \eqn{\phi > 0.60} indicate strong small-world structure
#' (ranges between 0 and 1)
#'
#' }
#'
#' @author Alexander P. Christensen <alexpaulchristensen@gmail.com>
#'
#' @references
#' \strong{omega} \cr
#' Telesford, Q. K., Joyce, K. E., Hayasaka, S., Burdette, J. H., & Laurienti, P. J. (2011).
#' The ubiquity of small-world networks.
#' \emph{Brain Connectivity}, \emph{1}(5), 367--375.
#'
#' \strong{S and analytical} \cr
#' Humphries, M. D., & Gurney, K. (2008).
#' Network 'small-world-ness': A quantitative method for determining canonical network equivalence.
#' \emph{PLoS ONE}, \emph{3}(4), e0002051.
#'
#' \strong{SWI} \cr
#' Neal, Z. P. (2015).
#' Making big communities small: Using network science to understand the ecological and behavioral requirements for community social capital.
#' \emph{American Journal of Community Psychology}, \emph{55}(3), 369--380.
#'
#' \strong{SWP} \cr
#' Muldoon, S. F., Bridgeford, E. W., & Bassett, D. S. (2016).
#' Small-world propensity and weighted brain networks.
#' \emph{Scientific Reports}, \emph{6}(1), 22057.
#'
#' @examples
#' # Get network
#' network <- network_estimation(basic_smallworld)
#'
#' # Compute SWP (default)
#' swp <- smallworldness(network)
#'
#' # Compute omega
#' omega <- smallworldness(network, method = "omega")
#'
#' # Compute analytical S
#' S <- smallworldness(network, method = "analytical")
#'
#' # Compute simulated S
#' S_sim <- smallworldness(network, method = "S")
#'
#' # Compute SWI
#' swi <- smallworldness(network, method = "SWI")
#'
#' # Compute weighted SWP
#' swp_w <- smallworldness(network, weighted = TRUE)
#'
#' # Compute weighted omega
#' omega_w <- smallworldness(network, method = "omega", weighted = TRUE)
#'
#' @export
#'
# Compute smallworldness ----
# Updated 25.03.2026
smallworldness <- function(
    network, lattice, method = c("analytical", "omega", "S", "SWI", "SWP"),
    weighted = FALSE, iter = 100
)
{

  # Ensure network is in absolute values
  network <- abs(network)

  # Check for missing arguments (argument, default, function)
  # Uses actual function they will be used in
  method <- set_default(method, "swp", smallworldness)

  # Obtain adjacency matrix
  A <- network != 0

  # Check for weights
  if(!weighted){
    network <- A
  }

  # Obtain edges
  degree <- colSums(A, na.rm = TRUE)

  # Get nodes and degree
  nodes  <- length(degree)

  # Set up distance matrix
  node_sequence <- seq_len(nodes)
  distance_matrix <- abs(outer(node_sequence, node_sequence, "-"))
  distance_matrix <- pmin(distance_matrix, nodes - distance_matrix)

  # Convert network to {igraph}
  I <- convert2igraph(abs(network))

  # Set up function switch
  smallworld_FUN <- switch(
    method,
    "analytical" = smallworldness_analytical,
    "omega" = smallworldness_omega,
    "s" = smallworldness_S,
    "swi" = smallworldness_SWI,
    "swp" = smallworldness_SWP
  )

  # Obtain empirical values
  ASPL <- igraph::mean_distance(I)
  CC <- igraph::transitivity(I, type = swiftelse(method == "s", "global", "average"))
  # Switch CC based on method

  # Compute smallworldness
  return(
    smallworld_FUN(
      network = network, I = I, ASPL = ASPL, CC = CC, degree = degree,
      lattice = lattice, weighted = weighted, distance_matrix = distance_matrix,
      iter = iter
    )
  )

}

#' @noRd
# Analytical ----
# Updated 22.03.2026
smallworldness_analytical <- function(network, ASPL, CC, degree, ...)
{

  # Obtain nodes and average degree
  nodes <- dim(network)[2]
  mean_degree <- mean(degree, na.rm = TRUE)

  # Return analytical
  return(
    (CC / (mean_degree / nodes)) / (ASPL / (log(nodes) / log(mean_degree)))
  )

}

#' @noRd
# Telesford et al. (2011) ----
# Updated 25.03.2026
smallworldness_omega <- function(
    network, I, ASPL, CC, degree, lattice, weighted, distance_matrix, iter, ...
)
{

  # Collect random graphs
  random_graphs <- lapply(
    seq_len(iter), function(i){
      igraph::sample_degseq(out.deg = degree, method = "edge.switching.simple")
    }
  )

  # Check for weighted
  if(weighted){

    # Obtain weighted graphs
    random_graphs <- lapply(random_graphs, function(x){

      # Get weighted graphs
      weighted_graph <- assign_weights(network, igraph2matrix(x), distance_matrix)

      # Return back as {igraph}
      return(convert2igraph(weighted_graph))

    })

  }

  # Compute ASPL
  random_ASPL <- mean(nvapply(random_graphs, igraph::mean_distance), na.rm = TRUE)

  # Check if lattice is missing
  if(missing(lattice)){
    lattice_CC <- attributes(proxswap_lattice(network, weighted = weighted))$CC
  }else{

    # Check for weighted
    if(weighted){
      lattice <- assign_weights(network, lattice, distance_matrix)
    }

    lattice_CC <- igraph::transitivity(convert2igraph(lattice), type = "average")
  }

  # Return omega
  return((random_ASPL / ASPL) - (CC / lattice_CC))

}

#' @noRd
# Humphries & Gurney (2008) ----
# Updated 25.03.2026
smallworldness_S <- function(network, I, ASPL, CC, weighted, distance_matrix, iter, ...)
{

  # Obtain nodes and edges
  nodes <- dim(network)[2]
  edges <- sum(network[lower.tri(network)] != 0)

  # Obtain random graphs
  random_graphs <- lapply(seq_len(iter), function(i){
    igraph::sample_gnm(n = nodes, m = edges)
  })

  # Check for weighted
  if(weighted){

    # Obtain weighted graphs
    random_graphs <- lapply(random_graphs, function(x){

      # Get weighted graphs
      weighted_graph <- assign_weights(network, igraph2matrix(x), distance_matrix)

      # Return back as {igraph}
      return(convert2igraph(weighted_graph))

    })

  }

  # Compute S
  S <- nvapply(
    random_graphs, function(x){
      (CC / igraph::transitivity(x, type = "global")) / (ASPL / igraph::mean_distance(x))
    }
  )

  # Return S
  return(mean(S, na.rm = TRUE))

}

#' @noRd
# Neal (2015) ----
# Updated 25.03.2026
smallworldness_SWI <- function(
    network, I, ASPL, CC, degree, lattice, weighted, distance_matrix, iter, ...
)
{

  # Collect random graphs
  random_graphs <- lapply(
    seq_len(iter), function(i){
      igraph::sample_degseq(out.deg = degree, method = "edge.switching.simple")
    }
  )

  # Check for weighted
  if(weighted){

    # Obtain weighted graphs
    random_graphs <- lapply(random_graphs, function(x){

      # Get weighted graphs
      weighted_graph <- assign_weights(network, igraph2matrix(x), distance_matrix)

      # Return back as {igraph}
      return(convert2igraph(weighted_graph))

    })

  }

  # Compute ASPL and CC
  random_ASPL <- mean(nvapply(random_graphs, igraph::mean_distance), na.rm = TRUE)
  random_CC <- mean(nvapply(random_graphs, igraph::transitivity, type = "average"), na.rm = TRUE)

  # Check if lattice is missing
  if(missing(lattice)){
    lattice <- proxswap_lattice(network, weighted = weighted)
  }else if(weighted){
    lattice <- assign_weights(network, lattice, distance_matrix)
  }

  # Ensure lattice is {igraph}
  lattice <- convert2igraph(lattice)

  # Compute ASPL and CC
  lattice_ASPL <- igraph::mean_distance(lattice)
  lattice_CC <- igraph::transitivity(lattice, type = "average")

  # Return SWI
  return(
    ((ASPL - lattice_ASPL) / (random_ASPL - lattice_ASPL)) *
    ((CC - random_CC) / (lattice_CC - random_CC))
  )

}

#' @noRd
# Boundaries for SWP ----
# Updated 22.03.2026
boundary_values <- function(empirical, random, regular)
{
  any(is.infinite(c(empirical, random, regular))) | any(is.na(c(empirical, random, regular)))
}

#' @noRd
# Muldoon et al. (2016) ----
# Updated 25.03.2026
smallworldness_SWP <- function(
    network, I, ASPL, CC, degree, lattice, weighted, distance_matrix, iter, ...
)
{

  # Collect random graphs
  random_graphs <- lapply(
    seq_len(iter), function(i){
      igraph::sample_degseq(out.deg = degree, method = "edge.switching.simple")
    }
  )

  # Check for weighted
  if(weighted){

    # Obtain weighted graphs
    random_graphs <- lapply(random_graphs, function(x){

      # Get weighted graphs
      weighted_graph <- assign_weights(network, igraph2matrix(x), distance_matrix)

      # Return back as {igraph}
      return(convert2igraph(weighted_graph))

    })

  }

  # Compute ASPL and CC
  random_ASPL <- mean(nvapply(random_graphs, igraph::mean_distance), na.rm = TRUE)
  random_CC <- mean(nvapply(random_graphs, igraph::transitivity, type = "average"), na.rm = TRUE)

  # Check if lattice is missing
  if(missing(lattice)){
    lattice <- proxswap_lattice(network, weighted = weighted)
  }else if(weighted){
    lattice <- assign_weights(network, lattice, distance_matrix)
  }

  # Ensure lattice is {igraph}
  lattice <- convert2igraph(lattice)

  # Compute ASPL and CC
  lattice_ASPL <- igraph::mean_distance(lattice)
  lattice_CC <- igraph::transitivity(lattice, type = "average")

  # Compute deltas
  delta_ASPL <- max(0, (ASPL - random_ASPL)) / (lattice_ASPL - random_ASPL)
  delta_CC <- max(0, (lattice_CC - CC)) / (lattice_CC - random_CC)

  # Boundary check
  delta_ASPL <- min(1, swiftelse(boundary_values(ASPL, random_ASPL, lattice_ASPL), 1, delta_ASPL))
  delta_CC <- min(1, swiftelse(boundary_values(CC, random_CC, lattice_CC), 1, delta_CC))

  # Return SWP
  return(1 - sqrt((delta_ASPL^2 + delta_CC^2) / 2))

}

#%%%%%%%%%%%%%%%%%%%%%%%%%%#
#### Proximal Operators ####
#%%%%%%%%%%%%%%%%%%%%%%%%%%#

#' @noRd
# Updated 13.01.2026
atan_proximal <- function(x, lambda, gamma = 0.01, ...)
{
  return(l1_proximal(x, atan_derivative(x, lambda, gamma)))
}

#' @noRd
# Updated 05.02.2026
exp_proximal <- function(x, lambda, gamma = 0.01, ...)
{
  return(l1_proximal(x, exp_derivative(x, lambda, gamma)))
}

#' @noRd
# Updated 05.02.2026
gumbel_proximal <- function(x, lambda, gamma = 0.01, ...)
{
  return(l1_proximal(x, gumbel_derivative(x, lambda, gamma)))
}

#' @noRd
# Updated 25.07.2025
l1_proximal <- function(x, lambda, ...)
{
  return(sign(x) * pmax(abs(x) - lambda, 0))
}

#' @noRd
# Updated 08.03.2026
log_proximal <- function(x, lambda, gamma = 0.10, ...)
{
  return(l1_proximal(x, log_derivative(x, lambda, gamma)))
}

#' @noRd
# Updated 04.03.2026
weibull_proximal <- function(x, lambda, gamma = 0.01, shape, ...)
{
  return(l1_proximal(x, weibull_derivative(x, lambda, gamma, shape)))
}

.onload <- function(libname, pkgname)
{
    library.dynam("L0ggm",package=pkgname,lib.loc=libname)
}

.onAttach <- function(libname, pkgname)
{
    msg <- styletext(styletext(paste("\nL0ggm (version ", packageVersion("L0ggm"), ")", sep = ""), defaults = "underline"), defaults = "bold")
    # msg <- paste(msg,'\n\nFor help getting started, see <https://r-ega.net>')
    msg <- paste(msg,"\n\nFor bugs and errors, submit an issue to <https://github.com/AlexChristensen/L0ggm/issues>")

    packageStartupMessage(msg)
}

#' @title Confusion Matrix Metrics for Edge Comparison and Recovery
#'
#' @description Computes many commonly used confusion matrix metrics
#'
#' @param base Matrix or data frame.
#' Network that will be treated as the "ground truth" such that
#' a false positive represents an edge that is present in \code{comparison}
#' but not in this network
#'
#' @param comparison Matrix or data frame.
#' Network that will be treated as the estimator such that a
#' false positive represents an edge that is present in this network
#' but not in \code{base}
#'
#' @param metric Character vector.
#' Defaults to \code{"all"} metrics.
#' Available options:
#'
#' \itemize{
#'
#' \item \code{"all"} --- All available metrics (default)
#'
#' \item \code{"sen"} --- Sensitivity (True Positive Rate):
#' \deqn{\frac{TP}{TP + FN}}
#'
#' \item \code{"spec"} --- Specificity (True Negative Rate):
#' \deqn{\frac{TN}{TN + FP}}
#'
#' \item \code{"ppv"} --- Positive Predictive Value (Precision):
#' \deqn{\frac{TP}{TP + FP}}
#'
#' \item \code{"npv"} --- Negative Predictive Value:
#' \deqn{\frac{TN}{TN + FN}}
#'
#' \item \code{"fdr"} --- False Discovery Rate:
#' \deqn{1 - PPV = \frac{FP}{TP + FP}}
#'
#' \item \code{"fom"} --- False Omission Rate:
#' \deqn{1 - NPV = \frac{FN}{TN + FN}}
#'
#' \item \code{"ba"} --- Balanced Accuracy:
#' \deqn{\frac{Sensitivity + Specificity}{2}}
#'
#' \item \code{"f1"} --- F1 Score (harmonic mean of PPV and Sensitivity):
#' \deqn{\frac{2TP}{2TP + FP + FN}}
#'
#' \item \code{"csi"} --- Critical Success Index (Jaccard / Threat Score):
#' \deqn{\frac{TP}{TP + FP + FN}}
#'
#' \item \code{"mcc"} --- Matthews Correlation Coefficient:
#' \deqn{\frac{TP \times TN - FP \times FN}{\sqrt{(TP+FP)(TP+FN)(TN+FP)(TN+FN)}}}
#'
#' }
#'
#' @param full.names Boolean (length = 1).
#' Whether full or abbreviated names should be used.
#' Defaults to \code{FALSE}.
#' Set to \code{TRUE} for full names
#'
#' @return A named numeric vector of the requested confusion matrix metrics.
#' Values are generally in \eqn{[0, 1]}, with the exception of \code{"mcc"}
#' (Matthews Correlation Coefficient), which ranges from \eqn{-1} to \eqn{1}
#' where 1 indicates perfect agreement, 0 indicates chance-level performance,
#' and -1 indicates perfect disagreement. The vector contains only the elements
#' specified by \code{metric} (all ten metrics when \code{metric = "all"},
#' which is the default). Names are abbreviated (e.g., \code{"sen"}) when
#' \code{full.names = FALSE} (default), or expanded
#' (e.g., \code{"Sensitivity"}) when \code{full.names = TRUE}. Any metric
#' whose denominator is zero (e.g., sensitivity when no edges exist in
#' \code{base}) is returned as \code{NA}.
#'
#' @author Alexander P. Christensen <alexpaulchristensen@gmail.com>
#'
#' @examples
#' # Set split
#' split <- sample(
#'   1:nrow(basic_smallworld),
#'   round(nrow(basic_smallworld) / 2)
#' )
#'
#' # Estimate networks
#' split1 <- network_estimation(basic_smallworld[split,])
#' split2 <- network_estimation(basic_smallworld[-split,])
#'
#' # Estimate metrics
#' edge_confusion(split1, split2)
#'
#' @export
#'
# Compute confusion matrix metrics ----
# Updated 26.02.2026
edge_confusion <- function(base, comparison, metric = c(
  "all", "sen", "spec",
  "ppv", "npv", "fdr", "fom",
  "ba", "f1", "csi", "mcc"
), full.names = FALSE)
{

  # Set names
  nice_names <- c(
    "sen" = "Sensitivity", "spec" = "Specificity", "ppv" = "Positive Predictive Value",
    "npv" = "Negative Predictive Value", "fdr" = "False Discovery Rate", "fom" = "False Omission Rate",
    "ba" = "Balanced Accuracy", "f1" = "F1 Score", "csi" = "Critical Success Index (Jaccard)",
    "mcc" = "Matthew's Correlation Coefficient"
  )

  # Set default
  metric <- set_default(metric, "all", edge_confusion, several.ok = TRUE)

  # Check for all
  if("all" %in% metric){
    metric <- c("sen", "spec", "ppv", "npv", "fdr", "fom", "ba", "f1", "csi", "mcc")
  }

  # Obtain lower triangle indices and networks
  lower_triangle <- lower.tri(base)
  base_lower <- base[lower_triangle]
  comp_lower <- comparison[lower_triangle]

  # Set up indices
  base_present <- base_lower != 0
  base_absent <- !base_present
  comp_present <- comp_lower != 0
  comp_absent <- !comp_present

  # Set up basic confusion matrix metrics
  TP <- sum(base_present & comp_present)
  FP <- sum(base_absent & comp_present)
  TN <- sum(base_absent & comp_absent)
  FN <- sum(base_present & comp_absent)

  # Totals
  P <- TP + FN
  N <- TN + FP
  TP2 <- 2 * TP

  # Check for zero denominators
  TPFP <- TP + FP
  TNFN <- TN + FN
  FPFN <- FP + FN
  TP2FPFN <- TP2 + FPFN
  TPFPFN <- TP + FPFN

  # Compute metrics
  sen <- swiftelse(P == 0, NA, TP / P)
  spec <- swiftelse(N == 0, NA, TN / N)
  ppv <- swiftelse(TPFP == 0, NA, TP / TPFP)
  npv <- swiftelse(TNFN == 0, NA, TN / TNFN)

  # Create metric vector
  metric_vector <- c(
    "sen" = sen, "spec" = spec, "ppv" = ppv, "npv" = npv,
    "fdr" = 1 - ppv, "fom" = 1 - npv, "ba" = (sen + spec) / 2,
    "f1" = swiftelse(TP2FPFN == 0, NA, TP2 / TP2FPFN),
    "csi" = swiftelse(TPFPFN == 0, NA, TP / TPFPFN),
    "mcc" = silent_call(cor(base_present, comp_present)) # equivalent
  )

  # Return metrics
  return(
    structure(
      metric_vector[metric],
      names = swiftelse(full.names, nice_names[metric], metric)
    )
  )

}

#%%%%%%%%%%%%%%%%%%%#
#### Derivatives ####
#%%%%%%%%%%%%%%%%%%%#

#' @noRd
# Updated 15.03.2026
atan_derivative <- function(x, lambda, gamma = 0.01, ...)
{
  # return(lambda * (gamma * (gamma + 2 / pi)) / (gamma^2 + x^2))

  # Return derivative
  return(
    .Call(
      "atan_derivative_c",
      as.numeric(x), lambda_ = lambda, gamma_ = gamma,
      PACKAGE = "L0ggm"
    )
  )

}

#' @noRd
# Updated 15.03.2026
exp_derivative <- function(x, lambda, gamma = 0.01, ...)
{
  # return(lambda * (1 / gamma) * exp(-(abs(x) / gamma)))

  # Return derivative
  return(
    .Call(
      "exp_derivative_c",
      as.numeric(x), lambda_ = lambda, gamma_ = gamma,
      PACKAGE = "L0ggm"
    )
  )

}

#' @noRd
# Updated 15.03.2026
gumbel_derivative <- function(x, lambda, gamma = 0.01, ...)
{

  # # Pre-compute values
  # gamma_x <- abs(x) / gamma
  #
  # # Return derivative
  # return((lambda / gamma) * (exp(-gamma_x - exp(-gamma_x)) + exp(gamma_x - exp(gamma_x))))

  # Return derivative
  return(
    .Call(
      "gumbel_derivative_c",
      as.numeric(x), lambda_ = lambda, gamma_ = gamma,
      PACKAGE = "L0ggm"
    )
  )

}

#' @noRd
# Updated 15.03.2026
log_derivative <- function(x, lambda, gamma = 0.10, ...)
{
  # return(lambda / ((gamma + abs(x)) * log(1 + 1 / gamma)))

  # Return derivative
  return(
    .Call(
      "log_derivative_c",
      as.numeric(x), lambda_ = lambda, gamma_ = gamma,
      PACKAGE = "L0ggm"
    )
  )

}

#' @noRd
# Updated 15.03.2026
weibull_derivative <- function(x, lambda, gamma = 0.01, shape, ...)
{

  # # Pre-compute components
  # abs_x <- pmax.int(abs(x), .Machine$double.eps)
  # x_gamma <- abs_x / gamma
  #
  # # Check for shape greater than 1
  # if(shape > 1){
  #
  #   # Pre-compute shape
  #   shape_one <- shape - 1
  #
  #   # Obtain peak
  #   peak <- gamma * ((shape - 1) / shape)^(1 / shape)
  #
  #   # Compute peak gamma
  #   peak_gamma <- peak / gamma
  #
  #   # Compute values
  #   peak_value <- (shape / gamma) * peak_gamma^shape_one * exp(-peak_gamma^shape)
  #
  #   # Return
  #   return(
  #     lambda * swiftelse(
  #       abs_x <= peak, peak_value,
  #       (shape / gamma) * x_gamma^shape_one * exp(-x_gamma^shape)
  #     )
  #   )
  #
  # }
  #
  # # Return derivative
  # return(lambda * (shape / gamma) * x_gamma^(shape - 1) * exp(-x_gamma^shape))

  # Return derivative
  return(
    .Call(
      "weibull_derivative_c",
      as.numeric(x), lambda_ = lambda,
      gamma_ = gamma, shape_ = shape,
      PACKAGE = "L0ggm"
    )
  )

}

#' Toy Small-world Network Data Example
#'
#' A toy small-world network example (\emph{n} = 500)
#'
#' @name basic_smallworld
#'
#' @docType data
#'
#' @usage data(basic_smallworld)
#'
#' @format A 500 \eqn{\times} 20 continuous response matrix
#'
#' @keywords datasets
#'
#' @examples
#' # Generated using:
#' # basic_smallworld <- simulate_smallworld(
#' #  nodes = 20, # 20 nodes in the network
#' #  density = 0.30, # moderate initial lattice connectivity
#' #  rewire = 0.20, # 20% rewiring probability
#' #  sample_size = 500 # number of cases = 500
#' # )$data
#' data("basic_smallworld")
#'
NULL
#----

#%%%%%%%%%%%%%%%%%%%%%%%#
#### {L0ggm} Helpers ####
#%%%%%%%%%%%%%%%%%%%%%%%#

# These helpers are internal functions that are repeatedly
# used across the package. They are implemented here once
# so that changes only need to be made here and not in
# multiple places across the package.
#
# The goal is to minimize redundant calls to improve reliability of code.
# Sections are separated by how they are associated with their utility
# in other functions.

#%%%%%%%%%%%%%%%%%%%%%%
# *APPLY FUNCTIONS ----
#%%%%%%%%%%%%%%%%%%%%%%

# These functions are merely wrappers over the `*apply`
# family to execute specific tasks that are often repeatedly
# done. Their purpose are to provide slightly cleaner code.

# These `*vapply` are specific versions of `vapply` that
# pre-specify the output. These versions are generally
# simpler (in function) and marginally faster than
# the more commonly used `sapply`. They are more strict
# in that they require knowing the output format (e.g., numeric)

#' @noRd
# Character vector/matrix output apply ----
# Updated 06.07.2023
cvapply <- function(X, FUN, ..., LENGTH = 1, USE.NAMES = TRUE)
{
  return(vapply(X = X, FUN = FUN, FUN.VALUE = character(LENGTH), ..., USE.NAMES = USE.NAMES))
}

#' @noRd
# Logical vector/matrix output apply ----
# Updated 06.07.2023
lvapply <- function(X, FUN, ..., LENGTH = 1, USE.NAMES = TRUE)
{
  return(vapply(X = X, FUN = FUN, FUN.VALUE = logical(LENGTH), ..., USE.NAMES = USE.NAMES))
}

#' @noRd
# Numeric vector/matrix output apply ----
# Updated 06.07.2023
nvapply <- function(X, FUN, ..., LENGTH = 1, USE.NAMES = TRUE)
{
  return(vapply(X = X, FUN = FUN, FUN.VALUE = numeric(LENGTH), ..., USE.NAMES = USE.NAMES))
}

#' @noRd
# Unlist `lapply` ----
# Updated 14.07.2023
ulapply <- function(X, FUN, ..., recursive = TRUE)
{
  return(unlist(lapply(X, FUN, ...), recursive = recursive))
}

# The `row_apply` and `column_apply` functions are
# not necessary faster than `apply` but they are
# more strict in that they return a matrix with
# the same dimensions that was input into 'X'.
# Therefore, there is never a need to transpose
# the output because it will always have the same
# dimensions as what went in

#' @noRd
# Stricter `apply` by row ----
# Slightly faster than `apply` but not appreciably
# Updated 06.07.2023
row_apply <- function(X, FUN, ...)
{

  # Ensure matrix
  X <- as.matrix(X)

  # Get dimensions of data
  dimensions <- dim(X)

  # Get appropriate function
  vapply_FUN <- switch( # in order of most likely
    typeof(X),
    "double" = nvapply,
    "integer" = nvapply,
    "logical" = lvapply,
    "character" = cvapply,
    stop(
      paste0(
        "\"", typeof(X), "\"",
        " is not available for `row_apply`. Only \"character\", \"logical\", and \"numeric\" are available"
      ), call. = FALSE
    )
  )

  # Return call
  return(
    matrix(
      vapply_FUN(
        split(X, seq_len(dimensions[1])),
        FUN = FUN, ..., LENGTH = dimensions[2],
        USE.NAMES = FALSE # names are supplied below
      ),
      dimensions, dimnames = dimnames(X), byrow = TRUE
    )
  )

}

#' @noRd
# Stricter `apply` by column ----
# Slightly faster than `apply` but not appreciably
# Updated 06.07.2023
column_apply <- function(X, FUN, ...)
{

  # Ensure matrix
  X <- as.matrix(X)

  # Get dimensions of data
  dimensions <- dim(X)

  # Get appropriate function
  vapply_FUN <- switch( # in order of most likely
    typeof(X),
    "double" = nvapply,
    "integer" = nvapply,
    "logical" = lvapply,
    "character" = cvapply,
    stop(
      paste0(
        "\"", typeof(X), "\"",
        " is not available for `row_apply`. Only \"character\", \"logical\", and \"numeric\" are available"
      ), call. = FALSE
    )
  )

  # Return call
  return(
    matrix(
      vapply_FUN(
        split(t(X), seq_len(dimensions[2])),
        FUN = FUN, ..., LENGTH = dimensions[1],
        USE.NAMES = FALSE # names are supplied below
      ),
      dimensions, dimnames = dimnames(X), byrow = FALSE
    )
  )

}

#' @noRd
# 3D Array apply ----
# Apply's a function across a 3D array to obtain a 2D array
# About 2x faster than `apply(simplify2array(X), 1:2, FUN)`
# Updated 06.07.2023
symmetric_matrix_lapply <- function(X, FUN, ...){

  # Get appropriate function
  vapply_FUN <- switch( # in order of most likely
    typeof(X[[1]]),
    "double" = nvapply,
    "integer" = nvapply,
    "logical" = lvapply,
    "character" = cvapply,
    stop(
      paste0(
        "\"", typeof(X[[1]]), "\"",
        " is not available for `symmetric_matrix_lapply`. Only \"character\", \"logical\", and \"numeric\" are available"
      ), call. = FALSE
    )
  )

  # Get dimensions of single matrix in list
  dimensions <- dim(X[[1]])

  # Pre-obtain lower triangle indices
  # Equivalent to: (`lower.tri(matrix, diag = TRUE)`)
  lower_triangle_index <- .row(dimensions) >= .col(dimensions)

  # Get lower triangles
  lower_matrix <- do.call(rbind, lapply(X, function(x){x[lower_triangle_index]}))

  # Apply function
  values <- vapply_FUN(as.data.frame(lower_matrix), FUN, ...)

  # Initialize return matrix
  new_matrix <- matrix(
    nrow = dimensions[1], ncol = dimensions[2],
    dimnames = dimnames(X[[1]])
  )

  # Add values to lower triangle
  new_matrix[lower_triangle_index] <- values

  # Transpose
  new_matrix <- t(new_matrix)

  # Add values again to lower triagle
  new_matrix[lower_triangle_index] <- values

  # Return new matrix
  return(new_matrix)

}

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# REPRODUCIBLE RANDOM GENERATION ----
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

# For these random generation functions, the goal
# is reproducibility, which runs counter to "random".
# However, for bootstrap results to be reproduced
# there is a need to generate "random" samples that
# can be reproduced so that the user arrives at the
# same results *without* influencing R's random number
# generation or changing a user-defined seed previous
# to, while, or after using {L0ggm} functions.
#
# Assistance in writing C and interfacing with R
# was provided by GPT-4

#' @noRd
# Generate integer seeds ----
# (non-randomly based on `seed`)
# `NULL` is reserved for actual pseudo-randomness
# Uses xoshiro256++ random number generation: https://prng.di.unimi.it/
# Updated 27.07.2023
reproducible_seeds <- function(n, seed = NULL)
{

  # Return call from C
  return(
    .Call(
      "r_xoshiro_seeds",
      as.integer(n), swiftelse(is.null(seed), 0, seed),
      PACKAGE = "L0ggm"
    )
  )

}

#' @noRd
# Generate uniform data ----
# Allows adjustment of range
# Updated 26.10.2023
runif_xoshiro <- function(n, min = 0, max = 1, seed = NULL)
{

  # Get values
  values <- .Call(
    "r_xoshiro_uniform",
    as.integer(n),
    swiftelse(is.null(seed), 0, seed),
    PACKAGE = "L0ggm"
  )

  # Check for changes to minimum and maximum
  if(min != 0 || max != 1){ # transform
    values <- min + (max - min) * values
  }


  # Return call from C
  return(values)

}

#' @noRd
# Shuffle (without replacement) ----
# Uses xoshiro256++ random number generation: https://prng.di.unimi.it/
# Updated 30.07.2023
shuffle <- function(x, size = length(x), seed = NULL)
{

  # Return call from C
  return(
    x[.Call(
      "r_xoshiro_shuffle",
      as.integer(seq_along(x)),
      swiftelse(is.null(seed), 0, seed),
      PACKAGE = "L0ggm"
    )][seq_len(size)]
  )

}

#' @noRd
# Weighted Shuffle (without replacement) ----
# Uses xoshiro256++ random number generation: https://prng.di.unimi.it/
# Updated 14.04.2026
weighted_shuffle <- function(x, size = length(x), prob, seed = NULL)
{

  # Return call from C
  return(
    x[.Call(
      "r_xoshiro_weighted_shuffle",
      as.integer(seq_along(x)), prob,
      swiftelse(is.null(seed), 0, seed),
      PACKAGE = "L0ggm"
    )][seq_len(size)]
  )

}

#' @noRd
# Shuffle (with replacement) ----
# Uses xoshiro256++ random number generation: https://prng.di.unimi.it/
# Updated 10.03.2026
shuffle_replace <- function(x, size = length(x), seed = NULL)
{

  # Return call from C
  return(
    x[.Call(
      "r_xoshiro_shuffle_replace",
      x, size, swiftelse(is.null(seed), 0, seed),
      PACKAGE = "L0ggm"
    )]
  )

}

#' @noRd
# Random normal generation with Ziggurat ----
# https://people.sc.fsu.edu/~jburkardt/cpp_src/ziggurat/ziggurat.html
# Updated 27.07.2023
rnorm_ziggurat <- function(n, seed = NULL)
{

  # Return call from C
  return(
    .Call(
      "r_ziggurat",
      as.integer(n),
      swiftelse(is.null(seed), 0, seed),
      PACKAGE = "L0ggm"
    )
  )

}

#%%%%%%%%%%%%%%%%
# CATEGORIZE ----
#%%%%%%%%%%%%%%%%

#' @noRd
# Adds skew to a single variable ----
# Updated 30.11.2022
skew_single_variable <- function(data, skew_values){

  # Categorize biased data with updated thresholds
  for(i in (length(skew_values) + 1):1){

    # First category
    if(i == 1){
      data[data < skew_values[i]] <- i
    }else if(i == length(skew_values) + 1){ # Last category
      data[data >= skew_values[i-1]] <- i
    }else{ # Middle category
      data[data >= skew_values[i-1] & data < skew_values[i]] <- i
    }

  }

  # Return skewed data
  return(data)

}

# Based on
# Garrido, L. E., Abad, F. J., & Ponsoda, V. (2011).
# Performance of Velicer’s minimum average partial factor retention
# method with categorical variables.
# Educational and Psychological Measurement, 71(3), 551-570.
# https://doi.org/10.1177/0013164410389489
#
#' @noRd
# Generates skewed data for continuous data ----
# Updated 13.03.2026
skew_continuous <- function(skew, data = NULL, sample_size = 1e06, tolerance = 1e-05)
{

  # Generate data
  if(is.null(data)){
    data <- rnorm_ziggurat(sample_size)
  }

  # Check for zero skew (add minimal skew)
  if(skew == 0){
    return(data)
  }

  # Obtain sign and absolute skew
  skew_sign <- sign(skew)
  absolute <- abs(skew)

  # Compute arcsinh once
  arcsinh_data <- asinh(data)

  # Initialize kurtosis
  kurtosis <- 1

  # Optimize for result
  objective <- function(kurtosis){

    # Optimize for skew
    optimize(
      f = function(x){
        abs(absolute - psych::skew(sinh(kurtosis * (arcsinh_data + x))))
      }, interval = c(0, absolute + 2), tol = tolerance
    )

  }

  # Repeat until convergence
  repeat{

    # Get result
    result <- silent_call(objective(kurtosis))

    # Check for minimum too low
    if(is.na(result$objective)){
      return(data)
    }

    # Check for tolerance
    if(result$objective < tolerance){
      break
    }

    # Update kurtosis
    kurtosis <- kurtosis + result$objective * 0.25 # learning rate

  }

  # Return result
  return(skew_sign * scale(sinh(kurtosis * (arcsinh_data + result$minimum))))

}

# Based on
# Garrido, L. E., Abad, F. J., & Ponsoda, V. (2011).
# Performance of Velicer’s minimum average partial factor retention
# method with categorical variables.
# Educational and Psychological Measurement, 71(3), 551-570.
# https://doi.org/10.1177/0013164410389489
#
#' @noRd
# Generates skew ----
# Updated 09.08.2022
skew_generator <- function(
    skewness, categories,
    reduction_factor = 0.75,
    sample_size = 1000000,
    initial_proportion = 0.50,
    tolerance = 0.00001
)
{

  # Initialize skew matrix
  skew_matrix <- matrix(
    0, nrow = categories, ncol = categories
  )

  # Initialize cases
  cases <- numeric(sample_size)

  # Loop through categories
  for(i in 2:categories){

    # Initialize (largest category) proportion
    proportion <- initial_proportion

    # Current proportion for rest of categories
    remaining_proportion <- 1 - proportion

    # Initialize categories allocations
    allocation <- 1
    allocation_1 <- 1

    # Loop through with reduction factor
    if(i > 2){
      for(j in 1:(i-2)){
        allocation_1 <- allocation_1 * reduction_factor
        allocation <- allocation + allocation_1
      }
    }

    # Divide remaining proportion by allocations
    divided_proportion <- remaining_proportion / allocation

    # Undefined objects
    propinf <- 1 / sample_size
    propsup <- initial_proportion
    E <- divided_proportion / reduction_factor

    # Loop through
    for(j in 1:i){
      cases[round(sample_size * propinf):round(sample_size * propsup)] <- j
      E <- E * reduction_factor
      propinf <- propsup
      propsup <- propinf + E
    }

    # Compute skewness
    skew_actual <- psych::skew(cases)

    # Limits
    limitsup <- 1
    limitsinf <- 0

    # Ensure skew within tolerance
    while(abs(skew_actual - skewness) > tolerance){

      # Skew greater than
      if(skew_actual < skewness){
        limitsinf <- proportion
        proportion <- (proportion + limitsup) / 2
      }else{
        limitsup <- proportion
        proportion <- (proportion + limitsinf) / 2
      }

      # Update
      # Current proportion for rest of categories
      remaining_proportion <- 1 - proportion

      # Divide remaining proportion by allocations
      divided_proportion <- remaining_proportion / allocation

      # Undefined objects
      propinf <- 1 / sample_size
      propsup <- proportion
      E <- divided_proportion / reduction_factor

      # Loop through
      for(j in 1:i){
        cases[round(sample_size * propinf):round(sample_size * propsup)] <- j
        E <- E * reduction_factor
        propinf <- propsup
        propsup <- propinf + E
      }

      # Compute skewness
      skew_actual <- psych::skew(cases)

    }

    # Set E
    E <- divided_proportion / reduction_factor
    cumulative_probability <- proportion

    # Update matrix
    for(j in 1:i){

      skew_matrix[i,j] <- cumulative_probability
      E <- E * reduction_factor
      cumulative_probability <- cumulative_probability + E

    }

  }

  # Normal inverse
  norm_inv_matrix <- qnorm(skew_matrix)

  # Set infinite values to zero
  norm_inv_matrix[is.infinite(norm_inv_matrix)] <- 0

  # Category probability
  category_probability <- matrix(
    0, nrow = categories, ncol = categories
  )

  # Fill first column
  category_probability[,1] <- skew_matrix[,1]

  # Loop through
  for(i in 2:categories){
    category_probability[,i] <- skew_matrix[,i] - skew_matrix[,i-1]
  }

  # Make -1 = 0
  category_probability[category_probability == -1] <- 0

  # Return skew
  result <- list(
    skew_matrix = norm_inv_matrix,
    probability = category_probability
  )
  return(result)

}

#%%%%%%%%%%%%%%%%%%%%%%
# FASTER SEQUENCES ----
#%%%%%%%%%%%%%%%%%%%%%%

# These functions aren't often used because
# dimensions of the data are usually used
# elsewhere in {L0ggm}; however, these
# functions use `seq_len` and `dim` which
# have minimal speed advantages over their
# respective `1:nrow` and `1:ncol`
#
# The primary benefit of using this functions
# is `seq_len` which throws an error with
# `seq_len(0)` where `1:0` will not
# (at least not at the top of the call)

#' @noRd
# Faster row sequences ----
# Updated 06.07.2023
nrow_sequence <- function(data)
{
  return(seq_len(dim(data)[1]))
}


#' @noRd
# Faster column sequences ----
# Updated 06.07.2023
ncol_sequence <- function(data)
{
  return(seq_len(dim(data)[2]))
}

#' @noRd
# Faster `ifelse` ----
# For single value replacements:
# 1.5x faster with 1 value
# 2.5x faster with 10 values
# >= 18x faster with >= 100 values
# Updated 26.02.2026
swiftelse <- function(condition, true, false)
{
  # Get condition length
  condition_length <- length(condition)

  # Check for single value
  if(condition_length == 1){

    # If TRUE
    if(condition){
      return(true)
    }else{ # Otherwise, FALSE
      return(false)
    }

  }

  # Initialize result
  result <- vector(typeof(false), condition_length)

  # Set TRUE condition
  if(length(true) == 1){
    result[condition] <- true
  }else{
    result[condition] <- true[condition]
  }

  # Set FALSE condition
  if(length(false) == 1){
    result[!condition] <- false
  }else{

    # Get opposite condition (slightly faster than repeated calls)
    opposite_condition <- !condition
    result[opposite_condition] <- false[opposite_condition]
  }


  # Return result
  return(result)

}

#%%%%%%%%%%%%%%%%%%%%%%%%%%
# FORMATTING FUNCTIONS ----
#%%%%%%%%%%%%%%%%%%%%%%%%%%

#' @noRd
# Determine decimal places before non-zero ----
# From StackOverflow: https://stackoverflow.com/questions/35553244/count-leading-zeros-between-the-decimal-point-and-first-nonzero-digit
# Updated 10.11.2023
leading_zero <- function(number)
{
  floor(-log10(.Machine$double.eps + abs(number) - floor(abs(number))))
}

#' @noRd
# Determine number of digits in a number ----
# Updated 24.07.2023
digits <- function(number)
{
  # Obtain the lowest value of log base 10 and add 1
  return(floor(log10(abs(number))) + 1)

}

#' @noRd
# Determine format of bytes ----
# Updated 22.07.2023
byte_digits <- function(bytes)
{

  # Get number of digits
  ndigits <- digits(bytes)

  # Loop over values
  if(ndigits > 11){
    value <- paste0(round(bytes / 1e+12, 2), " TB")
  }else if(ndigits > 8){
    value <- paste0(round(bytes / 1e+09, 2), " GB")
  }else if(ndigits > 5){
    value <- paste0(round(bytes / 1e+06, 2), " MB")
  }else if(ndigits > 2){
    value <- paste0(round(bytes / 1e+03, 2), " KB")
  }else{
    value <- paste0(round(bytes, 2), " B")
  }

  # Return configured value
  return(value)

}

#' @noRd
# Format number with certain decimals ----
# Mainly for naming and printing
# Updated 14.06.2024
format_decimal <- function(numbers, places)
{

  return(
    formatC(
      x = numbers,
      digits = places,
      format = "f", flag = "0"
    )
  )

}

#' @noRd
# Format number with certain integer ----
# Mainly for naming and printing
# Updated 14.06.2024
format_integer <- function(numbers, places)
{

  return(
    formatC(
      x = numbers,
      digits = places,
      format = "d", flag = "0"
    )
  )

}

#%%%%%%%%%%%%%%%%%%%%%%
# OBJECT FUNCTIONS ----
#%%%%%%%%%%%%%%%%%%%%%%

#' @noRd
# Faster data frame initialization ----
# Initializes a matrix and converts to a data frame
# Data frames are initialized as individual vectors
# carrying extra overhead
# Updated 06.07.2023
fast.data.frame <- function(
    data = NA, nrow = 1, ncol = 1,
    rownames = NULL, colnames = NULL,
    ...
)
{

  return(
    as.data.frame(
      matrix(
        data = data,
        nrow = nrow, ncol = ncol,
        dimnames = list(rownames, colnames)
      ), make.names = FALSE,
      ...
    )
  )

}

#' @noRd
# Determine object type ----
# Updated 05.07.2023
get_object_type <- function(object)
{

  # Remove attributes and class from object
  unclassed_object <- remove_attributes(object)

  # Add object type to object's attributes
  ## In order of precedence
  if(is(object, "tbl")){ # needs precedence over 'data.frame'
    return("tibble")
  }else if(is.data.frame(object)){ # needs precedence over 'list'
    return("data.frame")
  }else if(is.list(unclassed_object)){ # needs precedence over 'vector'
    return("list")
  }else if(is.factor(object)){
    return("factor")
  }else if(is.vector(unclassed_object)){
    return("vector")
  }else if(is.matrix(unclassed_object)){
    return("matrix")
  }else if(is.array(unclassed_object)){
    return("array")
  }

}

#' @noRd
# Force vector ----
# For 1D matrix or data frames only
# Updated 05.07.2023
force_vector <- function(object)
{

  # Get object type
  object_type <- get_object_type(object)

  # Branch across types
  if(object_type == "vector"){
    return(object) # already a vector
  }else if(object_type %in% c("matrix", "data.frame", "tibble", "factor")){

    # Ensure matrix
    new_matrix <- as.matrix(object)

    # Get dimensions
    dimensions <- dim(new_matrix)

    # Check for more than one dimension
    if(all(dimensions) > 1){
      stop("Cannot sufficiently force object into 'vector'.", call. = FALSE)
    }

    # Set as vector
    new_vector <- as.vector(new_matrix)

    # Set names based on dimension
    names(new_vector) <- unlist(dimnames(object)[dimensions != 1])

    # Return new vector
    return(new_vector)

  }

}

#' @noRd
# Force numeric ----
# For factors
# Updated 05.07.2023
force_numeric <- function(object)
{

  # Check for whether object is already numeric
  if(is.numeric(object)){
    return(object) # just return
  }else if(is.factor(object)){ # factors are trickier...

    # Switch based on type in the levels
    if(is.numeric(levels(object))){

      # Get original values rather than factor values
      return(as.numeric(as.character(object)))

    }else{

      # Just used factor assigned values
      return(as.numeric(object))

    }

  }

}

#' @noRd
# Convert matrix to {igraph} network ----
# Updated 09.08.2023
convert2igraph <- function (A, diagonal = 0)
{

  # Convert to matrix
  A <- as.matrix(A)

  # Change diagonal (to zero)
  diag(A) <- diagonal

  # Return {igraph} network
  return(
    silent_call(
      igraph::graph_from_adjacency_matrix(
        A, weighted = TRUE, mode = "undirected",
        add.colnames = FALSE
      )
    )
  )

}

#' @noRd
# Convert {igraph} network to matrix ----
# Updated 09.08.2023
igraph2matrix <- function(I)
{
  return(as.matrix(igraph::as_adjacency_matrix(I)))
}

#%%%%%%%%%%%%%%%%%%%%%
# COUNT FUNCTIONS ----
#%%%%%%%%%%%%%%%%%%%%%

#' @noRd
# Unique Length ----
# Counts length of unique variables (without `NA`)
# Updated 03.07.2023
unique_length <- function(x)
{
  return(length(unique(x)) - anyNA(x))
}

#' @noRd
# Edge count ----
# Counts the number of edges
# (assumes network is _symmetric_ matrix)
# Updated 11.10.2023
edge_count <- function(network, nodes, diagonal = FALSE)
{
  return((sum(network != 0, na.rm = TRUE) - swiftelse(diagonal, nodes, 0)) * 0.50)
}

#' @noRd
# Faster table ----
# Skips many checks in `table`
# Updated 22.07.2023
fast_table <- function(values)
{

  # Get factor values
  factor_values <- swiftelse(
    is.factor(values), values, factor(values)
  )

  # Get factor levels
  factor_levels <- levels(factor_values)

  # Return with names
  return(
    setNames(
      tabulate(factor_values, length(factor_levels)),
      factor_levels
    )
  )

}

#' @noRd
# Converts entire vectors to a single value ----
# Same logic as {plyr}'s `id_var` and `ninteraction`
# Avoids {plyr} dependency for single function
# Updated 06.07.2023
vector2factor <- function(data)
{

  # Use "paste" method
  stringed_values <- do.call(
    # Get total unique values each variable takes
    paste, rev(lapply( # `ninteraction`

      # Ensure data frame
      as.data.frame(data), function(x){

        # Get matched ID (`id_var`)
        matched_id <- match(x, sort(unique(x), na.last = TRUE))

        # Attached number of unique as attribute
        attr(matched_id, "n") <- max(matched_id)

        # Return result
        return(matched_id)

      }))
  )

  # Return unique ID
  return(factor(match(stringed_values, unique(stringed_values))))

}

#' @noRd
# Count table ----
# Provides counts of repeated rows in a data frame
# Same logic as {plyr}'s `count`
# Avoids {plyr} dependency for single function
# Updated 22.07.2023
count_table <- function(data, proportion = FALSE)
{

  # Ensure data frame
  data <- as.data.frame(data)

  # Get unique ID
  unique_id <- vector2factor(data)

  # Get unique solutions
  unique_solutions <- data[!duplicated(unique_id),,drop = FALSE]

  # Tabulate data
  Value <- fast_table(unique_id) # tabulate(unique_id, length(levels(unique_id)))

  # Check for proportion
  if(proportion){
    Value <- Value / dim(data)[1]
  }

  # Return data frame
  return(
    as.data.frame(
      cbind(unique_solutions[order(unique(unique_id)),], Value)
    )
  )

}

#%%%%%%%%%%%%%%%%%%%%
# DATA FUNCTIONS ----
#%%%%%%%%%%%%%%%%%%%%

#' @noRd
# Determine number of categories in data ----
# Updated 06.07.2023
data_categories <- function(data)
{
  return(nvapply(as.data.frame(data), unique_length))
}

#' @noRd
# All-purpose symmetric checker ----
# Faster but less robust than `isSymmetric`
# Defaults to floating point tolerance
# Updated 01.07.2023
is_symmetric <- function(data, tolerance = 1e-06){

  # Get dimensions
  dimensions <- dim(data)

  # Check for whether rows equal columns
  if(dimensions[1] != dimensions[2]){ # Not a (square) matrix
    return(FALSE)
  }else{

    # Convert to matrix
    data_matrix <- as.matrix(data)

    # Check that all are equal
    return(all(abs(data_matrix - t(data_matrix)) < tolerance))

  }

}

#' @noRd
# Ensure data has dimension names ----
# Updated 24.07.2023
ensure_dimension_names <- function(data)
{

  # Determine object type
  object_type <- get_object_type(data)

  # Branch for vector vs. matrix and data frame
  if(object_type == "vector"){

    # Get length of vector
    vector_length <- length(data)

    # Check for names
    if(is.null(names(data))){
      names(data) <- paste0(
        "V", format_integer(
          seq_len(vector_length),
          digits(vector_length) - 1
        )
      )
    }

  }else if(object_type %in% c("matrix", "data.frame")){

    # Get dimensions
    dimensions <- dim(data)

    # Get dimension names
    dimension_names <- dimnames(data)

    # Check for column names
    if(is.null(dimension_names[[2]])){

      # Standardize names
      dimnames(data)[[2]] <- paste0(
        "V", format_integer(
          seq_len(dimensions[2]),
          digits(dimensions[2]) - 1
        )
      )

    }

    # Check for symmetric matrix
    if(is_symmetric(data)){

      # Check for row names
      if(is.null(dimnames(data)[[1]]) | any(dimension_names[[1]] != dimension_names[[2]])){

        # Assign column names to row names
        dimnames(data)[[1]] <- dimnames(data)[[2]]

      }

    }

  }

  # Return named data
  return(data)

}

#' @noRd
# Transfer names from data to output ----
# Usually for matrix and data frame
# Updated 25.06.2023
transfer_names <- function(data, output)
{

  # Obtain column names
  column_names <- dimnames(data)[[2]]

  # Check for variable names
  if(!is.null(column_names)){

    # Add names to rows and columns
    dimnames(output) <- list(column_names, column_names)

  }

  # Return named output
  return(output)

}

#' @noRd
# Function to check for usable variables ----
# Updated 24.07.2023
usable_data <- function(data, verbose)
{

  # Turn data into a data matrix
  data_matrix <- data.matrix(data)

  # All missing after coercions
  remove_columns <- lvapply(as.data.frame(is.na(data_matrix)), all)

  # Send warning if there are any removals
  if(any(remove_columns)){

    # Send warning
    warning(
      paste0(
        "Some variables could not to be coerced to numeric values. These variables have been removed from the analysis:\n",
        paste0("'", dimnames(data_matrix)[[2]][remove_columns], "'", collapse = ", "),
        "\n\nIf these variables were not intended to be removed, then try converting them to numeric values before inputting the data into the function"
      ), call. = FALSE
    )

    # Remove these variables from `data` and `data_matrix`
    keep_columns <- !remove_columns
    data <- data[,keep_columns]
    data_matrix <- data_matrix[,keep_columns]

  }

  # Determine whether each variable was coerced
  coercions <- lvapply(as.data.frame(data_matrix != data), any, na.rm = TRUE)

  # Send warning for any coercions
  if(any(coercions) & verbose){

    # Send warning
    warning(
      paste0(
        "Several variables were coerced to numeric values. These variables were changed to numeric values:\n",
        paste0("'", dimnames(data_matrix)[[2]][coercions], "'", collapse = ", ")
      ), call. = FALSE
    )

  }

  # Send back usable data
  return(data_matrix)

}

#' @noRd
# Whether usable data is needed ----
# Updated 07.09.2023
needs_usable <- function(ellipse)
{
  return(!"needs_usable" %in% names(ellipse) || ellipse$needs_usable)
}

#' @noRd
# Obtain data, sample size, correlation matrix ----
# Generic function to get the usual needed inputs
# Updated 29.11.2025
obtain_sample_correlations <- function(data, n, corr, na_data, verbose, ...)
{

  # Check if data is a correlation matrix
  if(is_symmetric(data)){

    # Check for sample size
    if(is.null(n)){
      stop(
        "A symmetric matrix was provided in the 'data' argument but the sample size argument, 'n', was not set. Please input the sample size into the 'n' argument.",
        call. = FALSE
      )
    }

    # If symmetric and sample size is provided, then
    # data to correlation matrix
    correlation_matrix <- data

  }else{

    # Check for usable data
    if(needs_usable(list(...))){
      data <- usable_data(data, verbose)
    }

    # Obtain sample size
    n <- dim(data)[1]

    # Check for automatic correlations
    if(corr == "auto"){
      correlation_matrix <- auto_correlate(
        data = data, corr = "pearson", na_data = na_data,
        verbose = verbose, ...
      )
    }else{
      correlation_matrix <- cor(
        data, use = swiftelse(
          na_data == "pairwise",
          "pairwise.complete.obs",
          "complete.obs"
        ), method = corr
      )
    }

  }

  # Return results
  return(
    list(
      data = data, n = n,
      correlation_matrix = correlation_matrix
    )
  )

}

# Compute thresholds ----
#' @noRd
# Updated 22.07.2023
obtain_thresholds <- function(categorical_variable)
{

  # Obtain cumulative sums from frequency table
  cumulative_sum <- cumsum(fast_table(categorical_variable))

  # Obtain cumulative length
  cumsum_length <- length(cumulative_sum)

  # Obtain thresholds
  return(
    qnorm(
      cumulative_sum[-cumsum_length] /
        cumulative_sum[cumsum_length]
    )
  )

}

#%%%%%%%%%%%%%%%%%%%%%%%%%
# QUIET CALLS & LOADS ----
#%%%%%%%%%%%%%%%%%%%%%%%%%

# Lots of R packages and functions print
# messages that add to the noise of {L0ggm}
# outputs; these functions make other
# packages' messages go away
#
# NOTE: In some cases, some package messages
# are important and should not be ignored
# (don't use these functions in those cases)

#' @noRd
# General function to silently obtain output ----
# Updated 01.07.2023
silent_call <- function(...)
{

  # Make call
  sink <- capture.output(
    result <- suppressWarnings(
      suppressMessages(
        ...
      )
    )
  )

  # Return result
  return(result)

}

#' @noRd
# General function to silently load package ----
# Updated 01.07.2023
silent_load <- function(...)
{
  return(suppressPackageStartupMessages(...))
}

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# FUNCTIONS & ARGUMENTS FUNCTIONS ----
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#' @noRd
# Strip classes and attributes
# Updated 09.07.2023
remove_attributes <- function(object)
{

  # Remove all except special attributes (see `?structure`)
  attributes(object) <- attributes(object)[
    names(attributes(object)) %in% c(
      "dim", "dimnames", "names", "tsp", "levels"
    )
  ]

  # Return unclassed object
  return(unclass(object))

}

#' @noRd
# Set default argument (cleaner missing function) ----
# Updated 28.07.2023
set_default <- function(argument, default, FUN, several.ok = FALSE)
{

  # Check for type error
  typeof_error(argument, c(typeof(default), "closure"), "set_default")

  # Return argument if it's a function
  if(is.function(argument)){
    return(argument)
  }

  # Return default if NULL
  if(is.null(argument)){
    return(default)
  }

  # Return default if length > 1 and several are NOT okay
  if(length(argument) > 1 & !several.ok){
    return(default)
  }

  # Get argument name
  argument_name <- as.character(substitute(argument))

  # Get choices for the function
  if(!is.function(FUN)){ # choices are provided in 'FUN'
    choices <- FUN
  }else{ # get formal arguments from the function itself

    # Get formal argument choices from default call
    choices <- formals(FUN)[[argument_name]]

    # Check if choices is a call
    if(is.call(choices)){

      # Get choices
      choices <- as.character(choices)

      # Remove "c" from call
      choices <- choices[-which(choices == "c")]

    }

  }

  # Make argument lowercase (if necessary)
  argument <- tolower(argument); choices <- tolower(choices)

  # Get arguments not in choices
  if(any(!argument %in% choices)){ # Repeated calls OK since returning error

    # Get arguments not in choices
    not_in_choices <- !argument %in% choices

    # Get number not in choices
    number_not <- sum(not_in_choices)

    # Send error for bad argument
    if(number_not > 0){

      # Only one not in choices
      if(number_not == 1){
        argument_text <- paste0(" = \"", argument[not_in_choices], "\"")
      }else{ # Multiple not in choices
        argument_text <- paste0(
          " = c(",
          paste0("\"", argument[not_in_choices], "\"", collapse = ", "),
          ")"
        )
      }

      # Throw error
      stop(
        paste0(
          paste0("Invalid argument: ", argument_name, argument_text),
          "\n\nPlease choose from: ",
          paste0("\"", choices, "\"", collapse = ", ")
        ), call. = FALSE
      )

    }

  }

  # Return argument
  return(argument)

}

#' @noRd
# Overwrite Arguments ----
# Updated 24.06.2023
overwrite_arguments <- function(main, ARGS)
{

  # Obtain replacement arguments
  target_args <- ARGS[names(ARGS) %in% names(main)]

  # Replace the arguments with the same name
  main[names(target_args)] <- target_args

  # Return main arguments
  return(main)

}

#' @noRd
# Function to obtain arguments ----
# Updated 06.07.2023
obtain_arguments <- function(FUN, FUN.args)
{

  # Overwrite arguments
  FUN.formals <- overwrite_arguments(formals(FUN), FUN.args)

  # Remove ellipses
  FUN.formals <- FUN.formals[names(FUN.formals) != "..."]

  # Remove call arguments (assume they are supplied elsewhere)
  return(FUN.formals[!lvapply(FUN.formals, is, "call")])

}

#%%%%%%%%%%%%%%%%%%%%%
# ERROR FUNCTIONS ----
#%%%%%%%%%%%%%%%%%%%%%

#' @noRd
# Experimental warning ----
# Updated 03.08.2023
experimental <- function(function_name)
{
  warning(
    paste0(
      "This implementation of `", function_name,
      "` is ", styletext("experimental", "italics"), ". \n\nThe underlying function ",
      "and/or output may change until the results have been appropriately vetted and validated."
    ), call. = FALSE
  )
}

#' @noRd
# Error for class ----
# Updated 04.09.2023
class_error <- function(input, expected_class, function_name){

  # Check for object types
  if(!is(input, expected_class)){
    .handleSimpleError(
      h = stop,
      msg = paste0(
        "Input into '", deparse(substitute(input)),
        "' is an object with ", paste0("'", class(input), "'", collapse = ", "), " class(es)",
        ". Input is expected to be ", paste0("'", expected_class, "'", collapse = ", "), " class(es)",
        "\n\n For more details on how to fix this error, see:\n",
        "https://r-ega.net/articles/errors.html#class-error"
      ),
      call = function_name
    )
  }

}

#' @noRd
# Error for object type ----
# Updated 04.09.2023
object_error <- function(input, expected_type, function_name){

  # Get input type
  input_type <- get_object_type(input)

  # Check for object types
  if(!input_type %in% expected_type){
    .handleSimpleError(
      h = stop,
      msg = paste0(
        "Input into '", deparse(substitute(input)),
        "' is a ", paste0("'", input_type, "'", collapse = ", "), " object",
        ". Input is expected to be ", paste0("'", expected_type, "'", collapse = ", "), " object",
        "\n\n For more details on how to fix this error, see:\n",
        "https://r-ega.net/articles/errors.html#object-error"
      ),
      call = function_name
    )
  }

}

#' @noRd
# Error for `typeof` ----
# Updated 04.09.2023
typeof_error <- function(input, expected_value, function_name){

  # Switch out "closure" with "function"
  if("closure" %in% expected_value){
    expected_value[expected_value == "closure"] <- "function"
  }

  # Get type of input
  typeof_input <- typeof(input)

  # Convert "closure" to "function"
  if("closure" %in% typeof_input){
    typeof_input[typeof_input == "closure"] <- "function"
  }

  # Convert "integer" and "double" to "numeric"
  ## Input
  typeof_input <- swiftelse(
    typeof_input %in% c("integer", "double"),
    "numeric", typeof_input
  )
  ## Expected value
  expected_value <- swiftelse(
    any(expected_value %in% c("integer", "double")),
    "numeric", expected_value
  )

  # Check for value
  if(!typeof_input %in% expected_value){
    .handleSimpleError(
      h = stop,
      msg = paste0(
        "Input into '", deparse(substitute(input)),
        "' is ", paste("'", typeof_input, "'", sep = "", collapse = ", "), " type",
        ". Input is expected to be ",
        paste0("'", expected_value, "'", collapse = " or "), " type",
        "\n\n For more details on how to fix this error, see:\n",
        "https://r-ega.net/articles/errors.html#typeof-error"
        # can use "or" because `typeof` won't ever be more than two
      ),
      call = function_name
    )
  }

}

#' @noRd
# Error for `length` ----
# Updated 04.09.2023
length_error <- function(input, expected_lengths, function_name){

  # Check for length of input in expected length
  if(!length(input) %in% expected_lengths){
    .handleSimpleError(
      h = stop,
      msg = paste0(
        "Length of '", deparse(substitute(input)),
        "' (", length(input),") does not match expected length(s). Length must be: ",
        paste0("'", expected_lengths, "'", collapse = " or "),
        "\n\n For more details on how to fix this error, see:\n",
        "https://r-ega.net/articles/errors.html#length-error"
      ),
      call = function_name
    )
  }

}

#' @noRd
# Error for `range` ----
# Updated 04.09.2023
range_error <- function(input, expected_ranges, function_name){

  # Obtain expected maximum and minimum values
  expected_maximum <- max(expected_ranges, na.rm = TRUE)
  expected_minimum <- min(expected_ranges, na.rm = TRUE)

  # Obtain maximum and minimum values
  actual_maximum <- round(max(input, na.rm = TRUE), 3)
  actual_minimum <- round(min(input, na.rm = TRUE), 3)

  # Check for maximum of input in expected range
  if(actual_maximum > expected_maximum){
    .handleSimpleError(
      h = stop,
      msg = paste0(
        "Maximum value of '", deparse(substitute(input)),
        "' (", actual_maximum,") does not match expected range(s). Values must range between: ",
        paste0("'", expected_ranges, "'", collapse = " and "),
        "\n\n For more details on how to fix this error, see:\n",
        "https://r-ega.net/articles/errors.html#range-error"
      ),
      call = function_name
    )
  }

  # Check for maximum of input in expected range
  if(actual_minimum < expected_minimum){
    .handleSimpleError(
      h = stop,
      msg = paste0(
        "Minimum value of '", deparse(substitute(input)),
        "' (", actual_minimum,") does not match expected range(s). Values must range between: ",
        paste0("'", expected_ranges, "'", collapse = " and "),
        "\n\n For more details on how to fix this error, see:\n",
        "https://r-ega.net/articles/errors.html#range-error"
      ),
      call = function_name
    )
  }

}

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# COVARIANCE CONVERSION FUNCTIONS ----
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#' @noRd
# Zero-order correlations to partial correlations ----
# Updated 29.06.2023
cor2pcor <- function(correlation_matrix)
{

  # Convert to inverse correlations to partial correlations
  partial_correlations <- -cov2cor(solve(correlation_matrix))

  # Set diagonal to zero
  diag(partial_correlations) <- 0

  # Return
  return(partial_correlations)

}

#' @noRd
# Inverse covariances to partial correlations ----
# Updated 07.03.2026
inv2pcor <- function(precision)
{

  # Convert to inverse correlations to partial correlations
  partial_correlations <- -cov2cor(precision)

  # Set diagonal to zero
  diag(partial_correlations) <- 0

  # Return
  return(partial_correlations)

}

#' @noRd
# Partial correlations to zero-order correlations ----
# Updated 29.06.2023
pcor2cor <- function(partial_correlations)
{

  # Set diagonal to negative 1
  diag(partial_correlations) <- -1

  # Return partial correlations as zero-order correlations
  return(cov2cor(solve(-partial_correlations)))

}

#' @noRd
# Partial correlations to inverse covariances ----
# Updated 25.09.2024
pcor2inv <- function(partial_correlations)
{

  # Set diagonal to negative 1
  diag(partial_correlations) <- -1

  # Return inverse covariance matrix
  return(solve(-partial_correlations))

}

#%%%%%%%%%%%%%%%%%%%%%%%%
## NETWORK FUNCTIONS ----
#%%%%%%%%%%%%%%%%%%%%%%%%

#' @noRd
# Create sparse network ----
# Updated 08.11.2023
sparse_network <- function(network)
{

  # Get number of nodes
  nodes <- dim(network)[2]

  # Get node sequence
  node_sequence <- seq_len(nodes)

  # Create data frame
  sparse_df <- data.frame(
    row = rep(node_sequence, each = nodes),
    col = rep(node_sequence, times = nodes),
    weight = as.vector(network)
  )

  # Return lower triangle
  return(sparse_df[sparse_df$row < sparse_df$col,])

}

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# MATH & STATS FUNCTIONS ----
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%

# There are some redundancies in the math functions because
# it's often faster to call some functions directly
# rather than nesting non-base R functions in functions

#' @noRd
# Positive definite matrix ----
# Logical for whether a matrix is positive definite
# Updated 05.11.2024
is_positive_definite <- function(data)
{
  return(all(eigen(data, symmetric = TRUE, only.values = TRUE)$values > .Machine$double.eps))
}

#' @noRd
# Fast kappa exact ----
# Updated 11.03.2026
fast_kappa <- function(X)
{
  # Obtain eigenvalues
  E <- abs(eigen(X, symmetric = TRUE, only.values = TRUE)$values)

  # Return kappa
  return(max(E) / min(E))

}

#' @noRd
# Wrapper for `eigen` ----
# Extracts eigenvalues only
# Updated 11.03.2026
matrix_eigenvalues <- function(X)
{
  return(eigen(X, symmetric = TRUE, only.values = TRUE)$values)
}

#' @noRd
# Trace of matrix ----
# Updated 25.06.2023
trace <- function(object)
{
  return(sum(diag(object), na.rm = TRUE))
}

#' @noRd
# r to Fisher's z ----
# Updated 27.03.2025
r2z <- function(r)
{
  return(0.5 * log((1 + r) / (1 - r)))
}

#' @noRd
# Fisher's z to r ----
# Updated 27.03.2025
z2r <- function(z)
{
  return((exp(2 * z) - 1) / (1 + exp(2 * z)))
}

#' @noRd
# Partial correlation p-values ----
# Updated 27.03.2025
p_partial <- function(z, variables, sample_size)
{

  # Get test statistic
  test <- z / sqrt(1 / (sample_size - variables - 5))

  return(pnorm(test, lower.tail = FALSE) * 2)

}

#' @noRd
# Min-max normalization ----
# Updated 11.03.2026
min_max <- function(x)
{

  # Obtain minimum and maximum
  x_range <- range(x, na.rm = TRUE)

  # Return min-max normalization
  return(
    swiftelse(
      x_range[1] == x_range[2], x,
      (x - x_range[1]) / (x_range[2] - x_range[1])
    )
  )

}

#%%%%%%%%%%%%%%%%%%%%%%%
# SYSTEM FUNCTIONS ----
#%%%%%%%%%%%%%%%%%%%%%%

#' @noRd
# OS and System Check ----
# Updated 28.11.2023
system.check <- function()
{

  # Get OS usage
  OS <- unname(tolower(Sys.info()["sysname"]))

  # Get RStudio usage
  RSTUDIO <- swiftelse(Sys.getenv("RSTUDIO") == "1", TRUE, FALSE)

  # Return list
  return(
    list(
      OS = OS,
      R = paste0(R.version$major, ".", R.version$minor),
      RSTUDIO = RSTUDIO,
      TEXT = swiftelse(!RSTUDIO & OS != "linux", FALSE, TRUE)
    )
  )

}

#' @noRd
# Colorize text ----
# Updated 08.09.2020
colortext <- function(text, number = NULL, defaults = NULL)
{
  # Check system
  sys.check <- system.check()

  if(sys.check$TEXT)
  {
    # Defaults for number (white text)
    if(is.null(number) || number < 0 || number > 231)
    {number <- 15}

    # Check for default color
    if(!is.null(defaults))
    {
      # Adjust highlight color based on background color
      if(defaults == "highlight")
      {
        number <- 208
      }else{

        number <- switch(defaults,
                         message = 204,
                         red = 9,
                         orange = 208,
                         yellow = 11,
                         "light green" = 10,
                         green = 34,
                         cyan = 14,
                         blue = 12,
                         magenta = 13,
                         pink = 211,
        )

      }

    }

    return(paste("\033[38;5;", number, "m", text, "\033[0m", sep = ""))

  }else{return(text)}
}

#' @noRd
# Style text ----
# Updated 26.07.2023
styletext <- function(
    text, defaults = c(
      "bold", "italics", "highlight",
      "underline", "strikethrough"
    )
)
{

  if(system.check()$TEXT)
  {

    if(missing(defaults)){
      number <- 0
    }else{

      number <- switch(
        defaults,
        bold = 1, italics = 3,
        underline = 4, highlight = 7,
        strikethrough = 9
      )

    }

    return(paste("\033[", number, ";m", text, "\033[0m", sep = ""))

  }else{
    return(text)
  }

}

#' @noRd
# Symbols ----
# Updated 26.07.2024
textsymbol <- function(
    symbol = c(
      "alpha", "beta", "chi", "delta",
      "eta", "gamma", "lambda", "omega",
      "phi", "pi", "rho", "sigma", "tau",
      "theta", "square root", "infinity",
      "check mark", "x", "bullet"
    )
)
{
  # Return code
  return(
    switch(
      symbol,
      alpha = "\u03B1", beta = "\u03B2", chi = "\u03C7",
      delta = "\u03B4", eta = "\u03B7", gamma = "\u03B3",
      lambda = "\u03BB", omega = "\u03C9", phi = "\u03C6",
      pi = "\u03C0", rho = "\u03C1", sigma = "\u03C3", tau = "\u03C4",
      theta = "\u03B8", "square root" = "\u221A", infinity = "\u221E",
      "check mark" = "\u2713", x = "\u2717", bullet = "\u2022"
    )
  )

}

#' @noRd
# Title Case ----
# Not truly title case -- just capitializes
# the first letter of each word
# Uses `tools::toTitleCase` for actual title case
# Updated 25.06.2023
totitle <- function(string)
{

  # Split by spaces
  words <- unlist(strsplit(string, split = " "))

  # Set first letters to uppercase
  titleCased <- cvapply(words, function(x){

    # Stitch together letters
    return(
      paste0(
        toupper(substr(x, 1, 1)),
        tolower(substr(x, 2, nchar(x)))
      )
    )

  })

  # Paste words back together
  return(paste(titleCased, collapse = " "))

}

#' @noRd
# General function to check for packages ----
# Updated 13.07.2023
check_package <- function(packages)
{

  # Performs what original `installed.packages()` does
  # but without additional fluff
  installed <- packages %in% ulapply(.libPaths(), list.files)

  # Determine which packages are not installed
  not_installed <- packages[!installed]

  # Print error with missing packages
  if(length(not_installed) != 0){

    # Organize packages error output
    if(length(not_installed) > 1){

      # Get missing packages
      missing_packages <- paste0("{", packages , "}", collapse = ", ")
      packages <- paste0("\"", packages, "\"", collapse = ", ")

      # Stop and tell user to install package
      stop(
        paste0(
          missing_packages,
          " are not installed but are required for this function. ",
          "Please run \n\n",
          "`install.packages(c(", packages, "))`",
          "\n\nOnce installed, re-run this function (you may need to restart R/RStudio)."
        ), call. = FALSE
      )

    }else{

      # Get missing packages
      missing_packages <- paste0("{", packages, "}")
      packages <- paste0("\"", packages, "\"")

      # Stop and tell user to install package
      stop(
        paste0(
          missing_packages,
          " is not installed but is required for this function. ",
          "Please run \n\n",
          "`install.packages(", packages, ")`",
          "\n\nOnce installed, re-run this function (you may need to restart R/RStudio)."
        ), call. = FALSE
      )

    }

  }

}

#' Predict Weibull Parameters for Edge Weight Distributions
#'
#' @description
#' Predicts the shape and scale parameters of a Weibull distribution that
#' characterizes the absolute partial correlation edge weights of a psychometric
#' network, given its number of nodes and sample size. Parameter estimates are
#' derived from a Seemingly Unrelated Regression (SUR) model fitted to empirical
#' network data from Huth et al. (2025), where absolute partial correlations were
#' found to follow a Weibull distribution more consistently than Beta, Gamma, or
#' log-normal alternatives.
#'
#' @param nodes Integer. The number of nodes (variables) in the network.
#' Must be between 8 and 54, reflecting the range of the empirical networks
#' used to fit the underlying SUR model.
#'
#' @param sample_size Integer. The sample size of the dataset from which the
#' network is estimated.
#'
#' @param snr Numeric (length = 1).
#' Signal-to-noise ratio of partial correlations (\eqn{\bar{|w|} / \mathrm{SD}(|w|)}).
#' Values less than 1 indicate wider range of partial correlations (\eqn{w}) whereas
#' values greater than 1 indicate narrower range.
#' Defaults to \code{1} where the mean of the partial correlations (\eqn{\bar{|w|}})
#' is equal to the standard deviation (\eqn{\mathrm{SD}(|w|)})
#'
#' @param bootstrap Logical. If \code{TRUE}, a randomly sampled residual from
#' the SUR model fit is added to each predicted parameter, introducing
#' empirically grounded variability suitable for use in simulation or
#' bootstrapping contexts. Defaults to \code{FALSE}.
#'
#' @details
#' The shape and scale parameters are predicted from derived network descriptors
#' that differ between the two SUR equations:
#'
#' \describe{
#'   \item{\code{rlp}}{The reciprocal log of the number of nodes,
#'     \eqn{1 / \log(p)}, capturing the diminishing marginal effect of network
#'     size on edge weight distributions. Used in both the shape and scale
#'     equations.}
#'   \item{\code{scaling}}{The standard error of partial correlations defined as
#'     \eqn{\sqrt{1 / (n - p)}}, where \eqn{n} is the sample size and
#'     \eqn{p} is the number of nodes. Larger values indicate greater sampling
#'     uncertainty in the partial correlation estimates. Used in the scale
#'     equation only.}
#' }
#'
#' The two SUR equations have an asymmetric structure reflecting different
#' theoretical roles for sampling precision. Shape — which governs the
#' concentration of the edge weight distribution — is determined solely by
#' signal characteristics of the network via \code{snr} and \code{rlp}.
#' Scale — which governs the typical magnitude of edge weights — additionally
#' depends on \code{scaling}, as the expected size of partial correlations is
#' directly affected by estimation precision. This asymmetry is both
#' empirically supported (dropping \code{scaling} from the shape equation
#' costs \eqn{\Delta R^2 < 0.006}) and theoretically coherent.
#'
#' These predictors enter two SUR equations whose coefficients are stored in
#' the internal \code{weibull_weights} dataset. SUR was used to account for the
#' correlated residuals between the shape and scale equations across networks
#' (residual correlation = 0.261). Model fit was strong: the shape equation
#' achieved \eqn{R^2 = 0.887} (RMSE = 0.048) and the scale equation achieved
#' \eqn{R^2 = 0.885} (RMSE = 0.011). Shape residuals were normally distributed
#' (Shapiro-Wilk W = 0.990, p = 0.173). Scale residuals showed a modest
#' departure from normality (Shapiro-Wilk W = 0.973, p < 0.001), consistent
#' with slight right skew in the scale outcome and test sensitivity at n = 194
#' rather than a substantive violation. Heteroskedasticity was detected in both
#' the shape equation (Breusch-Pagan p < 0.001) and the scale equation
#' (Breusch-Pagan p < 0.001); robust standard errors (HC3) were used for
#' inference. Multicollinearity among predictors in the scale equation was
#' negligible (VIF \eqn{\leq} 1.60); the shape equation contains only two
#' predictors with no multicollinearity concern.
#'
#' \code{nodes} influences predicted shape and scale via \code{rlp}.
#' \code{sample_size} influences predicted scale via \code{scaling}, and
#' both parameters via their joint contribution to \code{snr} when it is
#' estimated from data rather than supplied directly.
#'
#' Empirically, shape values ranged from approximately 0.72 to 1.63
#' (M = 1.07, SD = 0.14) and scale values from approximately 0.03 to 0.19
#' (M = 0.10, SD = 0.03) across the 194 networks used to fit the model. Shape
#' values near 1 indicate approximately exponential edge weight distributions;
#' values above 1 indicate a rising hazard (mode-bearing distribution).
#'
#' When \code{bootstrap = TRUE}, residuals are drawn via \code{shuffle()} --
#' a random sampling without replacement -- from the empirical SUR residuals,
#' preserving the observed marginal residual distribution.
#'
#' @return A named numeric vector of length 2:
#' \describe{
#'   \item{\code{shape}}{The predicted Weibull shape parameter (\eqn{k > 0}).}
#'   \item{\code{scale}}{The predicted Weibull scale parameter (\eqn{\lambda > 0}).}
#' }
#'
#' @examples
#' # Predict parameters for a 10-node network with n = 500
#' weibull_parameters(nodes = 10, sample_size = 500)
#'
#' # With bootstrapped residuals for use in simulation
#' weibull_parameters(nodes = 10, sample_size = 500, bootstrap = TRUE)
#'
#' @author Alexander P. Christensen <alexpaulchristensen@gmail.com>
#'
#' @references
#' Huth, K. B. S., Haslbeck, J. M. B., Keetelaar, S., Van Holst, R. J., &
#' Marsman, M. (2025). Statistical evidence in psychological networks.
#' \emph{Nature Human Behaviour}.
#'
#' @export
#'
# Predict Weibull parameters ----
# Updated 29.04.2026
weibull_parameters <- function(nodes, sample_size, snr = 1, bootstrap = FALSE)
{

  # Get Weibull model
  weibull_weights <- get(data("weibull_weights", package = "L0ggm", envir = environment()))

  # Compute descriptive parameters
  shape_parameters <- c(snr = snr, rlp = 1 / log(nodes))
  scale_parameters <- c(shape_parameters, scaling = sqrt(1 / (sample_size - nodes)))

  # Compute Weibull parameters
  shape <- unname(
    weibull_weights$shape$coefficients[1] + sum(weibull_weights$shape$coefficients[-1] * shape_parameters)
  )
  scale <- unname(
    weibull_weights$scale$coefficients[1] + sum(weibull_weights$scale$coefficients[-1] * scale_parameters)
  )

  # Bootstrap residuals
  if(bootstrap){

    # Draw index
    index <- shuffle(seq_along(weibull_weights$shape$residuals), 1)

    # Set parameters with residuals
    shape <- shape + weibull_weights$shape$residuals[index]
    scale <- scale + weibull_weights$scale$residuals[index]

  }

  # Return parameters
  return(c(shape = shape, scale = scale))

}

#' Simulates Small-World GGM Data
#'
#' @description
#' Simulates data from a Gaussian Graphical Model (GGM) with a small-world
#' network structure. The generative process proceeds in three stages. First,
#' a ring lattice is constructed with \code{neighbors + 1} nearest-neighbor
#' connections per node, where \code{neighbors} is derived from \code{nodes}
#' and \code{density}. Second, the lattice is randomly pruned to the target
#' \code{density}, introducing degree heterogeneity from the outset. Third,
#' edges are rewired with probability \code{rewire}, where rewired edges are
#' placed preferentially on node pairs with higher combined degree (degree-
#' weighted rewiring). This approach produces more realistic degree
#' distributions than standard Watts-Strogatz rewiring while preserving
#' the local clustering structure of the lattice. Edge weights are assigned
#' by structural priority: edges retained from the pruned lattice receive
#' larger partial correlation weights based on their original neighbor
#' distance, grounded in the empirical observation that shorter-distance
#' connections tend to carry stronger weights in psychometric networks.
#' The resulting network is used to generate multivariate normal (or skewed)
#' data. Parameters do not have default values (except
#' \code{negative_proportion}, \code{snr}, \code{target_condition},
#' \code{max_correlation}, and \code{max_iterations}) and must each be set.
#' See Details and Examples to get started.
#'
#' @param nodes Numeric (length = 1).
#' Number of nodes in the network.
#' Minimum of three nodes. The total number of nodes should be between 8
#' and 54 to remain within the range of the empirical networks used to fit
#' the Weibull parameter model; values outside this range are accepted but
#' will trigger extrapolation and may produce a warning from
#' \code{\link{weibull_parameters}}.
#'
#' @param density Numeric (length = 1).
#' Target edge density of the network after pruning the initial ring lattice.
#' Controls the number of nearest neighbors \code{k} each node is connected
#' to via \eqn{k = \mathrm{round}((\text{nodes} \times
#' \frac{\text{nodes}-1}{2} \times \text{density}) / \text{nodes})}, subject
#' to a minimum of 2 and a maximum of
#' \eqn{\lfloor (\text{nodes}-1)/2 \rfloor}. A ring lattice with
#' \code{neighbors + 1} connections per node is first generated and then
#' pruned to this density, introducing degree heterogeneity before rewiring.
#' Must be between 0 and 1. A minimum density sufficient to maintain a
#' connected graph is enforced; values below this threshold will produce an
#' informative error.
#'
#' @param rewire Numeric (length = 1).
#' Probability of rewiring each edge. Unlike the standard Watts-Strogatz
#' model, rewired edges are placed using degree-weighted random selection:
#' node pairs with higher combined degree have a greater probability of
#' receiving a rewired edge (see Details). Values near 0 preserve the
#' pruned lattice structure; values near 1 produce approximately random
#' networks. The small-world regime typically occurs at intermediate values
#' (roughly 0.01 to 0.30). Must be between 0 and 1.
#'
#' @param snr Numeric (length = 1).
#' Signal-to-noise ratio of the absolute partial correlation weights,
#' defined as \eqn{\bar{|w|} / \mathrm{SD}(|w|)}. Values less than 1
#' produce wider, more heterogeneous weight distributions; values greater
#' than 1 produce narrower, more homogeneous distributions. Empirically
#' observed SNR values ranged from 0.648 to 1.712; values outside this
#' range are accepted but will trigger a warning. Defaults to \code{1}.
#'
#' @param negative_proportion Numeric (length = 1).
#' Proportion of edges that are negative (inhibitory).
#' Must be between 0 and 0.50. The upper bound of 0.50 is a mathematical
#' constraint of the sign-flipping procedure used to assign negative edges
#' (see Details). If not provided, a value is sampled from a truncated
#' normal distribution reflecting the empirical distribution of true
#' negative partial correlations across 194 psychometric networks:
#' mean = 0.34, SD = 0.086, bounded to \eqn{[0.083, 0.50]}.
#'
#' @param sample_size Numeric (length = 1).
#' Number of observations to generate from the population multivariate
#' normal distribution. Also influences the predicted Weibull scale
#' parameter via \code{\link{weibull_parameters}}: larger samples are
#' associated with smaller, more precisely estimated edge weights.
#'
#' @param skew Numeric (length = 1 or \code{nodes}).
#' Skew applied to each variable after generation from the multivariate
#' normal. Can be a single value (applied to all variables) or one value
#' per variable. Values are rounded to the nearest increment of 0.05 in
#' the range \eqn{[-2, 2]}. Defaults to \code{0} (no skew).
#'
#' @param skew_range Numeric (length = 2).
#' If provided, a skew value is drawn uniformly from this range for each
#' variable, overriding \code{skew}. Both values must be between -2 and 2.
#'
#' @param target_condition Numeric (length = 1).
#' Target condition number (using \code{\link{kappa}} with \code{exact =
#' TRUE}) applied when ridge regularization is needed to recover a positive
#' definite precision matrix. The smallest ridge penalty \eqn{\lambda} that
#' brings the condition number to this target is found via root-finding
#' (\code{\link{uniroot}}), subject to a maximum shrinkage of approximately
#' 23\% following Peeters et al. (2020). Lower values produce
#' better-conditioned matrices. Defaults to \code{30}. Values up to
#' \code{100} are accepted but not recommended.
#'
#' @param max_correlation Numeric (length = 1).
#' Maximum allowed absolute pairwise correlation in the population
#' correlation matrix \code{R}. Any draw where
#' \code{max(abs(R[lower.tri(R)])) > max_correlation} is rejected and a
#' new attempt is made. Must be between 0 and 1. Defaults to \code{0.80}.
#'
#' @param max_iterations Numeric (length = 1).
#' Maximum number of attempts to find (1) a connected network structure,
#' (2) a network satisfying the small-world screening criterion, and (3) a
#' valid set of edge weights. The error message reports a frequency table
#' of rejection reasons to assist with diagnosing convergence failures.
#' Defaults to \code{100}.
#'
#' @details
#' \strong{Lattice generation and pruning}
#'
#' The generative process begins by constructing a ring lattice with
#' \code{neighbors + 1} nearest-neighbor connections per node via
#' \code{\link[igraph]{sample_smallworld}} with \code{p = 0}. The
#' \code{+1} overshoot ensures the lattice always has more edges than the
#' target density, guaranteeing that pruning can proceed. The lattice is
#' then randomly pruned to the target \code{density} by removing edges
#' uniformly at random, subject to the constraint that the pruned graph
#' remains connected. Because removal is random, different nodes lose
#' different numbers of edges, producing a heterogeneous degree distribution
#' before any rewiring occurs. This heterogeneity provides a non-uniform
#' prior for the subsequent degree-weighted rewiring step.
#'
#' \strong{Degree-weighted rewiring}
#'
#' Each edge in the pruned lattice is independently selected for rewiring
#' with probability \code{rewire}. For each selected edge \eqn{(i, j)},
#' node \eqn{i} is kept fixed and the \eqn{j} endpoint is redirected to a
#' new target node. Valid targets are restricted to node pairs involving
#' node \eqn{i} that (1) are not currently connected, and (2) have never
#' been occupied during the current rewiring pass (i.e., were absent in
#' the original pruned lattice). Among valid targets, the new endpoint is
#' selected with probability proportional to \eqn{\sqrt{d_i + d_k}}, where
#' \eqn{d_i} and \eqn{d_k} are the current degrees of the two nodes in
#' the candidate pair. The square-root transformation moderates the
#' rich-get-richer tendency of linear preferential attachment, producing
#' degree heterogeneity consistent with the empirical range of psychometric
#' networks without generating extreme hubs. Node degrees are updated
#' incrementally after each rewire so that subsequent rewiring steps
#' reflect the current graph state.
#'
#' \strong{Edge weight assignment}
#'
#' Edge weights are assigned by ranking edges according to their distance
#' in the pruned lattice before rewiring. Edges retained from the pruned
#' lattice receive their original neighbor distance (1 = nearest neighbor,
#' 2 = second nearest, etc.); rewired edges are assigned a distance of
#' \code{neighbors + 1}, placing them at the bottom of the priority order.
#' Absolute edge weights are drawn from a Weibull distribution whose
#' parameters are predicted from the network size, sample size, and signal-
#' to-noise ratio using a Seemingly Unrelated Regression (SUR) model fitted
#' to 194 empirical psychometric networks (Huth et al., 2025). The sorted
#' (descending) Weibull weights are then mapped onto the distance ranking
#' so that the largest weights go to the shortest-distance edges. This
#' assignment is grounded in the empirical observation that shorter-distance
#' local connections tend to carry larger partial correlation weights in
#' psychometric networks.
#'
#' \strong{Negative edge assignment}
#'
#' Negative edges are introduced by flipping the signs of a subset of
#' nodes — multiplying all edges incident to those nodes by -1. The number
#' of nodes to flip is derived from \code{negative_proportion} via the
#' inversion formula \eqn{(1 - \sqrt{1 - 2p}) / 2}, where \eqn{p} is the
#' target proportion of negative edges. This formula assumes that an edge
#' is negative if and only if exactly one of its endpoints is flipped,
#' giving an expected negative proportion of \eqn{2 \cdot (k/n) \cdot
#' (1 - k/n)} for \eqn{k} flipped nodes out of \eqn{n} total. The formula
#' requires \eqn{p \leq 0.50}, which is why \code{negative_proportion} is
#' bounded at 0.50.
#'
#' @return A named list with four elements:
#'
#' \item{data}{Numeric matrix of dimension \code{sample_size x nodes}
#' containing the simulated observations drawn from the population GGM.
#' Rows are cases; columns are variables named \code{V01}, \code{V02},
#' etc. Values are continuous (or skewed continuous when \code{skew != 0}).
#' To produce ordinal data, pass the columns through
#' \code{\link{categorize}}.}
#'
#' \item{parameters}{
#' A list of input, derived, and estimated parameters:
#' \itemize{
#'   \item \code{nodes} --- Number of nodes (as supplied)
#'   \item \code{density} --- Target edge density (as supplied)
#'   \item \code{neighbors} --- Number of nearest neighbors \code{k} derived
#'   from \code{nodes} and \code{density}; the initial ring lattice uses
#'   \code{neighbors + 1} connections before pruning
#'   \item \code{rewire} --- Edge rewiring probability (as supplied)
#'   \item \code{negative_proportion} --- Proportion of negative edges,
#'   either as supplied or as sampled from the empirical distribution
#'   \item \code{sample_size} --- Number of simulated observations
#'   \item \code{skew} --- Named numeric vector of per-variable skew values
#'   actually applied (after rounding and possible resampling)
#'   \item \code{weibull} --- Named numeric vector of length 2 giving the
#'   Weibull \code{shape} and \code{scale} parameters of the absolute edge
#'   weight distribution actually used. If ridge conditioning was applied,
#'   these are re-estimated from the conditioned network via MLE.
#'   \item \code{omega} --- Smallworldness omega statistic of the generated
#'   network (Telesford et al., 2011). Values near zero indicate small-world
#'   structure; negative values indicate lattice-like structure; positive
#'   values indicate random-like structure (see \code{\link[L0ggm]{smallworldness}})
#'   \item \code{Q} --- Newman-Girvan modularity of the population network
#'   (\code{Omega}) computed via \code{igraph::modularity} on absolute edge
#'   weights, using the community partition returned by
#'   \code{igraph::cluster_leiden} with \code{objective_function =
#'   "modularity"}. Because \code{cluster_leiden} is a heuristic algorithm,
#'   \code{Q} reflects a high-quality but not necessarily optimal partition.
#'   Values near zero indicate absence of community structure, consistent
#'   with the network theory of psychopathology.
#' }
#' }
#'
#' \item{population}{
#' Population-level network parameters:
#' \itemize{
#'   \item \code{R} --- Population correlation matrix derived from the GGM
#'   via \code{pcor2cor}
#'   \item \code{Omega} --- Population partial correlation matrix (the GGM
#'   edge weight matrix), with zeros for absent edges
#' }
#' }
#'
#' \item{convergence}{
#' Iteration and conditioning diagnostics:
#' \itemize{
#'   \item \code{iterations} --- Number of sampling attempts needed to find
#'   a valid network
#'   \item \code{rejections} --- Character vector recording the rejection
#'   reason for each failed attempt. Common reasons include disconnected
#'   graph structure, omega exceeding the small-world threshold, condition
#'   number exceeding \code{target_condition}, and maximum correlation
#'   exceeding \code{max_correlation}
#'   \item \code{lambda} --- Ridge regularization parameter \eqn{\lambda}
#'   added to the diagonal of the precision matrix to ensure positive
#'   definiteness; \code{NA} if no conditioning was required
#'   \item \code{condition} --- Condition number of the final population
#'   correlation matrix \code{R}
#' }
#' }
#'
#' @examples
#' # Basic small-world network (moderate density, moderate rewiring)
#' result <- simulate_smallworld(
#'   nodes = 20,
#'   density = 0.30,
#'   rewire = 0.20,
#'   sample_size = 500
#' )
#'
#' # Lattice-like structure (low rewiring preserves local connectivity)
#' result <- simulate_smallworld(
#'   nodes = 20,
#'   density = 0.30,
#'   rewire = 0.01,
#'   sample_size = 500
#' )
#'
#' # Random-like structure (high rewiring destroys lattice regularity)
#' result <- simulate_smallworld(
#'   nodes = 20,
#'   density = 0.30,
#'   rewire = 0.80,
#'   sample_size = 500
#' )
#'
#' # Fix the proportion of negative edges
#' result <- simulate_smallworld(
#'   nodes = 20,
#'   density = 0.30,
#'   rewire = 0.20,
#'   sample_size = 500,
#'   negative_proportion = 0.20
#' )
#'
#' @author Alexander P. Christensen <alexpaulchristensen@gmail.com>
#'
#' @references
#' \strong{Seminal introduction to the Watts-Strogatz small-world model} \cr
#' Watts, D. J., & Strogatz, S. H. (1998).
#' Collective dynamics of 'small-world' networks.
#' \emph{Nature}, \emph{393}(6684), 440--442.
#'
#' \strong{Logic for weight assignments} \cr
#' Muldoon, S. F., Bridgeford, E. W., & Bassett, D. S. (2016).
#' Small-world propensity and weighted brain networks.
#' \emph{Scientific Reports}, \emph{6}(1), 22057.
#'
#' \strong{Analytical approximation of random-graph average path length} \cr
#' Newman, M. E. J., Strogatz, S. H., & Watts, D. J. (2001).
#' Random graphs with arbitrary degree distributions and their applications.
#' \emph{Physical Review E}, \emph{64}(2), 026118.
#'
#' \strong{Empirical network data used to fit the Weibull SUR model} \cr
#' Huth, K. B. S., Haslbeck, J. M. B., Keetelaar, S., Van Holst, R. J., &
#' Marsman, M. (2025). Statistical evidence in psychological networks.
#' \emph{Nature Human Behaviour}.
#'
#' \strong{Maximum ridge shrinkage bound} \cr
#' Peeters, C. F., van de Wiel, M. A., & van Wieringen, W. N. (2020).
#' The spectral condition number plot for regularization parameter evaluation.
#' \emph{Computational Statistics}, \emph{35}(2), 629--646.
#'
#' @export
#'
# Simulate Small-world GGM data ----
# Updated 15.05.2026
simulate_smallworld <- function(
    nodes, density, rewire, snr = 1, negative_proportion,
    sample_size, skew = 0, skew_range = NULL,
    target_condition = 30, max_correlation = 0.80,
    max_iterations = 100
)
{

  # Check for missing negative proportion
  if(missing(negative_proportion)){

    # Compute proportion of negative edges based on empirical values
    negative_proportion <- pmin(
      pmax(
        0.34 + rnorm_ziggurat(1) * 0.086, # empirical mean +/- 1 SD
        0.083 # empirical minimum
      ), 0.50 # empirical maximum (0.54945055)
      # needs to be capped at 0.50 due to flipping variables
    )

  }

  # Check for input errors
  simulate_smallworld_errors(
    nodes, density, rewire, snr, negative_proportion, sample_size,
    skew, target_condition, max_correlation, max_iterations
  )

  # Set neighbors
  neighbors <- get_neighbors(nodes, density)

  # Create lattice network
  lattice <- lattice_generate(nodes, neighbors + 1)

  # Obtain lower triangle
  lower_triangle <- lower.tri(lattice)

  # Initialize iterations
  iter <- 0

  # Initialize condition
  search <- TRUE

  # Collect rejections
  rejections <- character(length = max_iterations + 1)

  # Generation loop
  while(search){

    # Increase iterations
    iter <- iter + 1

    # Stop on greater than max iterations
    if(iter > max_iterations){

      # Collect rejections
      rejection_table <- fast_table(rejections)

      # Remove NULL
      rejection_table <- rejection_table[names(rejection_table) != ""]

      # Return error
      stop(
        paste0(
          "Reached maximum iterations: Could not find weights that converge for the small-world structure.\n",
          "Resulting rejections were due to:\n\n",
          paste0(names(rejection_table), " = ", rejection_table, collapse = "\n")
        )
      )

    }

    # Prune lattice
    pruned_lattice <- prune2density(lattice, density, nodes, lower_triangle)

    # Small-world adjacency
    A <- smallworld_generate(pruned_lattice, rewire, lower_triangle)

    # Check adjacency
    if(!igraph::is_connected(convert2igraph(A))){

      # Add rejection reason
      rejections[iter] <- "Could not find structure where every node was connected."

      # Move to next iteration
      next

    }

    # Try to get good weights
    output <- try(
      smallworld_weights(
        A, pruned_lattice, nodes, sample_size,
        snr, negative_proportion, target_condition,
        max_correlation, lower_triangle
      ), silent = TRUE
    )

    # Check for error
    if(is(output, "try-error")){

      # Add rejection reason
      rejections[iter] <- trimws(strsplit(output[1], split = "\n")[[1]][2])

      # Move to next iteration
      next

    }

    # Update search
    search <- FALSE



  }

  # Generate data
  data_output <- simulate_data(
    n = sample_size, p = nodes, R = output$R, skew = skew, skew_range = skew_range
  )

  # Estimate modularity using optimal modularity detection
  igraph_network <- convert2igraph(abs(output$network))

  # Return parameters
  return(
    list(
      data = data_output$data,
      parameters = list(
        nodes = nodes,
        density = density,
        neighbors = neighbors,
        rewire = rewire,
        negative_proportion = negative_proportion,
        sample_size = sample_size,
        skew = data_output$skew,
        weibull = output$params,
        omega = smallworldness(
          network = output$network, lattice = pruned_lattice, method = "omega"
        ),
        Q = igraph::modularity(
          igraph_network, igraph::cluster_leiden(
            igraph_network, objective_function =  "modularity"
          )$membership
        )
      ),
      population = list(R = output$R, Omega = output$network),
      convergence = list(
        iterations = iter,
        rejections = rejections,
        lambda = output$lambda,
        condition = output$condition
      )
    )
  )

}

# Bug checking ----
# nodes = 20; density = 0.50; rewire = 0.10
# negative_proportion = 0.35; sample_size = 1000
# skew = 0; skew_range = NULL; snr = 1
# target_condition = 30;
# max_correlation = 0.80; max_iterations = 100

#' @noRd
# Errors ----
# Updated 16.03.2026
simulate_smallworld_errors <- function(
    nodes, density, rewire, snr, negative_proportion, sample_size,
    skew, target_condition, max_correlation, max_iterations
)
{

  # Errors for 'nodes'
  typeof_error(nodes, "numeric", "simulate_smallworld")
  length_error(nodes, 1, "simulate_smallworld")
  range_error(nodes, c(3, Inf), "simulate_smallworld")

  # Errors for 'density'
  typeof_error(density, "numeric", "simulate_smallworld")
  length_error(density, 1, "simulate_smallworld")
  range_error(density, c(0, 1), "simulate_smallworld")

  # Minimum edges required for a connected graph
  min_edges <- nodes - 1
  target_edges <- round(density * (nodes * (nodes - 1) / 2))

  # Check for whether connected graph is possible
  if(target_edges < min_edges){
    stop(
      paste0(
        "Target density (", round(density, 3), ") is too low to maintain ",
        "a connected graph for ", nodes, " nodes. ",
        "Minimum required density: ", round(min_edges / (nodes * (nodes - 1) / 2), 3)
      )
    )
  }

  # Errors for 'rewire'
  typeof_error(rewire, "numeric", "simulate_smallworld")
  length_error(rewire, 1, "simulate_smallworld")
  range_error(rewire, c(0, 1), "simulate_smallworld")

  # Error for 'snr'
  typeof_error(snr, "numeric", "simulate_smallworld")
  length_error(snr, 1, "simulate_smallworld")
  range_error(snr, c(0.01, Inf), "simulate_smallworld")

  # Send warning if beyond bounds
  if((snr < 0.64) | (snr > 1.72)){
    warning(paste0(
      "The signal-to-ratio specified (", round(snr, 2), ") is beyond ",
      "the bounds of cataloged empirical data (0.64-1.72). "
    ))
  }

  # Errors for 'negative_proportion'
  typeof_error(negative_proportion, "numeric", "simulate_smallworld")
  length_error(negative_proportion, 1, "simulate_smallworld")
  range_error(negative_proportion, c(0, 0.50), "simulate_smallworld")

  # Errors for 'sample_size'
  typeof_error(sample_size, "numeric", "simulate_smallworld")
  length_error(sample_size, 1, "simulate_smallworld")
  range_error(sample_size, c(1, Inf), "simulate_smallworld")

  # Errors for 'skew'
  typeof_error(skew, "numeric", "simulate_smallworld")
  length_error(skew, c(1, nodes), "simulate_smallworld")
  range_error(skew, c(-2, 2), "simulate_smallworld")

  # Errors for 'target_condition'
  typeof_error(target_condition, "numeric", "simulate_smallworld")
  length_error(target_condition, 1, "simulate_smallworld")
  range_error(target_condition, c(1, 100), "simulate_smallworld")

  # Errors for 'max_correlation'
  typeof_error(max_correlation, "numeric", "simulate_smallworld")
  length_error(max_correlation, 1, "simulate_smallworld")
  range_error(max_correlation, c(0, 1), "simulate_smallworld")

  # Errors for 'max_iterations'
  typeof_error(max_iterations, "numeric", "simulate_smallworld")
  length_error(max_iterations, 1, "simulate_smallworld")
  range_error(max_iterations, c(1, Inf), "simulate_smallworld")

}

#' @noRd
# Get number of neighbors based on density ----
# Updated 08.03.2026
get_neighbors <- function(nodes, density)
{

  # Degrees of freedom
  half_df <- (nodes - 1) / 2

  # Total edges with density
  edges <- nodes * half_df * density

  # Return with minimum neighbor check
  return(
    max(
      min(
        max(round(edges / nodes), 1), # maximum neighbors
        floor(half_df) # minimum degrees of freedom
      ), 2 # at least two neighbors
    )
  )

}

#' @noRd
# Generate lattice ----
# Updated 18.03.2026
lattice_generate <- function(nodes, neighbors)
{

  # Set node sequence
  node_sequence <- seq_len(nodes)

  # Set up distance matrix
  distance_matrix <- abs(outer(node_sequence, node_sequence, "-"))
  distance_matrix <- pmin(distance_matrix, nodes - distance_matrix)

  # Convert distance matrix to zeros and ones
  distance_matrix[] <- as.numeric(distance_matrix <= neighbors)

  # Ensure zero diagonal
  diag(distance_matrix) <- 0

  # Return lattice
  return(distance_matrix)

}

#' @noRd
# Prune neighbors to density ----
# Updated 13.04.2026
prune2density <- function(lattice, density, nodes, lower_triangle)
{

  # Store original lattice
  original_lattice <- lattice

  # Initialize lower triangle
  lattice_copy <- original_lattice[lower_triangle]

  # Compute target edges
  target_edges <- round(
    density * (nodes * (nodes - 1) / 2)
  )

  # Initilaize connectedness
  connected <- FALSE

  # Connecteness loop
  while(!connected){

    # Initialize lower lattice
    lattice_lower <- lattice_copy

    # Obtain current edges
    nonzero <- which(lattice_lower != 0)
    current_edges <- length(nonzero)

    # Obtain number of nodes to remove
    remove_edges <- current_edges - target_edges

    # Check for whether edges is zero or negative
    if(remove_edges < 1){
      return(lattice)
    }

    # Select edges and remove them
    lattice_lower[shuffle(nonzero, remove_edges)] <- 0

    # Reconstruct lattice
    lattice[] <- 0
    lattice[lower_triangle] <- lattice_lower
    lattice <- lattice + t(lattice)

    # Check connectedness
    connected <- igraph::is_connected(convert2igraph(lattice))

  }

  # Return lattice
  return(lattice)

}

#' @noRd
# Generate smallworld ----
# Updated 16.03.2026
smallworld_generate <- function(lattice, rewire, lower_triangle)
{

  # Obtain degree
  degree <- colSums(lattice, na.rm = TRUE)

  # Create sparse network data frame
  sparse_lattice <- sparse_network(lattice)

  # Add degrees
  sparse_lattice$degree <- outer(degree, degree, "+")[lower_triangle]

  # Determine rewired edges
  current_edges <- sparse_lattice$weight != 0
  edge_indices <- which(current_edges)
  valid_targets <- which(!current_edges)
  rewire_edges <- edge_indices[runif_xoshiro(length(edge_indices)) < rewire]
  n_rewire <- length(rewire_edges)

  # Check for no rewiring
  if(n_rewire > 0){

    # Loop over edges to rewire
    for(edge in rewire_edges){

      # Check for no more updates
      if(length(valid_targets) == 0){
        break
      }

      # Get edge indices
      node_i <- sparse_lattice$row[edge]
      node_j <- sparse_lattice$col[edge]

      # Get candidate edges
      candidate_edges <- intersect(
        which((sparse_lattice$row == node_i) | (sparse_lattice$col == node_i)),
        valid_targets
      )

      # Check for candidate edges
      if(length(candidate_edges) == 0){
        next
      }

      # Get weights based on degree
      zero_degree <- sqrt(sparse_lattice$degree[candidate_edges])

      # Get rewire location
      location <- weighted_shuffle(x = candidate_edges, size = 1, prob = zero_degree / sum(zero_degree))

      # Update edges
      sparse_lattice$weight[edge] <- 0
      sparse_lattice$weight[location] <- 1

      # Get rewired index
      rewired_j <- swiftelse(
        sparse_lattice$row[location] == node_i,
        sparse_lattice$col[location],
        sparse_lattice$row[location]
      )

      # Obtain indices
      node_j_index <- (sparse_lattice$row == node_j) | (sparse_lattice$col == node_j)
      rewired_j_index <- (sparse_lattice$row == rewired_j) | (sparse_lattice$col == rewired_j)

      # Update degrees
      sparse_lattice$degree[node_j_index] <- sparse_lattice$degree[node_j_index] - 1
      sparse_lattice$degree[rewired_j_index] <- sparse_lattice$degree[rewired_j_index] + 1

      # Update valid_targets
      valid_targets <- setdiff(valid_targets, location)

    }

  }

  # Obtain nonzeros
  nonzero_network <- sparse_lattice[sparse_lattice$weight == 1,]

  # Rebuild network
  lattice[] <- 0
  lattice[cbind(nonzero_network$row, nonzero_network$col)] <-
  lattice[cbind(nonzero_network$col, nonzero_network$row)] <- 1


  # Return network
  return(lattice)

}

#' @noRd
# Smallworld weight generation ----
# Updated 13.04.2026
smallworld_weights <- function(
    A, lattice, nodes, sample_size,
    snr, negative_proportion, target_condition,
    max_correlation, lower_triangle
)
{

  # Set node sequence
  node_sequence <- seq_len(nodes)

  # Set up distance matrix
  distance_matrix <- abs(outer(node_sequence, node_sequence, "-"))
  distance_matrix <- pmin(distance_matrix, nodes - distance_matrix)

  # Set edges
  lattice_edges <- lattice == 1
  A_edges <- A == 1

  # Set up smallworld distances
  smallworld <- matrix(0, nrow = nodes, ncol = nodes)

  # Get sorted rewired distances
  rewired_distances <- sort(distance_matrix[lower_triangle][(lattice_edges & !A_edges)[lower_triangle]])

  # Set index for rewired indices
  rewired_index <- !lattice_edges & A_edges

  # Set up lower triangle with rewired distances
  smallworld[lower_triangle][rewired_index[lower_triangle]] <- rewired_distances[
    rank(distance_matrix[lower_triangle][rewired_index[lower_triangle]], ties.method = "random")
  ]

  # Make symmetric
  smallworld <- smallworld + t(smallworld)

  # Multiply retained by distance matrix
  smallworld <- smallworld + ((lattice_edges & A_edges) * distance_matrix)

  # Obtain nonzero edges
  nonzero <- A[lower_triangle] != 0

  # Initialize network
  network <- matrix(0, nrow = nodes, ncol = nodes)

  # Generate edges (returned edges are sorted)
  edges <- generate_edges(nonzero = sum(nonzero), n = sample_size, p = nodes, snr = snr)

  # Set weights order
  # Follows: Muldoon, Bridgeford, & Bassett's (2016) implementation
  weight_order <- rank(smallworld[lower_triangle][nonzero], ties.method = "random")
  network[lower_triangle][nonzero] <- edges[weight_order]
  network <- network + t(network) # make symmetric

  # Get signs
  network <- determine_signs(network, nodes, negative_proportion)

  # Get population values
  R <- silent_call(pcor2cor(network))

  # Set lambda
  lambda <- NA

  # Check for positive definite
  if(anyNA(R) || !is_positive_definite(R)){

    # Try to condition network
    output <- try(condition_network(network, target_condition), silent = TRUE)

    # Check for whether correlation matrix is conditioned
    if(is(output, "try-error")){
      stop("Ill-conditioned")
    }

    # Store outputs
    lambda <- output$lambda

    # Update network and correlations
    P <- cor2pcor(output$R)
    P[network == 0] <- 0
    network <- P
    R <- pcor2cor(network)

    # Check for positive definite again
    if(anyNA(R) || !is_positive_definite(R)){
      stop("Not positive definite")
    }

  }

  # Check for condition
  condition <- fast_kappa(R)
  if(condition > (target_condition + 1)){
    stop("Condition number exceeds target")
  }

  # Check for maximum correlation
  if(max(abs(R[lower_triangle])) > max_correlation){
    stop("Max correlation exceeds target")
  }

  # Return results
  return(
    list(
      R = R,
      network = network,
      params = attr(edges, "params"),
      lambda = lambda,
      condition = condition
    )
  )

}

#' @noRd
# Determine balance of signs ----
# Updated 11.03.2026
determine_signs <- function(network, nodes, negative_proportion)
{

  # Check for zero negative weights
  if(negative_proportion == 0){
    return(network)
  }

  # Convert negative_proportion (edge-level) to node flip proportion
  flip_proportion <- (1 - sqrt(1 - 2 * negative_proportion)) / 2
  flip_number <- round(nodes * flip_proportion)

  # Flip at least one
  flip_number <- max(1, min(flip_number, floor(nodes / 2)))

  # Determine node indices
  index <- shuffle(seq_len(nodes), flip_number)

  # Flip nodes
  network[index,] <- -network[index,]
  network[,index] <- -network[,index]

  # Return the network
  return(network)

}

#' @title L0ggm-package
#'
#' @description Implements L0 norm regularization penalties for Gaussian graphical models
#'
#' @author Alexander P. Christensen <alexpaulchristensen@gmail.com>
#'
#' @useDynLib L0ggm, .registration = TRUE
#'
#' @importFrom methods is
#' @importFrom stats complete.cases cor cov2cor dist dnorm na.omit optimize pchisq pnorm predict qnorm rnorm sd setNames uniroot
#' @importFrom utils capture.output data packageVersion
#'
"_PACKAGE"
#> [1] "_PACKAGE"
#L0ggm----

#' @noRd
# Bounds for Weibull distribution
# Updated 14.03.2026
check_bounds <- function(shape, scale)
{
  # The bounds were derived from the 222 empirical datasets
  # from Huth et al. (2025), representing most "standard"
  # conditions that occur in empirical data (round up and down)
  # (see `?weibull_descriptives` for selection criterion)
  # These ranges will be updated as more empirical data
  # are aggregated
  any((shape < 0.70) | (shape > 1.70) | (scale < 0.03) | (scale > 0.19))
}

#' @noRd
# Generate from Weibull distribution
# Updated 08.03.2026
weibull_xoshiro <- function(n, shape, scale)
{
  return(scale * (-log(runif_xoshiro(n)))^(1 / shape))
}

#' @noRd
# Generate edge values
# Updated 16.03.2026
generate_edges <- function(nonzero, n, p, snr)
{

  # Calculate possible edges
  possible_edges <- p * (p - 1) / 2

  # Check for empirical parameter bounds and maximal partial correlation
  greater_than <- TRUE

  # Generate until no edges are greater than one
  while(greater_than){

    # Generate Weibull parameters
    params <- weibull_parameters(nodes = p, sample_size = n, snr = snr, bootstrap = TRUE)

    # Check outside of bounds
    if(check_bounds(params[["shape"]], params[["scale"]])){
      next
    }

    # Generate edge weights
    weights <- weibull_xoshiro(possible_edges, shape = params[["shape"]], scale = params[["scale"]])

    # Update greater than
    greater_than <- any(weights >= 0.85)
    # No weights greater than max weight = 0.8452469

  }

  # Create weighted sampling of weights
  weighted_weights <- weights^2

  # Set weights
  weights <- sort(
    weighted_shuffle(
      x = weights, size = nonzero, prob = weighted_weights / sum(weighted_weights)
    ), decreasing = TRUE
  )

  # Attach attributes to weights
  attr(weights, "params") <- params

  # Return weights
  return(weights)

}

#' @noRd
# Minimally condition the network
# Updated 11.03.2026
condition_network <- function(network, target_condition)
{

  # Obtain pre-inversion matrix
  Omega <- -network
  diag(Omega) <- 1

  # Compute one plus minimum eigenvalue
  min_eigen <- 1 + abs(min(matrix_eigenvalues(Omega)))

  # Condition network
  opt <- uniroot(
    f = function(lambda){

      # Create precision
      diag(Omega) <- min_eigen + lambda

      # Compute the precision matrix
      K <- try(solve(Omega), silent = TRUE)

      # Catch bad matrices
      if(is(K, "try-error")){
        return(target_condition)
      }else{
        return(fast_kappa(cov2cor(K)) - target_condition)
      }

    }, interval = c(1e-05, 0.30) # max ~23% shrinkage (Peeters et al., 2020)
  )

  # Set diagonal on precision
  diag(Omega) <- min_eigen + opt$root

  # Return results
  return(list(R = cov2cor(solve(Omega)), lambda = opt$root))

}

#' @noRd
# Simulate data
# Updated 14.03.2026
simulate_data <- function(n, p, R, skew, skew_range)
{

  # Check for skew range
  if(!is.null(skew_range)){
    typeof_error(skew_range, "numeric") # object type error
    length_error(skew_range, 2) # object length error
    range_error(skew_range, c(-2, 2)) # object range error
    possible_skews <- seq(-2, 2, 0.05) # possible skews
    skew_range <- round(skew_range, 2) # get to hundredths digit
    min_range <- abs(min(skew_range) - possible_skews) # difference for minimum
    min_skew <- possible_skews[which.min(min_range)] # get minimum skew
    max_range <- abs(max(skew_range) - possible_skews) # difference for maximum
    max_skew <- possible_skews[which.min(max_range)] # get maximum skew
    skew <- seq(min_skew, max_skew, 0.05) # obtain skews
  }

  # Generate data
  data <- matrix(rnorm_ziggurat(n * p), nrow = n, ncol = p) %*% chol(R)

  # Set skew
  n_skew <- length(skew)
  if(n_skew == 1){
    skew <- rep(skew, p)
  }else if(n_skew != p){
    skew <- shuffle_replace(skew, p)
  }

  # Loop through columns
  node_sequence <- seq_len(p)
  for(i in node_sequence){
    data[,i] <- skew_continuous(skew = skew[i], data = data[,i])
  }

  # Add column names to data
  names(skew) <- colnames(data) <- paste0(
    "V", format_integer(node_sequence, digits(p) - 1)
  )

  # Return data and skew
  return(list(data = data, skew = skew))

}

#' Simulates Stochastic Block Model Data
#'
#' @description
#' Simulates data from a Gaussian Graphical Model (GGM) with a stochastic
#' block model (SBM) structure. Nodes are partitioned into communities, with
#' edge density controlled separately within and between communities via a
#' \code{blocks x blocks} density matrix. Absolute edge weights are drawn from
#' a Weibull distribution whose parameters are predicted from the network size,
#' sample size, and signal-to-noise ratio using a Seemingly Unrelated
#' Regression (SUR) model fitted to 194 empirical psychometric networks
#' (Huth et al., 2025). The resulting population partial correlation matrix is
#' used to generate multivariate normal (or skewed) data.
#'
#' A \code{diffusion} parameter controls the minimum proportion of the
#' strongest edges that are reserved for within-community positions. Because
#' the remaining edges are shuffled randomly, the effective within-community
#' advantage will generally exceed \code{1 - diffusion}; see Details.
#'
#' Parameters do not have default values (except \code{negative_proportion},
#' \code{diffusion}, \code{target_condition}, \code{max_correlation}, and
#' \code{max_iterations}) and must each be set. See Details and Examples to
#' get started.
#'
#' @param nodes Numeric (length = 1 or \code{blocks}).
#' Number of nodes per community block. Can be a single value applied to all
#' blocks, or a vector of length \code{blocks} specifying each block's size
#' individually. Minimum of three nodes per block. The total number of nodes
#' (\code{sum(nodes)}) should be between 8 and 54 to remain within the range
#' of the empirical networks used to fit the Weibull parameter model; values
#' outside this range are accepted but will trigger extrapolation and may
#' produce a warning from \code{\link{weibull_parameters}}.
#'
#' @param blocks Numeric (length = 1).
#' Number of community blocks.
#'
#' @param density_matrix Matrix (dim = \code{blocks x blocks}).
#' A symmetric numeric matrix specifying edge probabilities within and between
#' community blocks. Diagonal entry \eqn{[i,i]} gives the within-block edge
#' probability for community \eqn{i}; off-diagonal entry \eqn{[i,j]}
#' (\eqn{i \neq j}) gives the between-block edge probability for the pair of
#' communities \eqn{i} and \eqn{j}. All entries must be in \eqn{[0, 1]} and
#' the matrix must be symmetric (i.e., \code{density_matrix[i,j] ==
#' density_matrix[j,i]}). See Details for construction guidance.
#'
#' @param snr Numeric (length = 1).
#' Signal-to-noise ratio of the absolute partial correlation weights,
#' defined as \eqn{\bar{|w|} / \mathrm{SD}(|w|)}. This ratio governs the
#' shape and scale of the Weibull distribution used to generate edge weights:
#' values below 1 produce wider, more heterogeneous weight distributions;
#' values above 1 produce narrower, more homogeneous distributions. Note that
#' the same SNR can arise from different combinations of mean and standard
#' deviation (e.g., mean = 0.05, SD = 0.05 and mean = 0.15, SD = 0.15 both
#' give SNR = 1), so SNR alone does not determine the absolute magnitude of
#' edge weights, which is additionally governed by \code{nodes} and
#' \code{sample_size} via \code{\link{weibull_parameters}}. Empirically
#' observed SNR values ranged from 0.648 to 1.712; values outside this range
#' are accepted but will trigger a warning. Defaults to \code{1}.
#'
#' @param negative_proportion Numeric (length = 1).
#' Proportion of between-community edges assigned a negative sign (inhibitory
#' connections). Each between-community edge is independently signed negative
#' with this probability (i.e., a Bernoulli draw per edge), so the realized
#' proportion will vary around the specified value. Must be between 0 and 1.
#' Applies only to between-community edges; all within-community edges are
#' positive. If not provided, a value is sampled from a truncated normal
#' distribution reflecting the empirical distribution of true negative partial
#' correlations across 194 psychometric networks: mean = 0.34, SD = 0.086,
#' bounded to \eqn{[0.083, 0.55]}.
#'
#' @param diffusion Numeric (length = 1).
#' Controls the minimum proportion of the top-ranked edges (by absolute
#' weight) that are guaranteed to be placed within communities rather than
#' distributed freely across the network. Specifically, \code{1 - diffusion}
#' of the within-community edges are reserved for the highest-weight draws;
#' the remaining edges (both within- and between-community) are filled from
#' a randomly shuffled pool. Because shuffled edges can land within communities
#' by chance, the actual proportion of top edges that end up within communities
#' will on average exceed \code{1 - diffusion}. Lower values of
#' \code{diffusion} (e.g., \code{0.10}) therefore produce stronger community
#' contrast than higher values (e.g., \code{0.90}), but \code{diffusion}
#' should be interpreted as a floor on within-community weight concentration,
#' not an exact control. Defaults to \code{0.30}. Must be between 0 and 1.
#'
#' @param diffusion_range Numeric (length = 2).
#' If provided, overrides \code{diffusion} by drawing the diffusion proportion
#' uniformly from this interval on each call. Useful for introducing
#' replication-to-replication variability in community contrast. For example,
#' \code{diffusion_range = c(0.05, 0.20)} samples a value between 5\% and
#' 20\% on each draw. The same floor interpretation applies as for
#' \code{diffusion}. Both values must be between 0 and 1.
#'
#' @param sample_size Numeric (length = 1).
#' Number of observations to generate from the population multivariate
#' distribution. Also influences the predicted Weibull scale parameter via
#' \code{\link{weibull_parameters}}: larger samples are associated with
#' smaller, more precisely estimated edge weights.
#'
#' @param skew Numeric (length = 1 or \code{sum(nodes)}).
#' Skew applied to each variable after generation from the multivariate
#' normal. Can be a single value applied to all variables, or one value per
#' variable. Values are rounded to the nearest 0.05 increment and must be
#' in \eqn{[-2, 2]}. Defaults to \code{0} (no skew).
#'
#' @param skew_range Numeric (length = 2).
#' If provided, overrides \code{skew} by drawing a skew value independently
#' and uniformly from this interval for each variable. Both values must be
#' in \eqn{[-2, 2]}.
#'
#' @param target_condition Numeric (length = 1).
#' Target condition number (computed via \code{\link{kappa}} with
#' \code{exact = TRUE}) used when ridge regularization is required to recover
#' a positive definite precision matrix. The smallest ridge penalty
#' \eqn{\lambda} that brings the condition number to this target is found via
#' root-finding (\code{\link{uniroot}}), subject to a maximum shrinkage of
#' approximately 23\% following Peeters et al. (2020). After conditioning,
#' the Weibull bounds are re-verified on the updated edge weights; draws that
#' fall outside the empirical bounds after conditioning are rejected. Lower
#' values produce better-conditioned (more stable) matrices. Defaults to
#' \code{30}. Values up to \code{100} are accepted but not recommended.
#'
#' @param max_correlation Numeric (length = 1).
#' Maximum allowed absolute pairwise correlation in the population correlation
#' matrix \code{R}. Any draw where \code{max(abs(R[lower.tri(R)])) >
#' max_correlation} is rejected and a new attempt is made. Must be between
#' 0 and 1. Defaults to \code{0.80}.
#'
#' @param max_iterations Numeric (length = 1).
#' Maximum number of attempts to find a connected network with valid edge
#' weights before stopping with an error. The error message reports a
#' frequency table of rejection reasons to assist with diagnosing convergence
#' failures. Defaults to \code{100}.
#'
#' @details
#' \strong{Constructing \code{density_matrix}}
#'
#' The \code{density_matrix} is a \code{blocks x blocks} symmetric matrix
#' where entry \eqn{[i,i]} gives the within-block edge probability for
#' community \eqn{i}, and entry \eqn{[i,j]} (for \eqn{i \neq j}) gives the
#' between-block edge probability for communities \eqn{i} and \eqn{j}. The
#' simplest construction uses a uniform off-diagonal density with
#' block-specific diagonals:
#'
#' \preformatted{
#' # Uniform within (0.90) and between (0.20) density for 3 blocks
#' dm <- matrix(0.20, nrow = 3, ncol = 3)
#' diag(dm) <- 0.90
#' }
#'
#' For asymmetric community structure, each diagonal entry can differ:
#'
#' \preformatted{
#' # Varying within-block density per community
#' dm <- matrix(0.20, nrow = 3, ncol = 3)
#' diag(dm) <- c(0.85, 0.90, 0.95)
#' }
#'
#' For full pairwise control over between-block densities, specify the
#' complete symmetric matrix directly:
#'
#' \preformatted{
#' dm <- matrix(c(
#'   0.90, 0.20, 0.10,
#'   0.20, 0.85, 0.30,
#'   0.10, 0.30, 0.95
#' ), nrow = 3, ncol = 3)
#' }
#'
#' \strong{Diffusion and within-community weight concentration}
#'
#' The \code{diffusion} parameter does not exactly fix the proportion of
#' top-ranked edges placed within communities. Instead, \code{1 - diffusion}
#' of the within-community edge slots are filled deterministically from the
#' highest-weight draws. The remaining edge slots (within- and
#' between-community alike) are filled from a randomly shuffled pool of
#' lower-ranked weights, meaning some additional high-weight edges will land
#' within communities by chance. The realized within-community weight
#' advantage will therefore always be at least as large as implied by
#' \code{1 - diffusion}, and typically larger. The \code{Q} field in the
#' returned \code{parameters} list (Newman-Girvan modularity) provides a
#' post-hoc summary of the actual community contrast achieved.
#'
#' @return A named list with four elements:
#'
#' \item{data}{Numeric matrix of dimension \code{sample_size x sum(nodes)}
#' containing the simulated observations drawn from the population GGM.
#' Rows are cases; columns are variables named \code{V01}, \code{V02},
#' etc. Values are continuous (or skewed continuous when \code{skew != 0}).
#' To produce ordinal data, pass the columns through
#' \code{\link{categorize}}.}
#'
#' \item{parameters}{
#' A list of input, derived, and estimated parameters:
#' \itemize{
#'   \item \code{nodes} --- Integer vector of length \code{blocks} giving the
#'   number of nodes per block (scalar input is expanded to this length)
#'   \item \code{blocks} --- Number of community blocks
#'   \item \code{sample_size} --- Number of simulated observations
#'   \item \code{skew} --- Named numeric vector of per-variable skew values
#'   actually applied (after rounding and possible resampling)
#'   \item \code{density_matrix} --- The \code{blocks x blocks} density matrix
#'   as supplied
#'   \item \code{negative_proportion} --- The proportion of between-community
#'   edges assigned a negative sign, either as supplied or as sampled from the
#'   empirical distribution
#'   \item \code{weibull} --- Named numeric vector of length 2 giving the
#'   Weibull \code{shape} and \code{scale} parameters of the absolute edge
#'   weight distribution actually used. If ridge conditioning was applied,
#'   these are re-estimated from the conditioned network via MLE.
#'   \item \code{diffusion} --- Numeric vector of length 2 giving the
#'   within-block reservation range as a proportion of within-community edges,
#'   on the internal scale used by the sampler (i.e., \code{range(1 -
#'   diffusion)} or \code{range(1 - diffusion_range)}). Both values are equal
#'   when \code{diffusion} is scalar. Note this is the complement of the
#'   user-supplied \code{diffusion} value and represents the fraction of
#'   within-community slots filled from the top-ranked draws.
#'   \item \code{omega} --- Smallworldness omega statistic of the generated
#'   network (Telesford et al., 2011). Values near zero indicate small-world
#'   structure; negative values indicate lattice-like structure; positive
#'   values indicate random-like structure (see \code{\link[L0ggm]{smallworldness}})
#'   \item \code{Q} --- Newman-Girvan modularity of the population network
#'   (\code{Omega}) with respect to the block membership, computed via
#'   \code{igraph::modularity} on absolute edge weights. Provides a summary
#'   of the community contrast actually achieved after weight assignment and
#'   any ridge conditioning.
#' }
#' }
#'
#' \item{population}{
#' Population-level network parameters:
#' \itemize{
#'   \item \code{R} --- Population correlation matrix derived from the GGM
#'   via \code{pcor2cor}
#'   \item \code{Omega} --- Population partial correlation matrix (the GGM
#'   edge weight matrix), with zeros for absent edges
#'   \item \code{membership} --- Named integer vector of length
#'   \code{sum(nodes)} giving the community block assignment (1 to
#'   \code{blocks}) for each node
#' }
#' }
#'
#' \item{convergence}{
#' Iteration and conditioning diagnostics:
#' \itemize{
#'   \item \code{iterations} --- Number of sampling attempts needed to find a
#'   valid network (including graph structure and edge weight draws)
#'   \item \code{rejections} --- Character vector of length
#'   \code{max_iterations + 1} recording the rejection reason for each failed
#'   attempt; entries for successful or unused iterations are empty strings.
#'   Common reasons include disconnected graph structure, condition number
#'   exceeding \code{target_condition}, maximum correlation exceeding
#'   \code{max_correlation}, and Weibull parameters falling outside empirical
#'   bounds after ridge conditioning.
#'   \item \code{lambda} --- Ridge regularization parameter \eqn{\lambda}
#'   added to the diagonal of the precision matrix to ensure positive
#'   definiteness; \code{NA} if no conditioning was required
#'   \item \code{condition} --- Condition number of the final population
#'   correlation matrix \code{R}, computed via \code{kappa} with
#'   \code{exact = TRUE}
#' }
#' }
#'
#' @examples
#' # Construct density matrix for 3 blocks with uniform densities
#' dm <- matrix(0.20, nrow = 3, ncol = 3)
#' diag(dm) <- 0.90
#'
#' # Basic 3-block simulation with equal-sized communities
#' result <- simulate_sbm(
#'   nodes = 6, # 6 nodes per block = 18 total
#'   blocks = 3,
#'   sample_size = 500,
#'   density_matrix = dm
#' )
#'
#' # Unequal block sizes
#' result <- simulate_sbm(
#'   nodes = c(4, 6, 8), # 18 total nodes
#'   blocks = 3,
#'   sample_size = 500,
#'   density_matrix = dm
#' )
#'
#' # Varying within-block density per community
#' dm_varying <- matrix(0.20, nrow = 3, ncol = 3)
#' diag(dm_varying) <- c(0.85, 0.90, 0.95)
#'
#' result <- simulate_sbm(
#'   nodes = 6,
#'   blocks = 3,
#'   sample_size = 500,
#'   density_matrix = dm_varying
#' )
#'
#' # Full pairwise between-block density control
#' dm_pairwise <- matrix(c(
#'   0.90, 0.20, 0.10,
#'   0.20, 0.85, 0.30,
#'   0.10, 0.30, 0.95
#' ), nrow = 3, ncol = 3)
#'
#' result <- simulate_sbm(
#'   nodes = 6,
#'   blocks = 3,
#'   sample_size = 500,
#'   density_matrix = dm_pairwise
#' )
#'
#' # Fix the proportion of negative between-community edges
#' result <- simulate_sbm(
#'   nodes = 6,
#'   blocks = 3,
#'   sample_size = 500,
#'   density_matrix = dm,
#'   negative_proportion = 0.20
#' )
#'
#' # Introduce variability in diffusion across replications
#' result <- simulate_sbm(
#'   nodes = 6,
#'   blocks = 3,
#'   sample_size = 500,
#'   density_matrix = dm,
#'   diffusion_range = c(0.30, 0.70)
#' )
#'
#' @author Alexander P. Christensen <alexpaulchristensen@gmail.com>
#'
#' @references
#' \strong{Seminal introduction to Stochastic Block Models} \cr
#' Holland, P. W., Laskey, K. B., & Leinhardt, S. (1983).
#' Stochastic blockmodels: First steps.
#' \emph{Social Networks}, \emph{5}(2), 109--137.
#'
#' \strong{Empirical network data used to fit the Weibull SUR model} \cr
#' Huth, K. B. S., Haslbeck, J. M. B., Keetelaar, S., Van Holst, R. J., &
#' Marsman, M. (2025). Statistical evidence in psychological networks.
#' \emph{Nature Human Behaviour}.
#'
#' \strong{Maximum ridge shrinkage bound} \cr
#' Peeters, C. F., van de Wiel, M. A., & van Wieringen, W. N. (2020).
#' The spectral condition number plot for regularization parameter evaluation.
#' \emph{Computational Statistics}, \emph{35}(2), 629--646.
#'
#' @export
#'
# Simulate SBM GGM data ----
# Updated 15.05.2026
simulate_sbm <- function(
    nodes, blocks, density_matrix,
    snr = 1, diffusion = 0.30, diffusion_range = NULL,
    negative_proportion, sample_size,
    skew = 0, skew_range = NULL,
    target_condition = 30, max_correlation = 0.80,
    max_iterations = 100
)
{

  # Check for missing negative proportion
  if(missing(negative_proportion)){

    # Compute proportion of negative edges based on empirical values
    negative_proportion <- pmin(
      pmax(
        0.34 + rnorm_ziggurat(1) * 0.086, # empirical mean +/- 1 SD
        0.083 # empirical minimum
      ), 0.55 # empirical maximum
    )

  }

  # Check for input errors
  ## Returns updated nodes to ensure same length as blocks
  nodes <- simulate_sbm_errors(
    nodes, blocks, density_matrix,
    snr, negative_proportion, sample_size, skew, diffusion,
    target_condition, max_correlation, max_iterations
  )

  # Determine total number of nodes
  total_nodes <- sum(nodes)

  # Get community sequence
  community_sequence <- seq_len(blocks)

  # Set membership
  membership <- sort(rep(community_sequence, times = nodes))

  # Set up membership matrix
  membership_matrix <-  outer(membership, membership, "==")

  # Obtain lower triangle
  lower_triangle <- lower.tri(membership_matrix)

  # Initialize iterations
  iter <- 0

  # Initialize condition
  search <- TRUE

  # Collect rejections
  rejections <- character(length = max_iterations + 1)

  # Generation loop
  while(search){

    # Increase iterations
    iter <- iter + 1

    # Stop on greater than max iterations
    if(iter > max_iterations){

      # Collect rejections
      rejection_table <- fast_table(rejections)

      # Remove NULL
      rejection_table <- rejection_table[names(rejection_table) != ""]

      # Return error
      stop(
        paste0(
          "Reached maximum iterations: Could not find weights that converge for the SBM structure. ",
          "Resulting rejections were due to:\n\n",
          paste0(names(rejection_table), " = ", rejection_table, collapse = "\n")
        )
      )

    }

    # Get block matrix
    block_matrix <- generate_sbm(
      membership_matrix, blocks, membership,
      membership_matrix, density_matrix, lower_triangle
    )

    # Check block matrix
    if(!igraph::is_connected(convert2igraph(block_matrix))){

      # Add rejection reason
      rejections[iter] <- "Could not find structure where every node was connected."

      # Move to next iteration
      next

    }

    # Try to get good weights
    output <- try(
      sbm_weights(
        block_matrix, membership, membership_matrix, sample_size,
        snr, total_nodes, negative_proportion,
        diffusion, diffusion_range, target_condition,
        max_correlation, lower_triangle
      ), silent = TRUE
    )

    # Check for error
    if(is(output, "try-error")){

      # Add rejection reason
      rejections[iter] <- trimws(strsplit(output[1], split = "\n")[[1]][2])

      # Move to next iteration
      next

    }

    # Update search
    search <- FALSE

  }

  # Generate data
  data_output <- simulate_data(
    n = sample_size, p = total_nodes, R = output$R, skew = skew, skew_range = skew_range
  )

  # Return parameters
  return(
    list(
      data = data_output$data,
      parameters = list(
        nodes = nodes,
        blocks = blocks,
        sample_size = sample_size,
        skew = data_output$skew,
        density_matrix = density_matrix,
        negative_proportion = negative_proportion,
        weibull = output$params,
        diffusion = output$diffusion,
        omega = smallworldness(network = output$network, method = "omega"),
        Q = igraph::modularity(convert2igraph(abs(output$network)), output$membership)
      ),
      population = list(
        R = output$R,
        Omega = output$network,
        membership = output$membership
      ),
      convergence = list(
        iterations = iter,
        rejections = rejections,
        lambda = output$lambda,
        condition = output$condition
      )
    )
  )

}

# Bug checking ----
# nodes = 6; blocks = 3; sample_size = 1000
# density_matrix = matrix(0.20, nrow = blocks, ncol = blocks)
# diag(density_matrix) <- 0.90
# snr = 1; negative_proportion = 0.35
# skew = 0; skew_range = NULL
# diffusion = 0.05; diffusion_range = NULL
# target_condition = 10;
# max_correlation = 0.80; max_iterations = 100

#' @noRd
# Errors ----
# Updated 14.03.2026
simulate_sbm_errors <- function(
    nodes, blocks, density_matrix,
    snr, negative_proportion, sample_size, skew, diffusion,
    target_condition, max_correlation, max_iterations
)
{

  # Errors for 'blocks'
  typeof_error(blocks, "numeric", "simulate_sbm")
  length_error(blocks, 1, "simulate_sbm")
  range_error(blocks, c(1, Inf), "simulate_sbm")

  # Errors for 'nodes'
  typeof_error(nodes, "numeric", "simulate_sbm")
  length_error(nodes, c(1, blocks), "simulate_sbm")
  range_error(nodes, c(3, Inf), "simulate_sbm")

  # Ensure 'nodes' is the length of blocks
  if(length(nodes) != blocks){
    nodes <- rep(nodes, blocks)
  }

  # Object error for 'density_matrix'
  object_error(density_matrix, c("matrix", "data.frame", "tibble"), "simulate_sbm")

  # Check for matrix
  if(get_object_type(density_matrix) != "matrix"){
    density_matrix <- as.matrix(as.data.frame(density_matrix))
  }

  # Errors for 'density_matrix'
  typeof_error(density_matrix, "numeric", "simulate_sbm")
  length_error(density_matrix, blocks * blocks, "simulate_sbm")
  range_error(density_matrix, c(0, 1), "simulate_sbm")

  # Symmetric error for 'density_matrix'
  if(!is_symmetric(density_matrix)){
    stop("`simulate_sbm` can only handle symmetric 'density_matrix' input.")
  }

  # Error for 'snr'
  typeof_error(snr, "numeric", "simulate_sbm")
  length_error(snr, 1, "simulate_sbm")
  range_error(snr, c(0.01, Inf), "simulate_sbm")

  # Send warning if beyond bounds
  if((snr < 0.64) | (snr > 1.72)){
    warning(paste0(
      "The signal-to-ratio specified (", round(snr, 2), ") is beyond ",
      "the bounds of cataloged empirical data (0.64-1.72). "
    ))
  }

  # Errors for 'negative_proportion'
  typeof_error(negative_proportion, "numeric", "simulate_sbm")
  length_error(negative_proportion, 1, "simulate_sbm")
  range_error(negative_proportion, c(0, 1), "simulate_sbm")

  # Errors for 'sample_size'
  typeof_error(sample_size, "numeric", "simulate_sbm")
  length_error(sample_size, 1, "simulate_sbm")
  range_error(sample_size, c(1, Inf), "simulate_sbm")

  # Errors for 'skew'
  typeof_error(skew, "numeric", "simulate_sbm")
  length_error(skew, c(1, sum(nodes)), "simulate_sbm")
  range_error(skew, c(-2, 2), "simulate_sbm")

  # Errors for 'diffusion'
  typeof_error(diffusion, "numeric", "simulate_sbm")
  length_error(diffusion, 1, "simulate_sbm")
  range_error(diffusion, c(0, 1), "simulate_sbm")

  # Errors for 'target_condition'
  typeof_error(target_condition, "numeric", "simulate_sbm")
  length_error(target_condition, 1, "simulate_sbm")
  range_error(target_condition, c(1, 100), "simulate_sbm")

  # Errors for 'max_correlation'
  typeof_error(max_correlation, "numeric", "simulate_sbm")
  length_error(max_correlation, 1, "simulate_sbm")
  range_error(max_correlation, c(0, 1), "simulate_sbm")

  # Errors for 'max_iterations'
  typeof_error(max_iterations, "numeric", "simulate_sbm")
  length_error(max_iterations, 1, "simulate_sbm")
  range_error(max_iterations, c(1, Inf), "simulate_sbm")

  # Return updated nodes
  return(nodes)

}

#' @noRd
# Generate SBM ----
# Updated 14.03.2026
generate_sbm <- function(
    block_matrix, blocks, membership,
    membership_matrix, density_matrix, lower_triangle
)
{

  # Set upper triangles to FALSE
  block_matrix[!lower_triangle] <- FALSE

  # Set up lower block matrix
  for(i in seq_len(blocks)){

    # Get community index
    community_index <- membership == i

    # Get block edges
    community_block <- block_matrix[community_index, community_index]

    # Compute probabilities
    community_block[community_block] <- runif_xoshiro(sum(community_block)) < density_matrix[i,i]

    # Insert block back into block matrix
    block_matrix[community_index, community_index] <- community_block

    # Check for off-blocks
    if(i != blocks){

      # Set off-block
      for(j in (i + 1):blocks){

        # Get off-block index
        off_index <- membership == j

        # Get block edges
        off_block <- block_matrix[off_index, community_index]

        # Set index
        not_index <- !off_block

        # Compute probabilities
        off_block[not_index] <- runif_xoshiro(sum(not_index)) < density_matrix[i,j]

        # Insert block back into block matrix
        block_matrix[off_index, community_index] <- off_block

      }

    }

  }

  # Return block matrix
  return(block_matrix + t(block_matrix))

}

#' @noRd
# SBM weight generation ----
# Updated 14.03.2026
sbm_weights <- function(
    block_matrix, membership, membership_matrix, sample_size,
    snr, total_nodes, negative_proportion,
    diffusion, diffusion_range, target_condition,
    max_correlation, lower_triangle
)
{

  # Check for diffusion range
  if(!is.null(diffusion_range)){
    typeof_error(diffusion_range, "numeric") # object type error
    length_error(diffusion_range, 2) # object length error
    range_error(diffusion_range, c(0, 1)) # object range error
    diffusion <- diffusion_range
  }else{
    diffusion <- c(diffusion, diffusion)
  }

  # Reverse diffusion
  diffusion <- range(1 - diffusion)

  # Set lower triangles
  block_lower <- block_matrix[lower_triangle]
  membership_lower <- membership_matrix[lower_triangle]

  # Generate edges (returned edges are sorted)
  total_edges <- sum(block_lower)
  edges <- generate_edges(nonzero = total_edges, n = sample_size, p = total_nodes, snr = snr)

  # Obtain edges
  within_index <- membership_lower & block_lower
  total_within <- sum(within_index)

  # Set diffusion parameter
  within_reserved <- round(total_within * runif_xoshiro(1, min = diffusion[1], max = diffusion[2]))
  within_random <- total_within - within_reserved
  random_sequence <- seq_len(within_random)

  # Top edges guaranteed to go into communities
  reserved_index <- seq_len(within_reserved)

  # Identify between-community edges
  between_sequence <- total_edges - within_reserved
  between_index <- swiftelse(
    within_random > 0,
    seq_len(between_sequence)[-random_sequence],
    seq_len(between_sequence)
  )
  n_between <- length(between_index)

  # Remaining edge weights
  remaining_weights <- shuffle(edges[-reserved_index])

  # Check that there is between weights to assign
  if(n_between > 0){

    # Get signs
    signs <- swiftelse(runif_xoshiro(n_between) < negative_proportion, -1, 1)

    # Set between-community edges
    block_matrix[lower_triangle][!membership_lower & block_lower] <-
    shuffle(remaining_weights[between_index]) * signs

  }

  # Set within-community edges
  block_matrix[lower_triangle][within_index] <- shuffle(
    c(edges[reserved_index], remaining_weights[random_sequence])
  )

  # Set upper triangle to zero
  block_matrix[!lower_triangle] <- 0

  # Make symmetric
  network <- block_matrix + t(block_matrix)

  # Add names to network
  names(membership) <- row.names(network) <- colnames(network) <- paste0(
    "V", format_integer(seq_len(total_nodes), digits(total_nodes) - 1)
  )

  # Get population values
  R <- silent_call(pcor2cor(network))

  # Set condition and lambda
  lambda <- NA

  # Check for positive definite
  if(anyNA(R) || !is_positive_definite(R)){

    # Try to condition network
    output <- try(condition_network(network, target_condition), silent = TRUE)

    # Check for whether correlation matrix is conditioned
    if(is(output, "try-error")){
      stop("Ill-conditioned")
    }

    # Store outputs
    lambda <- output$lambda

    # Update network and correlations
    P <- cor2pcor(output$R)
    P[network == 0] <- 0
    network <- P
    R <- pcor2cor(network)

    # Check for positive definite again
    if(anyNA(R) || !is_positive_definite(R)){
      stop("Not positive definite")
    }

  }

  # Check for condition
  condition <- fast_kappa(R)
  if(condition > (target_condition + 1)){
    stop("Condition number exceeds target")
  }

  # Check for maximum correlation
  if(max(abs(R[lower_triangle])) > max_correlation){
    stop("Max correlation exceeds target")
  }

  # Return results
  return(
    list(
      R = R,
      network = network,
      membership = membership,
      params = attr(edges, "params"),
      diffusion = diffusion,
      lambda = lambda,
      condition = condition
    )
  )

}

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%#
#### Maximum Likelihood Functions ####
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%#

#' @noRd
# MLE (Folded) Gumbel Scale Parameter ----
# Updated 11.04.2026
gumbel_mle <- function(x)
{

  # Set up MLE for scale
  scale_mle <- function(scale, x)
  {

    # Pre-compute reused values
    x_scale <- x / scale
    neg_x_scale <- -x_scale

    # Compute log density
    density <- log(
      exp(neg_x_scale - exp(neg_x_scale)) +
        exp(x_scale - exp(x_scale))
    ) - log(scale)

    # Return log-likelihood
    return(-sum(density))

  }

  # Return parameters
  return(optimize(f = scale_mle, interval = c(1e-04, max(x)), x = x)$minimum)

}

#' @noRd
# MLE Weibull Parameters ----
# Updated 06.04.2026
weibull_mle <- function(x)
{

  # Ensure no exact zeros to avoid log(0)
  x <- pmax.int(x, .Machine$double.eps)

  # Set up MLE for shape
  shape_mle <- function(k, x, n)
  {

    # Pre-compute reused values
    x_k <- x^k
    log_x <- log(x)

    # Return log-likelihood
    return(sum(x_k * log_x) / sum(x_k) - 1 / k - sum(log_x) / n)

  }

  # Obtain MLE estimate
  shape <- uniroot(
    f = shape_mle, interval = c(0.01, 10),
    x = x, n = length(x), extendInt = "upX"
  )$root

  # Return parameters
  return(c(shape = shape, scale = mean(x^shape)^(1 / shape)))

}

#' @title Construct a Degree-Preserving Ring Lattice via Proximity-Swap Construction
#'
#' @description Converts a network matrix into a connected ring lattice whose
#' degree sequence exactly matches the original, while maximizing the average
#' local clustering coefficient. The resulting lattice is intended as a null
#' model for small-world network analyses. An adjacency matrix is derived
#' automatically from \code{network} (non-zero entries become edges), so the
#' function accepts any weighted or binary network directly.
#'
#' Each pass randomly permutes the degree sequence onto ring positions, then
#' greedily assigns edges in strictly increasing ring-distance order (proximity
#' construction). Any residual degree deficit is resolved by a swap-repair
#' phase. Passes that yield a disconnected graph or an unsatisfied degree
#' sequence are discarded; among valid passes the one with the highest average
#' clustering coefficient is returned. \code{shuffles} independent passes are
#' attempted in total.
#'
#' @param network Matrix.
#' A square, symmetric numeric matrix representing a network (e.g., partial
#' correlations). Non-zero off-diagonal entries are treated as edges; the
#' binary adjacency is derived internally as \code{network != 0}.
#' Isolated nodes (degree zero) are supported.
#'
#' @param weighted Logical (length = 1).
#' Whether to return a weighted ring lattice. When \code{TRUE}, the edge
#' weights from \code{network} are reassigned to the lattice edges according
#' to ring distance following the implementation of Muldoon, Bridgeford, &
#' Bassett (2016): shorter-distance lattice edges receive larger weights,
#' preserving the overall weight distribution while concentrating stronger
#' connections locally. When \code{FALSE} (default), a binary adjacency
#' matrix is returned.
#'
#' @param shuffles Numeric (length = 1).
#' Number of independent random permutation passes to attempt. Each pass
#' assigns the observed degree sequence to ring positions in a new random
#' order, runs the proximity construction, and applies swap repair if needed.
#' Only passes producing a connected graph with zero degree error are
#' retained; the one with the highest clustering coefficient is returned.
#' Defaults to \code{100}.
#'
#' @return A square symmetric matrix of the same dimension as \code{network},
#' with row and column names preserved, representing the resulting ring
#' lattice. When \code{weighted = FALSE} (default), entries are binary integers
#' (\code{0L}/\code{1L}); when \code{weighted = TRUE}, entries contain the
#' reassigned edge weights from \code{network} with original signs preserved.
#' The average clustering coefficient of the returned graph is attached as the
#' attribute \code{"CC"} and can be retrieved with \code{attr(result, "CC")}.
#' When the empirical fallback is triggered (see Details), \code{"CC"} reflects
#' the empirical network's clustering coefficient rather than a lattice's.
#'
#' @details
#' ## Algorithm
#'
#' \strong{Pair precomputation.} A ring distance matrix is computed as
#' \eqn{d_{ij} = \min(|i - j|,\, n - |i - j|)}, giving true circular
#' distances bounded by \eqn{\lfloor n/2 \rfloor}. All unique unordered pairs
#' at each distance \eqn{r = 1, \ldots, \lfloor n/2 \rfloor} are extracted
#' once (retaining only entries where \eqn{i < j}) and cached for reuse
#' across all passes.
#'
#' \strong{Proximity construction.} The observed degree sequence is randomly
#' permuted onto ring positions. Starting from an empty graph, edges are added
#' in distance order. At each distance band the eligible pairs (both endpoints
#' have remaining degree budget and are not yet connected) are sorted by
#' descending \eqn{\min(\text{budget}_i, \text{budget}_j)} so that high-need
#' pairs receive short connections first. Pairs are then assigned sequentially
#' with per-pair budget re-checks. The pass exits early once all budgets reach
#' zero. This phase is implemented in compiled C via \code{proximity_pass_c}.
#'
#' \strong{Swap repair.} If any degree deficit remains after the proximity
#' pass, the highest-deficit node \eqn{i} is connected to its nearest
#' available ring neighbour, scanning clockwise and counter-clockwise
#' positions in interleaved distance order. When no direct partner with
#' remaining budget exists, an edge swap is performed: a nearby unconnected
#' node \eqn{j} is found, one of \eqn{j}'s existing edges to \eqn{k} is
#' removed, and a new edge \eqn{(i, j)} is added. Node \eqn{k} recovers its
#' budget for resolution in a subsequent iteration. This repeats until all
#' deficits are resolved or the iteration cap (\eqn{2n^2}) is reached. This
#' phase is implemented in compiled C via \code{swapping_pass_c}.
#'
#' \strong{Weight assignment.} When \code{weighted = TRUE}, edge weights are
#' reassigned after the binary topology is finalised. The observed weights are
#' sorted by \emph{absolute value} in descending order (largest magnitude first)
#' and mapped onto lattice edges ranked by ascending ring distance (i.e.,
#' \eqn{d_{ij} = \min(|i - j|,\, n - |i - j|)}), so that shorter (more local)
#' connections receive the largest-magnitude weights. Original signs are
#' preserved: the sorted weight vector — not its absolute values — is assigned
#' to the lattice edges. This directly implements the distance-weight principle
#' of Muldoon, Bridgeford, & Bassett (2016), using the ring's structural
#' distances rather than any network-derived proxy. Ties in ring distance are
#' broken at random.
#'
#' \strong{Pass selection.} A pass is valid only if the resulting graph is
#' connected and has zero residual degree error. Among valid passes, the one
#' with the highest average clustering coefficient is returned.
#'
#' \strong{Empirical fallback.} If no valid pass is found, or if the best
#' lattice clustering coefficient is lower than that of the original network,
#' the empirical adjacency (or weighted matrix, if \code{weighted = TRUE}) is
#' returned with a warning.
#'
#' @references
#' \strong{Logic for weight assignments} \cr
#' Muldoon, S. F., Bridgeford, E. W., & Bassett, D. S. (2016).
#' Small-world propensity and weighted brain connectivity.
#' \emph{Scientific Reports}, \emph{6}, 22057.
#'
#' @examples
#' # Get network
#' network <- network_estimation(basic_smallworld)
#'
#' # Construct binary ring lattice
#' L <- proxswap_lattice(network)
#'
#' # Retrieve the attached clustering coefficient
#' attr(L, "CC")
#'
#' # Degree sequences should match exactly
#' cbind(target = colSums(network != 0), achieved = colSums(L))
#'
#' # Construct weighted ring lattice
#' L_weighted <- proxswap_lattice(network, weighted = TRUE)
#'
#' # Retrieve the attached clustering coefficient
#' attr(L_weighted, "CC")
#'
#' @author Alexander P. Christensen <alexpaulchristensen@gmail.com>
#'
#' @export
#'
# Proximity-swap lattice construction ----
# Updated 26.03.2026
proxswap_lattice <- function(network, weighted = FALSE, shuffles = 100)
{

  # Automatically construct adjacency
  A <- network != 0

  # Check for weighted
  if(!weighted){
    network <- A
  }

  # Initialize nodes, degree, and connectedness flag
  nodes  <- dim(network)[2]
  degree <- colSums(A)
  check_connectedness <- all(degree != 0)

  # Set up distance matrix
  node_sequence <- seq_len(nodes)
  distance_matrix <- abs(outer(node_sequence, node_sequence, "-"))
  distance_matrix <- pmin(distance_matrix, nodes - distance_matrix)

  # Pre-compute unique node pairs at each ring distance
  pairs <- build_pairs(distance_matrix)

  # Compute empirical clustering coefficient for fallback check
  empirical_CC <- igraph::transitivity(convert2igraph(network), type = "average")

  # Initialize best result trackers
  best_swap <- best_ring <- NULL
  best_CC   <- -Inf

  # Initialize to an empty graph
  ring <- matrix(FALSE, nrow = nodes, ncol = nodes)

  # Run each shuffle pass
  for(k in seq_len(shuffles)){

    # Store swap order (returns proper degree sequence)
    swap_order <- shuffle(node_sequence)

    # Run proximity construction with shuffled degrees
    result <- proximity_pass(
      nodes = nodes, ring = ring, budget = degree[swap_order], pairs = pairs
    )

    # Perform swapping to
    result <- swapping_pass(
      nodes = nodes, ring = result$ring, budget = result$budget,
      total_budget = result$total_budget
    )

    # Check for deficit before expensive connectedness check
    if(result$remaining > 0){
      next
    }

    # Convert ring to {igraph}
    iring <- convert2igraph(result$ring)

    # Determine whether connectedness (and then check it)
    if(check_connectedness && (!igraph::is_connected(iring))){
      next
    }

    # Check for weighted
    if(weighted){

      # Get weights
      result$ring <- assign_weights(network, result$ring, distance_matrix)

      # Update ring to {igraph}
      iring <- convert2igraph(result$ring)

    }

    # With success, compute clustering coefficient for this pass
    pass_CC <- igraph::transitivity(iring, type = "average")

    # Update best pass
    if(pass_CC > best_CC){
      best_CC <- pass_CC; best_ring <- result$ring; best_swap <- swap_order
    }

  }

  # Determine whether solution was reached
  if(is.infinite(best_CC)){

    # Send warning
    warning(paste0(
      "The lattice solution did not converge. The empirical graph was returned.\n\n",
      "Try increasing `shuffles` to find a suitable solution"
    ))

    # Set empirical flag
    empirical_flag <- TRUE

    # Check for empirical and weighted
    ring <- swiftelse(weighted, network, A)

  }else{

    # Check whether empirical beats the best lattice
    empirical_flag <- empirical_CC > best_CC

    # Warn on empirical fallback
    if(empirical_flag){
      warning(paste0(
        "The lattice solution did not produce a better CC (", round(best_CC, 3),
        ") than the empirical (", round(empirical_CC, 3), ").\n",
        "Falling back to empirical solution..."
      ))
    }

    # Select between lattice and empirical
    original_order <- order(best_swap)

    # Check for empirical and weighted
    ring <- swiftelse(
      empirical_flag,
      swiftelse(weighted, network, A),
      best_ring[original_order, original_order]
    )

  }

  # Ensure named matrix
  dimnames(ring) <- dimnames(network)

  # Attach clustering coefficient
  attr(ring, "CC") <- swiftelse(empirical_flag, empirical_CC, best_CC)

  # Return ring
  return(ring)

}

#' @noRd
# Pre-compute unique node pairs per ring distance ----
# Updated 26.03.2026
build_pairs <- function(distance_matrix)
{

  # Loop over distances
  return(
    lapply(seq_len(max(distance_matrix)), function(i){

      # Obtain pairs
      pairs <- which(distance_matrix == i, arr.ind = TRUE)

      # Return pairs
      return(pairs[pairs[,"row"] < pairs[,"col"],])

    }
    )
  )

}

#' @noRd
# Proximity construction ----
# Updated 24.03.2026
proximity_pass <- function(nodes, ring, budget, pairs)
{

  # # Remaining edges to assign per node
  # total_budget <- sum(budget)
  #
  # # Add edges in increasing ring distance
  # for(i in distance_sequence){
  #
  #   # Retrieve precomputed node-index vectors for this distance band
  #   rows <- pairs[[i]][,"row"]
  #   columns <- pairs[[i]][,"col"]
  #
  #   # Determine eligible
  #   eligible <- (budget[rows] > 0) & (budget[columns] > 0) & (!ring[cbind(rows, columns)])
  #
  #   # Skip this distance band if nothing is eligible
  #   if(!any(eligible)){
  #     next
  #   }
  #
  #   # Eligible candidate indices
  #   row <- rows[eligible]
  #   col <- columns[eligible]
  #
  #   # Get row length
  #   row_length  <- length(row)
  #
  #   # Sort: highest min-budget pair first so high-need pairs get short edges
  #   if(row_length > 1){
  #     target <- order(pmin(budget[row], budget[col]), decreasing = TRUE)
  #     row  <- row[target]
  #     col  <- col[target]
  #   }
  #
  #   # Assign edges one at a time; re-check both budgets before each assignment
  #   for(l in seq_len(row_length)){
  #
  #     # Set targets
  #     target_row <- row[l]
  #     target_col <- col[l]
  #
  #     # Check budget (can change with loop)
  #     if((budget[target_row] > 0) && (budget[target_col] > 0)){
  #
  #       # Add to ring
  #       ring[target_row, target_col] <- ring[target_col, target_row] <- TRUE
  #
  #       # Update budget
  #       budget[target_row] <- budget[target_row] - 1
  #       budget[target_col] <- budget[target_col] - 1
  #       total_budget <- total_budget - 2
  #
  #     }
  #   }
  #
  #   # Early exit once all budgets are satisfied
  #   if(total_budget < 1){
  #     break
  #   }
  #
  # }
  #
  # # Return ring and total absolute degree error
  # return(list(ring = ring, budget = budget, total_budget = total_budget))

  # Flatten the pairs list-of-matrices into three parallel vectors
  pair_rows <- ulapply(pairs, function(p) p[, "row"])
  pair_cols  <- ulapply(pairs, function(p) p[, "col"])
  pair_counts <- lengths(lapply(pairs, function(p) p[, "row"]))

  # Coerce logical ring matrix to integer storage before any .Call
  storage.mode(ring) <- "integer"

  # Collect result
  result <- .Call(
    "proximity_pass_c",
    as.integer(nodes),
    as.integer(ring),
    as.integer(budget),
    as.integer(pair_rows),
    as.integer(pair_cols),
    as.integer(pair_counts),
    PACKAGE = "L0ggm"
  )

  # Set up ring matrix
  result$ring <- matrix(result$ring, nrow = nodes, ncol = nodes)

  # Return result
  return(result)

}

#' @noRd
# Swap to amend lattice ----
# Updated 24.03.2026
swapping_pass <- function(nodes, ring, budget, total_budget)
{

  # # Set maximum iterations
  # for(iter in seq_len(nodes * nodes * 2)){
  #
  #   # Stop when all budgets are satisfied
  #   if(total_budget == 0){
  #     break
  #   }
  #
  #   # Work on the node with the largest remaining deficit
  #   i <- which.max(budget)
  #   if(budget[i] == 0){
  #     break
  #   }
  #
  #   # Interleave clockwise and counter-clockwise ring positions by distance
  #   distance <- interleave(i, nodes, distance_sequence)
  #
  #   # Set avaiable ring (diagonal will be TRUE but will never be in 'distance')
  #   available_ring <- !ring
  #
  #   # Try direct connection to nearest node that has budget and no existing edge
  #   direct <- distance[(budget[distance] > 0) & (available_ring[i, distance])]
  #
  #   # Check connection
  #   if(length(direct) > 0){
  #
  #     # Set index
  #     target <- direct[1]
  #
  #     # Update ring
  #     ring[i, target] <- ring[target, i] <- TRUE
  #
  #     # Update budget
  #     budget[i] <- budget[i] - 1
  #     budget[target] <- budget[target] - 1
  #     total_budget <- total_budget - 2
  #
  #     # Continue
  #     next
  #
  #   }
  #
  #   # Initialize swap
  #   swapped <- FALSE
  #
  #   # Check nearby unconnected nodes
  #   candidates <- distance[available_ring[i, distance]]
  #
  #   # Loop over candidates
  #   for(j in candidates){
  #
  #     # Find a neighbor of the neighbor that is not connected to current node
  #     neighborhood <- which(ring[j,] & node_sequence != i & available_ring[i,])
  #
  #     # Check if any are available
  #     if(length(neighborhood) > 0){
  #
  #       # Select first available neighbor
  #       k <- neighborhood[1]
  #
  #       # Perform swap
  #       ring[j, k] <- ring[k, j] <- FALSE
  #       ring[i, j] <- ring[j, i] <- TRUE
  #
  #       # Update budgets
  #       budget[i] <- budget[i] - 1
  #       budget[k] <- budget[k] + 1
  #
  #       # Total budget is unchanged with swap
  #       swapped <- TRUE
  #
  #       # Break out of loop
  #       break
  #
  #     }
  #
  #   }
  #
  #   # If completely struck, just break
  #   if(!swapped){
  #     break
  #   }
  #
  # }
  #
  # # Return ring and remaining
  # return(list(ring = ring, remaining = total_budget))

  # Coerce output ring back to integer
  storage.mode(ring) <- "integer"

  # Collect result
  result <- .Call(
    "swapping_pass_c",
    as.integer(nodes),
    as.integer(ring),
    as.integer(budget),
    as.integer(total_budget),
    PACKAGE = "L0ggm"
  )

  # Set up ring matrix
  result$ring <- matrix(result$ring, nrow = nodes, ncol = nodes)

  # Return result
  return(result)

}

# # Interleaving function ----
# # Updated 24.03.2026
# interleave <- function(index, nodes, distance_sequence)
# {
#
#   return(
#     unique(
#       c(
#         rbind(
#           (index - 1 + distance_sequence) %% nodes + 1, # clockwise
#           (index - 1 - distance_sequence + nodes * distance_sequence) %% nodes + 1
#           # counterclockwise
#         )
#       )
#     )
#   )
#
# }

#' @noRd
# Assign weights based on distance ----
# Updated 26.03.2026
assign_weights <- function(network, A, distance_matrix)
{

  # Obtain weights
  lower_triangle <- lower.tri(network)
  network_nonzero <- network[lower_triangle] != 0
  weights <- network[lower_triangle][network_nonzero][
    order(abs(network)[lower_triangle][network_nonzero], decreasing = TRUE)
  ]

  # Set weights
  A_nonzero <- A[lower_triangle] != 0

  # Set weights order
  # Follows: Muldoon, Bridgeford, & Bassett's (2016) implementation
  weight_order <- rank(distance_matrix[lower_triangle][A_nonzero], ties.method = "random")
  A[lower_triangle][A_nonzero] <- weights[weight_order]
  A[!lower_triangle] <- 0
  A <- A + t(A) # make symmetric

  # Return weighted lattice
  return(A)

}

#' @title Automatic correlations
#'
#' @description Automatically computes the proper correlations between
#' continuous and categorical variables. \code{NA} values are not treated
#' as categories
#'
#' @param data Matrix or data frame.
#' Should consist only of variables to be used in the analysis
#'
#' @param corr Character (length = 1).
#' The standard correlation method to be used.
#' Defaults to \code{"pearson"}.
#' Using \code{"pearson"} will compute polychoric, tetrachoric, polyserial,
#' and biserial correlations for categorical and categorical/continuous correlations
#' by default. To obtain \code{"pearson"} correlations regardless, use \code{\link{cor}}.
#' Other options of \code{"kendall"} and \code{"spearman"} are provided for
#' completeness and use \code{\link{cor}}
#'
#' @param ordinal_categories Numeric (length = 1).
#' \emph{Up to} the number of categories \emph{before} a variable is considered continuous.
#' Defaults to \code{7} categories before \code{8} is considered continuous
#'
#' @param forcePD Boolean (length = 1).
#' Whether positive definite matrix should be enforced.
#' Defaults to \code{TRUE}
#'
#' @param na_data Character (length = 1).
#' How should missing data be handled?
#' Defaults to \code{"pairwise"}.
#' Available options:
#'
#' \itemize{
#'
#' \item \code{"pairwise"} --- Computes correlation for all available
#' cases between two variables
#'
#' \item \code{"listwise"} --- Computes correlation for all complete
#' cases in the dataset
#'
#' }
#'
#' @param empty_method Character (length = 1).
#' Method for empty cell correction in \code{\link[L0ggm]{polychoric_matrix}}.
#' Defaults to \code{"none"}
#' Available options:
#'
#' \itemize{
#'
#' \item \code{"none"} --- Adds no value (\code{empty_value = "none"})
#' to the empirical joint frequency table between two variables
#'
#' \item \code{"zero"} --- Adds \code{empty_value} to the cells with
#' zero in the joint frequency table between two variables
#'
#' \item \code{"all"} --- Adds \code{empty_value} to all
#' in the joint frequency table between two variables
#'
#' }
#'
#' @param empty_value Character (length = 1).
#' Value to add to the joint frequency table cells in \code{\link[L0ggm]{polychoric_matrix}}.
#' Defaults to \code{"none"}.
#' Accepts numeric values between 0 and 1 or specific methods:
#'
#' \itemize{
#'
#' \item \code{"none"} --- Adds no value (\code{0}) to the empirical
#' joint  frequency table between two variables
#'
#' \item \code{"point_five"} --- Adds \code{0.5} to the cells
#' defined by \code{empty_method}
#'
#' \item \code{"one_over"} --- Adds \code{1 / n} where \code{n} equals the
#' number of cells based on \code{empty_method}. For
#' \code{empty_method = "zero"}, \code{n} equals the number of zero cells
#'
#' }
#'
#' @param forceReturn Boolean (length = 1).
#' Whether correlation matrix should be forced to return.
#' Defaults to \code{FALSE}.
#' Set to \code{TRUE} to receive the correlation matrix "as is"
#'
#' @param verbose Boolean (length = 1).
#' Whether messages should be printed.
#' Defaults to \code{FALSE}
#'
#' @param ...
#' Not actually used but makes it easier for general functionality
#' in the package
#'
#' @return A symmetric numeric matrix of dimension \eqn{p \times p}, where
#' \eqn{p} is the number of variables in \code{data}, with values in
#' \eqn{[-1, 1]} and ones on the diagonal. The correlation method used for
#' each pair of variables depends on their types when \code{corr = "pearson"}
#' (the default): polychoric for ordinal-ordinal pairs, polyserial for
#' ordinal-continuous pairs, and Pearson's for continuous-continuous pairs,
#' where \emph{ordinal} means having at most \code{ordinal_categories} unique
#' values. When \code{corr} is \code{"kendall"} or \code{"spearman"},
#' \code{\link[stats]{cor}} is used for all pairs regardless of variable
#' type. Row and column names are inherited from \code{data}. When
#' \code{forcePD = TRUE} (default) and the resulting matrix is not positive
#' definite, the nearest positive definite matrix is returned via
#' \code{\link[Matrix]{nearPD}}.
#'
#' @author Alexander P. Christensen <alexpaulchristensen@gmail.com>
#'
#' @examples
#' # Obtain correlations
#' R <- auto_correlate(basic_smallworld)
#'
#' @export
#'
# Automatic correlations ----
# Updated 07.03.2026
auto_correlate <- function(
    data, # Matrix or data frame
    corr = c("kendall", "pearson", "spearman"), # allow changes to standard correlations
    ordinal_categories = 7, # consider ordinal up to 7 categories
    forcePD = TRUE, # ensure result is positive definite
    na_data = c("pairwise", "listwise"), # use available or complete values
    empty_method = c("none", "zero", "all"), # zero frequencies in categorical correlations
    empty_value = c("none", "point_five", "one_over"), # value to use in zero cells
    forceReturn = FALSE, # return even if bad correlation matrix
    verbose = FALSE, # don't print messages
    ... # not actually used
)
{

  # Argument errors (return data in case of tibble)
  data <- auto_correlate_errors(data, ordinal_categories, forcePD, forceReturn, verbose, ...)

  # Check for missing arguments (argument, default, function)
  corr <- set_default(corr, "pearson", auto_correlate)
  na_data <- set_default(na_data, "pairwise", auto_correlate)
  empty_method <- set_default(empty_method, "none", auto_correlate)
  empty_value <- set_default(empty_value, "none", auto_correlate)

  # Ensure matrix
  data <- as.matrix(data)

  # Ensure variable names
  data <- ensure_dimension_names(data)

  # Get variable names
  variable_names <- dimnames(data)[[2]]

  # Set dimensions
  dimensions <- dim(data)

  # Assumes some preprocessing has taken place
  # to only select for appropriate variables

  # Determine whether categorical correlations are necessary
  if(corr != "pearson"){

    # Obtain correlation matrix
    correlation_matrix <- cor(
      x = data, use = swiftelse(
        na_data == "pairwise",
        "pairwise.complete.obs",
        "complete.obs"
      ), method = corr
    )

  }else{ # Proceed with determination of categorical correlations

    # Obtain the number of categories for each variables
    categories <- data_categories(data)

    # Determine categorical variables
    categorical_variables <- categories <= ordinal_categories

    # Determine number of categorical variables
    categorical_number <- sum(categorical_variables)

    # Determine whether there are any categorical variables
    if(categorical_number != 0){

      # Determine continuous variables
      continuous_variables <- !categorical_variables

      # Set up correlation matrix
      correlation_matrix <- matrix(
        nrow = dimensions[2], ncol = dimensions[2],
        dimnames = list(variable_names, variable_names)
      )

      # Determine whether there are more than one categorical variables
      if(categorical_number > 1){

        # Compute categorical correlations (only correlation for categorical data)
        # Add correlations to correlation matrix
        correlation_matrix[
          categorical_variables, categorical_variables # ensure proper indexing
        ] <- polychoric_matrix(
          data = data[,categorical_variables], na_data = na_data,
          empty_method = empty_method, empty_value = empty_value,
          needs_usable = FALSE # skip usable data check
        )

      }

      # Determine whether there are more than one continuous variables
      if(sum(continuous_variables) > 1){

        # Compute continuous correlations (only correlation for continuous data)
        # Add correlations to correlation matrix
        correlation_matrix[
          continuous_variables, continuous_variables # ensure proper indexing
        ] <- cor(
          x = data[,continuous_variables],
          use = swiftelse(
            na_data == "pairwise",
            "pairwise.complete.obs",
            "complete.obs"
          ), method = corr
        )

      }

      # Determine whether there are mixed variables
      if(categorical_number != dimensions[2]){ # Check for mixed variables

        # Obtain continuous data (keep as matrix)
        continuous_data <- data[,continuous_variables, drop = FALSE]

        # Loop over categorical indices
        for(i in which(categorical_variables)){

          # Fill matrix
          correlation_matrix[continuous_variables, i] <-
            correlation_matrix[i, continuous_variables] <-
            polyserial_vector( # computes polyserial vector
              categorical_variable = data[,i], # drops to vector
              continuous_variables = continuous_data,
              na_data = na_data
            )

        }

      }

    }else{

      # Compute Pearson's correlations
      correlation_matrix <- cor(
        x = data, use = swiftelse(
          na_data == "pairwise",
          "pairwise.complete.obs",
          "complete.obs"
        ), method = corr
      )

    }

  }

  # Set diagonal as one
  diag(correlation_matrix) <- 1

  # Try to test for positive definite
  PD <- try(is_positive_definite(correlation_matrix), silent = TRUE)
  PD_error <- is(PD, "try-error")

  # Check for error
  if(PD_error && forceReturn){
    return(correlation_matrix)
  }

  # Determine whether matrix is positive definite
  if((PD_error || !PD) && forcePD){

    # Send warning to user (if `verbose`)
    if(verbose){
      warning(
        "Correlation matrix is not positive definite. Finding nearest positive definite matrix using `Matrix::nearPD`",
        call. = FALSE
      )
    }

    # Regardless, make matrix positive definite
    correlation_matrix <- as.matrix(
      Matrix::nearPD(
        correlation_matrix, corr = TRUE,
        ensureSymmetry = TRUE, keepDiag = TRUE
      )$mat
    )

  }

  # Return correlation matrix
  return(correlation_matrix)

}

#' @noRd
# Errors ----
# Updated 29.11.2025
auto_correlate_errors <- function(data, ordinal_categories, forcePD, forceReturn, verbose, ...)
{

  # 'data' errors
  object_error(data, c("matrix", "data.frame", "tibble"), "auto_correlate")

  # Check for tibble
  if(get_object_type(data) == "tibble"){
    data <- as.data.frame(data)
  }

  # 'ordinal_categories' errors
  length_error(ordinal_categories, 1, "auto_correlate")
  typeof_error(ordinal_categories, "numeric", "auto_correlate")
  range_error(ordinal_categories, c(2, 11), "auto_correlate")

  # 'forcePD' errors
  length_error(forcePD, 1, "auto_correlate")
  typeof_error(forcePD, "logical", "auto_correlate")

  # 'verbose' errors
  length_error(verbose, 1, "auto_correlate")
  typeof_error(verbose, "logical", "auto_correlate")

  # 'forceReturn' errors
  length_error(forceReturn, 1, "auto_correlate")
  typeof_error(forceReturn, "logical", "auto_correlate")

  # Check for usable data
  if(needs_usable(list(...))){
    data <- usable_data(data, verbose)
  }

  # Return usable data (in case of tibble)
  return(data)

}

# Compute polyserial correlation ----
#' For a single categorical variable, compute correlations
#' with \emph{n} continuous variables
#'
#' Uses two-step approximation from {polycor}'s \code{polyserial}
#'
#' @noRd
# Updated 16.03.2024
polyserial_vector <- function(
    categorical_variable, continuous_variables,
    na_data = c("pairwise", "listwise")
)
{

  # Determine cases based on `na_data` argument
  if(na_data == "pairwise"){

    # Pairwise cases
    categorical_cases <- colSums(
      !is.na(categorical_variable) *
      !is.na(continuous_variables)
    )

  }else if(na_data == "listwise"){

    # Complete cases
    complete_cases <- complete.cases(cbind(categorical_variable, continuous_variables))

    # Repeat cases for number of continuous variables
    categorical_cases <- rep(sum(complete_cases), dim(continuous_variables)[2])

  }

  # Compute correlation
  return(
    sqrt((categorical_cases - 1) / categorical_cases) *
      sd(categorical_variable, na.rm = TRUE) *
      # Correlations with scaled continuous variables
      cor(
        categorical_variable, scale(continuous_variables),
        use = swiftelse(
          na_data == "pairwise",
          "pairwise.complete.obs",
          "complete.obs"
        )
      ) /
      # Compute sum of thresholds
      sum(dnorm(obtain_thresholds(categorical_variable)), na.rm = TRUE)
  )

}

#' Categorize Continuous Data
#'
#' Categorizes continuous data based on Garrido, Abad and Ponsoda (2011; see references).
#' Categorical data with 2 to 6 categories can include skew between -2 to 2 in
#' increments of 0.05
#'
#' @param data Numeric (length = n).
#' A vector of continuous data with \emph{n} values.
#' For matrices, use \code{apply}
#'
#' @param categories Numeric (length = 1).
#' Number of categories to create.
#' Between 2 and 6 categories can be used with skew
#'
#' @param skew_value Numeric (length = 1).
#' Value of skew.
#' Ranges between -2 to 2 in increments of 0.05.
#' Skews not in this sequence will be converted to
#' the nearest value in this sequence.
#' Defaults to \code{0} or no skew
#'
#' @return An integer vector of length \code{n} with values from \code{1} to
#' \code{categories}, giving the category assignment for each observation.
#' Category \code{1} corresponds to the lowest values of \code{data} and
#' category \code{categories} to the highest. When \code{categories > 6},
#' \code{\link[base]{cut}} is used and skew is not applied.
#'
#' @examples
#' # Dichotomous data (no skew)
#' dichotomous <- categorize(
#'   data = rnorm(1000),
#'   categories = 2
#' )
#'
#' # Dichotomous data (with positive skew)
#' dichotomous_skew <- categorize(
#'   data = rnorm(1000),
#'   categories = 2,
#'   skew_value = 1.25
#' )
#'
#' # 5-point Likert scale (no skew)
#' five_likert <- categorize(
#'   data = rnorm(1000),
#'   categories = 5
#' )
#'
#' # 5-point Likert scale (negative skew)
#' five_likert <- categorize(
#'   data = rnorm(1000),
#'   categories = 5,
#'   skew_value = -0.45
#' )
#'
#' @author
#' Maria Dolores Nieto Canaveras <mnietoca@nebrija.es>,
#' Luis Eduardo Garrido <luisgarrido@pucmm.edu>,
#' Hudson Golino <hfg9s@virginia.edu>,
#' Alexander P. Christensen <alexpaulchristensen@gmail.com>
#'
#' @references
#' Garrido, L. E., Abad, F. J., & Ponsoda, V. (2011). \cr
#' Performance of Velicer’s minimum average partial factor retention method with categorical variables. \cr
#' \emph{Educational and Psychological Measurement}, \emph{71}(3), 551-570.
#'
#' Golino, H., Shi, D., Christensen, A. P., Garrido, L. E., Nieto, M. D., Sadana, R., ... & Martinez-Molina, A. (2020).
#' Investigating the performance of exploratory graph analysis and traditional techniques to identify the number of latent factors: A simulation and tutorial.
#' \emph{Psychological Methods}, \emph{25}(3), 292-320.
#'
#' @export
#'
# Categorization function
# Updated 23.03.2023
categorize <- function(data, categories, skew_value = 0)
{

  # Possible skew values
  possible_skews <- seq(-2, 2, 0.05)

  # Check if skew is in possible values
  if(!skew_value %in% possible_skews){

    # Round to hundredths digit
    skew <- round(skew_value, 2)

    # Get differences
    skew_difference <- abs(skew - possible_skews)

    # Get skew
    skew_value <- possible_skews[which.min(skew_difference)]

  }

  # Load skew tables
  skew_tables <- get(data(
    "skew_tables",
    package = "L0ggm",
    envir = environment()
  ))

  # Check for categories greater than 6
  if(categories > 6){

    # Make categorical using `cut` (skew will be zero)
    skewed_data <- as.numeric(cut(data, breaks = categories))

  }else{

    # Switch categories to character
    categories <- switch(
      as.character(categories),
      "2" = "two",
      "3" = "three",
      "4" = "four",
      "5" = "five",
      "6" = "six"
    )

    # Obtain skew table
    skew_table <- skew_tables[[categories]]

    # Sometimes R will treat numeric zeros
    # as having negatives (seriously, why?)
    # This fix prevents R from making
    # 0 = "-0.00" and ensures 0 = "0.00"
    if(skew_value == 0){
      skew_value <- abs(skew_value)
    }

    # Obtain skew values
    skew_values <- skew_table[,formatC(
      skew_value, digits = 2,
      format = "f", flag = "0"
    )]

    # Add skew to data (see `utils-latentFactoR` for function)
    skewed_data <- skew_single_variable(
      data = data, skew_values = skew_values
    )


  }

  # Return categorized data
  return(skewed_data)

}

#' @title Traditional Fit Metrics for Networks
#'
#' @description Computes several traditional fit metrics for networks including
#'
#' \itemize{
#'
#' \item chi-square (\eqn{\chi^2})
#'
#' \item root mean square error of approximation (RMSEA) with confidence intervals
#'
#' \item confirmatory fit index (CFI)
#'
#' \item Tucker-Lewis index (TLI)
#'
#' \item standardized root mean residual (SRMR)
#'
#' \item log-likelihood
#'
#' \item Akaike's information criterion (AIC)
#'
#' \item Bayesian information criterion (BIC)
#'
#' }
#'
#' @param network Matrix or data frame.
#' A p by p square network matrix
#'
#' @param n Numeric (length = 1).
#' Sample size
#'
#' @param S Matrix or data frame.
#' A p by p square zero-order correlation matrix corresponding
#' with the input \code{network}
#'
#' @param ci Numeric (length = 1).
#' Confidence interval for RMSEA.
#' Defaults to \code{0.95}
#'
#' @return A named numeric vector of traditional and likelihood-based fit
#' indices. The vector always contains the following elements:
#'
#' \describe{
#'
#' \item{\code{chisq}}{Chi-square statistic
#' (\eqn{\chi^2 = n \cdot F_{ML}}), where \eqn{F_{ML}} is the maximum
#' likelihood discrepancy function between the model-implied and empirical
#' correlation matrices}
#'
#' \item{\code{df}}{Degrees of freedom: total number of unique off-diagonal
#' correlations minus the number of non-zero edges in \code{network}}
#'
#' \item{\code{chisq.p.value}}{p-value for the chi-square test of exact fit
#' (H0: model-implied covariance equals the population covariance)}
#'
#' \item{\code{RMSEA}}{Root mean square error of approximation. Values
#' \eqn{\leq} 0.05 indicate close fit; values \eqn{\leq} 0.08 indicate
#' acceptable fit}
#'
#' \item{\code{RMSEA.XX.lower}, \code{RMSEA.XX.upper}}{Lower and upper
#' bounds of the \code{ci}-level confidence interval for RMSEA, where
#' \code{XX} is the integer percentage (e.g., \code{RMSEA.95.lower} and
#' \code{RMSEA.95.upper} for a 95\% CI)}
#'
#' \item{\code{RMSEA.p.value}}{p-value for the one-sided test of close fit
#' (H0: RMSEA \eqn{\leq} 0.05)}
#'
#' \item{\code{CFI}}{Comparative fit index, comparing the target model to
#' an independence (null) baseline. Values \eqn{\geq} 0.95 indicate
#' acceptable fit}
#'
#' \item{\code{TLI}}{Tucker-Lewis index (non-normed fit index). Values
#' \eqn{\geq} 0.95 indicate acceptable fit; can fall outside \eqn{[0, 1]}
#' for severely misspecified models}
#'
#' \item{\code{SRMR}}{Standardized root mean residual: the root mean squared
#' difference between the model-implied and observed correlation matrices.
#' Values \eqn{\leq} 0.08 indicate acceptable fit}
#'
#' \item{\code{logLik}}{Gaussian log-likelihood of the model-implied
#' correlation matrix, assuming zero mean structure (means are not
#' estimated)}
#'
#' \item{\code{AIC}}{Akaike's information criterion:
#' \eqn{-2 \cdot \log L + 2 \cdot E}, where \eqn{E} is the number of
#' non-zero edges in \code{network}}
#'
#' \item{\code{BIC}}{Bayesian information criterion:
#' \eqn{-2 \cdot \log L + E \cdot \log(n)}}
#'
#' }
#'
#' @author Alexander P. Christensen <alexpaulchristensen@gmail.com>
#'
#' @examples
#' # Obtain correlation matrix
#' S <- auto_correlate(basic_smallworld)
#'
#' # Obtain Weibull network
#' weibull_network <- network_estimation(data = basic_smallworld, LLA = TRUE)
#'
#' # Obtain fit (expects continuous variables!)
#' network_fit(network = weibull_network, n = nrow(basic_smallworld), S = S)
#' # Scaled metrics are not yet available for
#' # dichotomous or polytomous data!
#'
#' @references
#' Epskamp, S., Rhemtulla, M., & Borsboom, D. (2017).
#' Generalized network psychometrics: Combining network and latent variable models.
#' \emph{Psychometrika}, \emph{82}(4), 904–927.
#'
#' @export
#'
# Compute network fit statistics ----
# Updated 19.11.2025
network_fit <- function(network, n, S, ci = 0.95)
{

  # Obtain lower triangle indices
  lower_triangle <- lower.tri(network)

  # Obtain model-implied zero-order correlation matrix
  R <- pcor2cor(network)

  # Compute number of parameters
  p <- dim(network)[2]
  zero_parameters <- p * (p - 1) / 2
  model_parameters <- sum(network[lower_triangle] != 0)

  # Compute baseline
  baseline <- diag(1, nrow = p, ncol = p)
  baseline_ML <- log(det(baseline)) + sum(diag(S %*% solve(baseline))) - log(det(S)) - p
  baseline_chi_square <- n * baseline_ML
  baseline_tli <- baseline_chi_square / zero_parameters

  # Compute traditional SEM measures
  loglik_ML <- log(det(R)) + sum(diag(S %*% solve(R))) - log(det(S)) - p
  chi_square <- n * loglik_ML
  df <- zero_parameters - model_parameters
  chi_max <- max(chi_square - df, 0)
  nDF <- n * df
  rmsea_null <- nDF * 0.0025 # 0.05^2

  # log-likelihood
  loglik <- -(n / 2) * (p * log(2 * pi) + log(det(R)) + sum(diag(S %*% solve(R))))
  # Assumes no mean structure (or that all means are equal to zero)

  # Obtain RMSEA confidence intervals
  rmsea_cis <- rmsea_ci(chi_square, df, n, nDF, ci)

  # Get fit indices
  fit_indices <- c(
    # Traditional fit measures
    chisq = chi_square, df = df, chisq.p.value = 1 - pchisq(chi_square, df = df),
    RMSEA = sqrt(chi_max / nDF),
    rmsea_cis,
    RMSEA.p.value = 1 - pchisq(chi_max, df = df, ncp = rmsea_null),
    CFI = 1 - (chi_max / max(baseline_chi_square - zero_parameters, 0)),
    TLI = (baseline_tli - (chi_square / df)) / (baseline_tli - 1),
    SRMR = sqrt(mean((R[lower_triangle] - S[lower_triangle])^2)),
    # Gaussian log-likelihood measures
    logLik = loglik,
    AIC = -2 * loglik + 2 * model_parameters,
    BIC = -2 * loglik + model_parameters * log(n)
  )

  # Rename confidence intervals
  names(fit_indices)[names(fit_indices) %in% c("lower", "upper")] <-
    paste("RMSEA", format_integer(ci * 100, 1), c("lower", "upper"), sep = ".")

  # Return log-likelihood
  return(fit_indices)

}

#' @noRd
# RMSEA Confidence Intervals ----
# Follows {lavaan} version 0.6.19
# Updated 01.11.2024
rmsea_ci <- function(chi_square, df, n, nDF, ci)
{

  # Set up CI
  lower_ci <- 1 - (1 - ci) / 2
  upper_ci <- 1 - lower_ci

  # Internal function for finding RMSEA confidence intervals
  # (same as {lavann} version 0.6.19)
  find_lambda <- function(lambda, ci){
    pchisq(chi_square, df = df, ncp = lambda) - ci
  }

  # Find lower bound
  if(df < 1 || find_lambda(0, lower_ci) < 0){
    rmsea_lower <- 0
  }else{

    # Try uniroot
    lambda <- try(
      uniroot(f = find_lambda, lower = 0, upper = chi_square, ci = lower_ci)$root,
      silent = TRUE
    )

    # Determine if there was an error
    rmsea_lower <- swiftelse(is(lambda, "try-error"), NA, sqrt(lambda / nDF))

  }

  # Find upper bound
  N_RMSEA <- max(n, chi_square * 4)
  if(df < 1 || find_lambda(N_RMSEA, upper_ci) > 0 || find_lambda(0, upper_ci) < 0){
    rmsea_upper <- 0
  }else{

    # Try uniroot
    lambda <- try(
      uniroot(f = find_lambda, lower = 0, upper = N_RMSEA, ci = upper_ci)$root,
      silent = TRUE
    )

    # Determine if there was an error
    rmsea_upper <- swiftelse(is(lambda, "try-error"), NA, sqrt(lambda / nDF))

  }

  # Return confidence interval
  return(c(lower = rmsea_lower, upper = rmsea_upper))

}

#' SUR Model Coefficients and Residuals for Weibull Parameter Prediction
#'
#' A named list encoding the Seemingly Unrelated Regression (SUR) model fitted
#' to Weibull shape and scale parameters derived from 194
#' empirical networks. The object is consumed internally by \code{\link{weibull_parameters}}
#' to generate data-driven Weibull parameter estimates given a network's
#' number of nodes and sample size.
#'
#' @name weibull_weights
#'
#' @docType data
#'
#' @usage data(weibull_weights)
#'
#' @format A named list with two elements, \code{shape} and \code{scale},
#' each containing:
#' \describe{
#'   \item{\code{coefficients}}{A named numeric vector of regression
#'   coefficients from the SUR model. The equations are asymmetric: the shape
#'   equation includes an intercept and two predictors (\code{snr}, \code{rlp});
#'   the scale equation includes an intercept and three predictors
#'   (\code{snr}, \code{rlp}, \code{scaling}).
#'   Predictors are defined as follows:
#'   \describe{
#'     \item{\code{snr}}{Signal-to-noise ratio of the absolute partial
#'     correlations, computed as the mean divided by the standard deviation
#'     of the absolute edge weights. Used in both equations.}
#'     \item{\code{rlp}}{Reciprocal log of node count, \eqn{1 / \log(p)},
#'     where \eqn{p} is the number of nodes. Used in both equations.}
#'     \item{\code{scaling}}{Standard error of partial correlations,
#'     \eqn{\sqrt{1 / (n - p)}}, where \eqn{n} is the sample size and
#'     \eqn{p} is the number of nodes. Used in the scale equation only.}
#'   }}
#'   \item{\code{residuals}}{A numeric vector of residuals from the fitted SUR
#'   equation, used by \code{\link{weibull_parameters}} to introduce
#'   empirically grounded variability when \code{bootstrap = TRUE}.}
#' }
#'
#' @details
#' Absolute partial correlations from 222 deduplicated empirical networks
#' (Huth et al., 2025) were fitted to Beta, Gamma, log-normal, and Weibull
#' distributions via maximum likelihood. Weibull provided the best fit most
#' consistently: it outperformed each alternative by more than 2 log-likelihood
#' units far more often than the reverse (vs. Beta: 56--0; vs. Gamma: 15--2;
#' vs. log-normal: 155--13; vs. Exponential: 54--0).
#'
#' The resulting Weibull shape and scale parameters were then jointly modelled
#' as a function of network descriptors using Seemingly Unrelated Regression
#' (\code{systemfit}), which accounts for correlated residuals between the
#' shape and scale equations across networks (residual correlation = 0.261).
#' Prior to fitting, networks with fewer than eight nodes (\eqn{p < 8}), more
#' than 300,000 observations, or fewer than one observation per edge
#' (\eqn{\text{ope} \leq 1}) were excluded, as Huth et al. (2025) demonstrated
#' that networks in this regime show the weakest statistical evidence for edge
#' presence or absence, yielding unstable parameter estimates (n = 28 excluded).
#' This left n = 194 networks for analysis. Shape and scale parameters were
#' modelled on their original scales.
#'
#' The two equations have an asymmetric structure. Shape — which governs the
#' concentration of the edge weight distribution — is predicted from \code{snr}
#' and \code{rlp} only, reflecting that the shape of the distribution is a
#' property of the network's signal structure independent of sampling precision.
#' Scale — which governs typical edge weight magnitude — additionally includes
#' \code{scaling}, as the expected size of partial correlations is directly
#' affected by estimation precision. Dropping \code{scaling} from the shape
#' equation costs \eqn{\Delta R^2 < 0.006} and yields cleaner inference; all
#' shape predictors are significant under HC3-robust standard errors
#' (both \eqn{p < 0.001}).
#'
#' Variance inflation factors for the scale equation (\eqn{\leq} 1.60)
#' confirmed the absence of problematic multicollinearity; the shape equation
#' contains only two predictors with no multicollinearity concern.
#' Breusch-Pagan tests indicated statistically significant heteroskedasticity
#' in both equations; however, all predictors remained significant under
#' HC3-robust standard errors, indicating no material effect on inference.
#' Shape residuals were normally distributed (Shapiro-Wilk W = 0.990,
#' p = 0.173). Scale residuals showed a modest departure from normality
#' (Shapiro-Wilk W = 0.973, p < 0.001), consistent with slight right skew
#' in the scale outcome and test sensitivity at n = 194 rather than a
#' substantive violation, as confirmed by visual inspection of the residual
#' histogram. Model fit was strong: shape \eqn{R^2 = 0.887} (RMSE = 0.048);
#' scale \eqn{R^2 = 0.885} (RMSE = 0.011).
#'
#' @keywords datasets
#'
#' @references
#' Huth, K. B. S., Haslbeck, J. M. B., Keetelaar, S., Van Holst, R. J., &
#' Marsman, M. (2025). Statistical evidence in psychological networks.
#' \emph{Nature Human Behaviour}.
#'
#' @examples
#' data("weibull_weights")
#'
#' # Inspect SUR coefficients for each equation
#' weibull_weights$shape$coefficients
#' weibull_weights$scale$coefficients
#'
#' # Predict Weibull parameters for a new network
#' weibull_parameters(nodes = 12, sample_size = 500)
#'
NULL
#----

#' Skew Tables
#'
#' Tables for skew based on the number of categories (2, 3, 4, 5, or 6) in the data
#'
#' @name skew_tables
#'
#' @docType data
#'
#' @usage data(skew_tables)
#'
#' @format A list (length = 5)
#'
#' @keywords datasets
#'
#' @examples
#' data("skew_tables")
#'
NULL

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weibull_weights.R
\docType{data}
\name{weibull_weights}
\alias{weibull_weights}
\title{SUR Model Coefficients and Residuals for Weibull Parameter Prediction}
\format{
A named list with two elements, \code{shape} and \code{scale},
each containing:
\describe{
  \item{\code{coefficients}}{A named numeric vector of regression
  coefficients from the SUR model. The equations are asymmetric: the shape
  equation includes an intercept and two predictors (\code{snr}, \code{rlp});
  the scale equation includes an intercept and three predictors
  (\code{snr}, \code{rlp}, \code{scaling}).
  Predictors are defined as follows:
  \describe{
    \item{\code{snr}}{Signal-to-noise ratio of the absolute partial
    correlations, computed as the mean divided by the standard deviation
    of the absolute edge weights. Used in both equations.}
    \item{\code{rlp}}{Reciprocal log of node count, \eqn{1 / \log(p)},
    where \eqn{p} is the number of nodes. Used in both equations.}
    \item{\code{scaling}}{Standard error of partial correlations,
    \eqn{\sqrt{1 / (n - p)}}, where \eqn{n} is the sample size and
    \eqn{p} is the number of nodes. Used in the scale equation only.}
  }}
  \item{\code{residuals}}{A numeric vector of residuals from the fitted SUR
  equation, used by \code{\link{weibull_parameters}} to introduce
  empirically grounded variability when \code{bootstrap = TRUE}.}
}
}
\usage{
data(weibull_weights)
}
\description{
A named list encoding the Seemingly Unrelated Regression (SUR) model fitted
to Weibull shape and scale parameters derived from 194
empirical networks. The object is consumed internally by \code{\link{weibull_parameters}}
to generate data-driven Weibull parameter estimates given a network's
number of nodes and sample size.
}
\details{
Absolute partial correlations from 222 deduplicated empirical networks
(Huth et al., 2025) were fitted to Beta, Gamma, log-normal, and Weibull
distributions via maximum likelihood. Weibull provided the best fit most
consistently: it outperformed each alternative by more than 2 log-likelihood
units far more often than the reverse (vs. Beta: 56--0; vs. Gamma: 15--2;
vs. log-normal: 155--13; vs. Exponential: 54--0).

The resulting Weibull shape and scale parameters were then jointly modelled
as a function of network descriptors using Seemingly Unrelated Regression
(\code{systemfit}), which accounts for correlated residuals between the
shape and scale equations across networks (residual correlation = 0.261).
Prior to fitting, networks with fewer than eight nodes (\eqn{p < 8}), more
than 300,000 observations, or fewer than one observation per edge
(\eqn{\text{ope} \leq 1}) were excluded, as Huth et al. (2025) demonstrated
that networks in this regime show the weakest statistical evidence for edge
presence or absence, yielding unstable parameter estimates (n = 28 excluded).
This left n = 194 networks for analysis. Shape and scale parameters were
modelled on their original scales.

The two equations have an asymmetric structure. Shape — which governs the
concentration of the edge weight distribution — is predicted from \code{snr}
and \code{rlp} only, reflecting that the shape of the distribution is a
property of the network's signal structure independent of sampling precision.
Scale — which governs typical edge weight magnitude — additionally includes
\code{scaling}, as the expected size of partial correlations is directly
affected by estimation precision. Dropping \code{scaling} from the shape
equation costs \eqn{\Delta R^2 < 0.006} and yields cleaner inference; all
shape predictors are significant under HC3-robust standard errors
(both \eqn{p < 0.001}).

Variance inflation factors for the scale equation (\eqn{\leq} 1.60)
confirmed the absence of problematic multicollinearity; the shape equation
contains only two predictors with no multicollinearity concern.
Breusch-Pagan tests indicated statistically significant heteroskedasticity
in both equations; however, all predictors remained significant under
HC3-robust standard errors, indicating no material effect on inference.
Shape residuals were normally distributed (Shapiro-Wilk W = 0.990,
p = 0.173). Scale residuals showed a modest departure from normality
(Shapiro-Wilk W = 0.973, p < 0.001), consistent with slight right skew
in the scale outcome and test sensitivity at n = 194 rather than a
substantive violation, as confirmed by visual inspection of the residual
histogram. Model fit was strong: shape \eqn{R^2 = 0.887} (RMSE = 0.048);
scale \eqn{R^2 = 0.885} (RMSE = 0.011).
}
\examples{
data("weibull_weights")

# Inspect SUR coefficients for each equation
weibull_weights$shape$coefficients
weibull_weights$scale$coefficients

# Predict Weibull parameters for a new network
weibull_parameters(nodes = 12, sample_size = 500)

}
\references{
Huth, K. B. S., Haslbeck, J. M. B., Keetelaar, S., Van Holst, R. J., &
Marsman, M. (2025). Statistical evidence in psychological networks.
\emph{Nature Human Behaviour}.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/skew_tables.R
\docType{data}
\name{skew_tables}
\alias{skew_tables}
\title{Skew Tables}
\format{
A list (length = 5)
}
\usage{
data(skew_tables)
}
\description{
Tables for skew based on the number of categories (2, 3, 4, 5, or 6) in the data
}
\examples{
data("skew_tables")

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/categorize.R
\name{categorize}
\alias{categorize}
\title{Categorize Continuous Data}
\usage{
categorize(data, categories, skew_value = 0)
}
\arguments{
\item{data}{Numeric (length = n).
A vector of continuous data with \emph{n} values.
For matrices, use \code{apply}}

\item{categories}{Numeric (length = 1).
Number of categories to create.
Between 2 and 6 categories can be used with skew}

\item{skew_value}{Numeric (length = 1).
Value of skew.
Ranges between -2 to 2 in increments of 0.05.
Skews not in this sequence will be converted to
the nearest value in this sequence.
Defaults to \code{0} or no skew}
}
\value{
An integer vector of length \code{n} with values from \code{1} to
\code{categories}, giving the category assignment for each observation.
Category \code{1} corresponds to the lowest values of \code{data} and
category \code{categories} to the highest. When \code{categories > 6},
\code{\link[base]{cut}} is used and skew is not applied.
}
\description{
Categorizes continuous data based on Garrido, Abad and Ponsoda (2011; see references).
Categorical data with 2 to 6 categories can include skew between -2 to 2 in
increments of 0.05
}
\examples{
# Dichotomous data (no skew)
dichotomous <- categorize(
  data = rnorm(1000),
  categories = 2
)

# Dichotomous data (with positive skew)
dichotomous_skew <- categorize(
  data = rnorm(1000),
  categories = 2,
  skew_value = 1.25
)

# 5-point Likert scale (no skew)
five_likert <- categorize(
  data = rnorm(1000),
  categories = 5
)

# 5-point Likert scale (negative skew)
five_likert <- categorize(
  data = rnorm(1000),
  categories = 5,
  skew_value = -0.45
)

}
\references{
Garrido, L. E., Abad, F. J., & Ponsoda, V. (2011). \cr
Performance of Velicer’s minimum average partial factor retention method with categorical variables. \cr
\emph{Educational and Psychological Measurement}, \emph{71}(3), 551-570.

Golino, H., Shi, D., Christensen, A. P., Garrido, L. E., Nieto, M. D., Sadana, R., ... & Martinez-Molina, A. (2020).
Investigating the performance of exploratory graph analysis and traditional techniques to identify the number of latent factors: A simulation and tutorial.
\emph{Psychological Methods}, \emph{25}(3), 292-320.
}
\author{
Maria Dolores Nieto Canaveras <mnietoca@nebrija.es>,
Luis Eduardo Garrido <luisgarrido@pucmm.edu>,
Hudson Golino <hfg9s@virginia.edu>,
Alexander P. Christensen <alexpaulchristensen@gmail.com>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/network_estimation.R
\name{network_estimation}
\alias{network_estimation}
\title{L0 Norm Regularized Network Estimation}
\usage{
network_estimation(
  data,
  n = NULL,
  corr = c("auto", "pearson", "spearman"),
  na_data = c("pairwise", "listwise"),
  penalty = c("atan", "exp", "gumbel", "log", "weibull"),
  gamma = NULL,
  adaptive = TRUE,
  nlambda = 50,
  lambda_min_ratio = 0.01,
  penalize_diagonal = TRUE,
  ic = c("AIC", "AICc", "BIC", "BIC0", "EBIC", "MBIC"),
  ebic_gamma = 0.5,
  fast = TRUE,
  LLA = FALSE,
  LLA_threshold = 0.001,
  LLA_iter = 10000,
  network_only = TRUE,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{data}{Matrix or data frame.
Should consist only of variables to be used in the analysis}

\item{n}{Numeric (length = 1).
Sample size \strong{must} be provided if \code{data} provided is a correlation matrix}

\item{corr}{Character (length = 1).
Method to compute correlations.
Defaults to \code{"auto"}.
Available options:

\itemize{

\item \code{"auto"} --- Automatically computes appropriate correlations for
the data using Pearson's for continuous, polychoric for ordinal,
tetrachoric for binary, and polyserial/biserial for ordinal/binary with
continuous. To change the number of categories that are considered
ordinal, use \code{ordinal_categories}
(see \code{\link[L0ggm]{polychoric_matrix}} for more details)

\item \code{"pearson"} --- Pearson's correlation is computed for all
variables regardless of categories

\item \code{"spearman"} --- Spearman's rank-order correlation is computed
for all variables regardless of categories

}

For other similarity measures, compute them first and input them
into \code{data} with the sample size (\code{n})}

\item{na_data}{Character (length = 1).
How should missing data be handled?
Defaults to \code{"pairwise"}.
Available options:

\itemize{

\item \code{"pairwise"} --- Computes correlation for all available cases between
two variables

\item \code{"listwise"} --- Computes correlation for all complete cases in the dataset

}}

\item{penalty}{Character (length = 1).
Defaults to \code{"weibull"}.
Available options:

\itemize{

\item \code{"atan"} --- Arctangent (Wang & Zhu, 2016)
\deqn{\lambda \cdot (\gamma + \frac{2}{\pi}) \cdot \arctan\left(\frac{|x|}{\gamma}\right)}

\item \code{"exp"} --- EXP (Wang, Fan, & Zhu, 2018)
\deqn{\lambda \cdot (1 - e^{-\frac{|x|}{\gamma}})}

\item \code{"gumbel"} --- Gumbel
\deqn{\frac{\lambda}{1 - e^{-1}} \cdot \left(e^{-e^{-\frac{|x|}{\gamma}}} - e^{-1}\right)}

\item \code{"log"} --- Log (Candes, Wakin, & Boyd, 2008)
\deqn{\frac{\lambda \cdot \log\left(1 + \frac{|x|}{\gamma}\right)}{\log\left(1 + \frac{1}{\gamma}\right)}}

\item \code{"weibull"} --- Weibull
\deqn{\lambda \cdot \left(1 - e^{-\left(\frac{|x|}{\gamma}\right)^k}\right)}

}}

\item{gamma}{Numeric (length = 1).
Adjusts the shape of the penalty.
Defaults:

\itemize{

\item \code{"atan"} = 0.01

\item \code{"exp"} = 0.01

\item \code{"gumbel"} = 0.01

\item \code{"log"} = 0.10

\item \code{"weibull"} = 0.01

}}

\item{adaptive}{Boolean (length = 1).
Whether data-adaptive (gamma) parameters should be used.
Defaults to \code{TRUE}.
Set to \code{FALSE} to apply default gamma parameters for
adaptive penalties.
Available options:

\itemize{

\item \code{"exp"}

\item \code{"gumbel"}

\item \code{"weibull"}

}

When \code{adaptive = TRUE}, gamma is set to the 10th percentile of the
distribution fitted to the empirical partial correlations}

\item{nlambda}{Numeric (length = 1).
Number of lambda values to test.
Defaults to \code{50}}

\item{lambda_min_ratio}{Numeric (length = 1).
Ratio of lowest lambda value compared to maximal lambda.
Defaults to \code{0.01}}

\item{penalize_diagonal}{Boolean (length = 1).
Should the diagonal be penalized?
Defaults to \code{TRUE}}

\item{ic}{Character (length = 1).
What information criterion should be used for model selection?
Available options include:

\itemize{

\item \code{"AIC"} --- Akaike's information criterion: \eqn{-2L + 2E}

\item \code{"AICc"} --- AIC corrected: \eqn{AIC + \frac{2E^2 + 2E}{n - E - 1}}

\item \code{"BIC"} --- Bayesian information criterion: \eqn{-2L + E \cdot \log{(n)}}

\item \code{"BIC0"} --- Bayesian information criterion not (Dicker et al., 2013): \eqn{\log{\large(\frac{D}{n - E}\large)} + \large(\frac{\log{(n)}}{n}\large) \cdot E}

\item \code{"EBIC"} --- Extended BIC: \eqn{BIC + 4E \cdot \gamma \cdot \log{(E)}}

\item \code{"MBIC"} --- Modified Bayesian information criterion (Wang et al., 2018):  \eqn{\log{\large(\frac{D}{n - E}\large)} + \large(\frac{\log{(n)} \cdot E}{n}\large) \cdot \log{(\log{(p)}})}

}

Term definitions:

\itemize{

\item \eqn{n} --- sample size

\item \eqn{p} --- number of variables

\item \eqn{E} --- edges

\item \eqn{S} --- empirical correlation matrix

\item \eqn{K} --- estimated inverse covariance matrix (network)

\item \eqn{L = \frac{n}{2} \cdot \log \text{det} K - \sum_{i=1}^p (SK)_{ii}}

\item \eqn{D = n \cdot \sum_{i=1}^p (SK)_{ii} - \log \text{det} K}

}

Defaults to \code{"BIC"}}

\item{ebic_gamma}{Numeric (length = 1)
Value to set gamma parameter in EBIC (see above).
Defaults to \code{0.50}

\emph{Only used if \code{ic = "EBIC"}}}

\item{fast}{Boolean (length = 1).
Whether the \code{\link[glassoFast]{glassoFast}} version should be used
to estimate the GLASSO.
Defaults to \code{TRUE}.

The fast results \emph{may} differ by less than floating point of the original
GLASSO implemented by \code{\link[glasso]{glasso}} and should not impact reproducibility much (set to \code{FALSE} if concerned)}

\item{LLA}{Boolean (length = 1).
Should Local Linear Approximation be used to find optimal minimum?
Defaults to \code{FALSE} or a single-pass approximation, which can be
significantly faster (Zou & Li, 2008).
Set to \code{TRUE} to find global minimum based on convergence (\code{LLA_threshold})}

\item{LLA_threshold}{Numeric (length = 1).
When performing the Local Linear Approximation, the maximum threshold
until convergence is met.
Defaults to \code{1e-03}}

\item{LLA_iter}{Numeric (length = 1).
Maximum number of iterations to perform to reach convergence.
Defaults to \code{10000}}

\item{network_only}{Boolean (length = 1).
Whether the network only should be output.
Defaults to \code{TRUE}.
Set to \code{FALSE} to obtain all output for the
network estimation method}

\item{verbose}{Boolean (length = 1).
Whether messages and (insignificant) warnings should be output.
Defaults to \code{FALSE} (silent calls).
Set to \code{TRUE} to see all messages and warnings for every function call}

\item{...}{Additional arguments to be passed on to \code{auto.correlate}}
}
\value{
When \code{network_only = TRUE} (default), returns a \eqn{p \times p}
numeric matrix of partial correlations representing the estimated Gaussian
graphical model. Off-diagonal entry \eqn{[i,j]} is the partial correlation
between variables \eqn{i} and \eqn{j} controlling for all other variables,
with values in \eqn{[-1, 1]}; a value of zero indicates the absence of an
edge. Diagonal entries are zero. Row and column names are inherited from
\code{data}.

When \code{network_only = FALSE}, returns a named list with the following
elements:

\describe{

\item{\code{network}}{The \eqn{p \times p} partial correlation matrix
described above}

\item{\code{K}}{The \eqn{p \times p} estimated inverse covariance
(precision) matrix at the optimal lambda. Diagonal entries are the
conditional precisions; off-diagonal entries are proportional to
partial covariances}

\item{\code{R}}{The \eqn{p \times p} regularized covariance matrix
returned by GLASSO at the optimal lambda (the \code{w} component of the
GLASSO solution)}

\item{\code{penalty}}{Character string naming the penalty function used
(one of \code{"atan"}, \code{"exp"}, \code{"gumbel"}, \code{"log"},
\code{"weibull"})}

\item{\code{lambda}}{Numeric scalar giving the regularization parameter
value selected by the information criterion. Larger values correspond
to sparser networks}

\item{\code{gamma}}{Numeric scalar giving the shape parameter of the
penalty actually used. For adaptive penalties (\code{adaptive = TRUE}),
this is the data-derived value; otherwise it is the default or
user-supplied value}

\item{\code{correlation}}{The \eqn{p \times p} empirical correlation
matrix computed from \code{data}, used as input to the GLASSO}

\item{\code{criterion}}{Character string naming the information criterion
used for model selection (e.g., \code{"bic"})}

\item{\code{IC}}{Numeric scalar giving the value of the information
criterion at the optimal lambda}

}
}
\description{
A general function to estimate Gaussian graphical models using
L0 penalty approximations. All penalties are implemented using
either single-pass or full Local Linear Approximation (LLA: Fan & Li, 2001; Zou & Li, 2008)
}
\examples{
# Obtain default estimator (adaptive Weibull)
weibull_network <- network_estimation(
  data = basic_smallworld, LLA = TRUE
)

# Obtain Atan network
atan_network <- network_estimation(
  data = basic_smallworld, penalty = "atan", LLA = TRUE
)

# Obtain static EXP network
exp_network <- network_estimation(
  data = basic_smallworld, penalty = "exp",
  adaptive = FALSE, LLA = TRUE
)

}
\references{
\strong{Log penalty} \cr
Candes, E. J., Wakin, M. B., & Boyd, S. P. (2008).
Enhancing sparsity by reweighted l1 minimization.
\emph{Journal of Fourier Analysis and Applications}, \emph{14}(5), 877--905.

\strong{BIC0} \cr
Dicker, L., Huang, B., & Lin, X. (2013).
Variable selection and estimation with the seamless-L0 penalty.
\emph{Statistica Sinica}, \emph{23}(2), 929--962.

\strong{Local Linear Approximation} \cr
Fan, J., & Li, R. (2001).
Variable selection via nonconcave penalized likelihood and its oracle properties.
\emph{Journal of the American Statistical Association}, \emph{96}(456), 1348--1360.

\strong{EXP penalty} \cr
Wang, Y., Fan, Q., & Zhu, L. (2018).
Variable selection and estimation using a continuous approximation to the L0 penalty.
\emph{Annals of the Institute of Statistical Mathematics}, \emph{70}(1), 191--214.

\strong{Atan penalty} \cr
Wang, Y., & Zhu, L. (2016).
Variable selection and parameter estimation with the Atan regularization method.
\emph{Journal of Probability and Statistics}, \emph{2016}, 1--12.

\strong{Seminal simulation in network psychometrics} \cr
Williams, D. R. (2020).
Beyond lasso: A survey of nonconvex regularization in Gaussian graphical models.
\emph{PsyArXiv}.

\strong{One-step Local Linear Approximation} \cr
Zou, H., & Li, R. (2008).
One-step sparse estimates in nonconcave penalized likelihood models.
\emph{Annals of Statistics}, \emph{36}(4), 1509--1533.
}
\author{
Alexander P. Christensen <alexpaulchristensen@gmail.com>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/network_fit.R
\name{network_fit}
\alias{network_fit}
\title{Traditional Fit Metrics for Networks}
\usage{
network_fit(network, n, S, ci = 0.95)
}
\arguments{
\item{network}{Matrix or data frame.
A p by p square network matrix}

\item{n}{Numeric (length = 1).
Sample size}

\item{S}{Matrix or data frame.
A p by p square zero-order correlation matrix corresponding
with the input \code{network}}

\item{ci}{Numeric (length = 1).
Confidence interval for RMSEA.
Defaults to \code{0.95}}
}
\value{
A named numeric vector of traditional and likelihood-based fit
indices. The vector always contains the following elements:

\describe{

\item{\code{chisq}}{Chi-square statistic
(\eqn{\chi^2 = n \cdot F_{ML}}), where \eqn{F_{ML}} is the maximum
likelihood discrepancy function between the model-implied and empirical
correlation matrices}

\item{\code{df}}{Degrees of freedom: total number of unique off-diagonal
correlations minus the number of non-zero edges in \code{network}}

\item{\code{chisq.p.value}}{p-value for the chi-square test of exact fit
(H0: model-implied covariance equals the population covariance)}

\item{\code{RMSEA}}{Root mean square error of approximation. Values
\eqn{\leq} 0.05 indicate close fit; values \eqn{\leq} 0.08 indicate
acceptable fit}

\item{\code{RMSEA.XX.lower}, \code{RMSEA.XX.upper}}{Lower and upper
bounds of the \code{ci}-level confidence interval for RMSEA, where
\code{XX} is the integer percentage (e.g., \code{RMSEA.95.lower} and
\code{RMSEA.95.upper} for a 95\% CI)}

\item{\code{RMSEA.p.value}}{p-value for the one-sided test of close fit
(H0: RMSEA \eqn{\leq} 0.05)}

\item{\code{CFI}}{Comparative fit index, comparing the target model to
an independence (null) baseline. Values \eqn{\geq} 0.95 indicate
acceptable fit}

\item{\code{TLI}}{Tucker-Lewis index (non-normed fit index). Values
\eqn{\geq} 0.95 indicate acceptable fit; can fall outside \eqn{[0, 1]}
for severely misspecified models}

\item{\code{SRMR}}{Standardized root mean residual: the root mean squared
difference between the model-implied and observed correlation matrices.
Values \eqn{\leq} 0.08 indicate acceptable fit}

\item{\code{logLik}}{Gaussian log-likelihood of the model-implied
correlation matrix, assuming zero mean structure (means are not
estimated)}

\item{\code{AIC}}{Akaike's information criterion:
\eqn{-2 \cdot \log L + 2 \cdot E}, where \eqn{E} is the number of
non-zero edges in \code{network}}

\item{\code{BIC}}{Bayesian information criterion:
\eqn{-2 \cdot \log L + E \cdot \log(n)}}

}
}
\description{
Computes several traditional fit metrics for networks including

\itemize{

\item chi-square (\eqn{\chi^2})

\item root mean square error of approximation (RMSEA) with confidence intervals

\item confirmatory fit index (CFI)

\item Tucker-Lewis index (TLI)

\item standardized root mean residual (SRMR)

\item log-likelihood

\item Akaike's information criterion (AIC)

\item Bayesian information criterion (BIC)

}
}
\examples{
# Obtain correlation matrix
S <- auto_correlate(basic_smallworld)

# Obtain Weibull network
weibull_network <- network_estimation(data = basic_smallworld, LLA = TRUE)

# Obtain fit (expects continuous variables!)
network_fit(network = weibull_network, n = nrow(basic_smallworld), S = S)
# Scaled metrics are not yet available for
# dichotomous or polytomous data!

}
\references{
Epskamp, S., Rhemtulla, M., & Borsboom, D. (2017).
Generalized network psychometrics: Combining network and latent variable models.
\emph{Psychometrika}, \emph{82}(4), 904–927.
}
\author{
Alexander P. Christensen <alexpaulchristensen@gmail.com>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/basic_smallworld.R
\docType{data}
\name{basic_smallworld}
\alias{basic_smallworld}
\title{Toy Small-world Network Data Example}
\format{
A 500 \eqn{\times} 20 continuous response matrix
}
\usage{
data(basic_smallworld)
}
\description{
A toy small-world network example (\emph{n} = 500)
}
\examples{
# Generated using:
# basic_smallworld <- simulate_smallworld(
#  nodes = 20, # 20 nodes in the network
#  density = 0.30, # moderate initial lattice connectivity
#  rewire = 0.20, # 20\% rewiring probability
#  sample_size = 500 # number of cases = 500
# )$data
data("basic_smallworld")

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weibull_parameters.R
\name{weibull_parameters}
\alias{weibull_parameters}
\title{Predict Weibull Parameters for Edge Weight Distributions}
\usage{
weibull_parameters(nodes, sample_size, snr = 1, bootstrap = FALSE)
}
\arguments{
\item{nodes}{Integer. The number of nodes (variables) in the network.
Must be between 8 and 54, reflecting the range of the empirical networks
used to fit the underlying SUR model.}

\item{sample_size}{Integer. The sample size of the dataset from which the
network is estimated.}

\item{snr}{Numeric (length = 1).
Signal-to-noise ratio of partial correlations (\eqn{\bar{|w|} / \mathrm{SD}(|w|)}).
Values less than 1 indicate wider range of partial correlations (\eqn{w}) whereas
values greater than 1 indicate narrower range.
Defaults to \code{1} where the mean of the partial correlations (\eqn{\bar{|w|}})
is equal to the standard deviation (\eqn{\mathrm{SD}(|w|)})}

\item{bootstrap}{Logical. If \code{TRUE}, a randomly sampled residual from
the SUR model fit is added to each predicted parameter, introducing
empirically grounded variability suitable for use in simulation or
bootstrapping contexts. Defaults to \code{FALSE}.}
}
\value{
A named numeric vector of length 2:
\describe{
  \item{\code{shape}}{The predicted Weibull shape parameter (\eqn{k > 0}).}
  \item{\code{scale}}{The predicted Weibull scale parameter (\eqn{\lambda > 0}).}
}
}
\description{
Predicts the shape and scale parameters of a Weibull distribution that
characterizes the absolute partial correlation edge weights of a psychometric
network, given its number of nodes and sample size. Parameter estimates are
derived from a Seemingly Unrelated Regression (SUR) model fitted to empirical
network data from Huth et al. (2025), where absolute partial correlations were
found to follow a Weibull distribution more consistently than Beta, Gamma, or
log-normal alternatives.
}
\details{
The shape and scale parameters are predicted from derived network descriptors
that differ between the two SUR equations:

\describe{
  \item{\code{rlp}}{The reciprocal log of the number of nodes,
    \eqn{1 / \log(p)}, capturing the diminishing marginal effect of network
    size on edge weight distributions. Used in both the shape and scale
    equations.}
  \item{\code{scaling}}{The standard error of partial correlations defined as
    \eqn{\sqrt{1 / (n - p)}}, where \eqn{n} is the sample size and
    \eqn{p} is the number of nodes. Larger values indicate greater sampling
    uncertainty in the partial correlation estimates. Used in the scale
    equation only.}
}

The two SUR equations have an asymmetric structure reflecting different
theoretical roles for sampling precision. Shape — which governs the
concentration of the edge weight distribution — is determined solely by
signal characteristics of the network via \code{snr} and \code{rlp}.
Scale — which governs the typical magnitude of edge weights — additionally
depends on \code{scaling}, as the expected size of partial correlations is
directly affected by estimation precision. This asymmetry is both
empirically supported (dropping \code{scaling} from the shape equation
costs \eqn{\Delta R^2 < 0.006}) and theoretically coherent.

These predictors enter two SUR equations whose coefficients are stored in
the internal \code{weibull_weights} dataset. SUR was used to account for the
correlated residuals between the shape and scale equations across networks
(residual correlation = 0.261). Model fit was strong: the shape equation
achieved \eqn{R^2 = 0.887} (RMSE = 0.048) and the scale equation achieved
\eqn{R^2 = 0.885} (RMSE = 0.011). Shape residuals were normally distributed
(Shapiro-Wilk W = 0.990, p = 0.173). Scale residuals showed a modest
departure from normality (Shapiro-Wilk W = 0.973, p < 0.001), consistent
with slight right skew in the scale outcome and test sensitivity at n = 194
rather than a substantive violation. Heteroskedasticity was detected in both
the shape equation (Breusch-Pagan p < 0.001) and the scale equation
(Breusch-Pagan p < 0.001); robust standard errors (HC3) were used for
inference. Multicollinearity among predictors in the scale equation was
negligible (VIF \eqn{\leq} 1.60); the shape equation contains only two
predictors with no multicollinearity concern.

\code{nodes} influences predicted shape and scale via \code{rlp}.
\code{sample_size} influences predicted scale via \code{scaling}, and
both parameters via their joint contribution to \code{snr} when it is
estimated from data rather than supplied directly.

Empirically, shape values ranged from approximately 0.72 to 1.63
(M = 1.07, SD = 0.14) and scale values from approximately 0.03 to 0.19
(M = 0.10, SD = 0.03) across the 194 networks used to fit the model. Shape
values near 1 indicate approximately exponential edge weight distributions;
values above 1 indicate a rising hazard (mode-bearing distribution).

When \code{bootstrap = TRUE}, residuals are drawn via \code{shuffle()} --
a random sampling without replacement -- from the empirical SUR residuals,
preserving the observed marginal residual distribution.
}
\examples{
# Predict parameters for a 10-node network with n = 500
weibull_parameters(nodes = 10, sample_size = 500)

# With bootstrapped residuals for use in simulation
weibull_parameters(nodes = 10, sample_size = 500, bootstrap = TRUE)

}
\references{
Huth, K. B. S., Haslbeck, J. M. B., Keetelaar, S., Van Holst, R. J., &
Marsman, M. (2025). Statistical evidence in psychological networks.
\emph{Nature Human Behaviour}.
}
\author{
Alexander P. Christensen <alexpaulchristensen@gmail.com>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxswap_lattice.R
\name{proxswap_lattice}
\alias{proxswap_lattice}
\title{Construct a Degree-Preserving Ring Lattice via Proximity-Swap Construction}
\usage{
proxswap_lattice(network, weighted = FALSE, shuffles = 100)
}
\arguments{
\item{network}{Matrix.
A square, symmetric numeric matrix representing a network (e.g., partial
correlations). Non-zero off-diagonal entries are treated as edges; the
binary adjacency is derived internally as \code{network != 0}.
Isolated nodes (degree zero) are supported.}

\item{weighted}{Logical (length = 1).
Whether to return a weighted ring lattice. When \code{TRUE}, the edge
weights from \code{network} are reassigned to the lattice edges according
to ring distance following the implementation of Muldoon, Bridgeford, &
Bassett (2016): shorter-distance lattice edges receive larger weights,
preserving the overall weight distribution while concentrating stronger
connections locally. When \code{FALSE} (default), a binary adjacency
matrix is returned.}

\item{shuffles}{Numeric (length = 1).
Number of independent random permutation passes to attempt. Each pass
assigns the observed degree sequence to ring positions in a new random
order, runs the proximity construction, and applies swap repair if needed.
Only passes producing a connected graph with zero degree error are
retained; the one with the highest clustering coefficient is returned.
Defaults to \code{100}.}
}
\value{
A square symmetric matrix of the same dimension as \code{network},
with row and column names preserved, representing the resulting ring
lattice. When \code{weighted = FALSE} (default), entries are binary integers
(\code{0L}/\code{1L}); when \code{weighted = TRUE}, entries contain the
reassigned edge weights from \code{network} with original signs preserved.
The average clustering coefficient of the returned graph is attached as the
attribute \code{"CC"} and can be retrieved with \code{attr(result, "CC")}.
When the empirical fallback is triggered (see Details), \code{"CC"} reflects
the empirical network's clustering coefficient rather than a lattice's.
}
\description{
Converts a network matrix into a connected ring lattice whose
degree sequence exactly matches the original, while maximizing the average
local clustering coefficient. The resulting lattice is intended as a null
model for small-world network analyses. An adjacency matrix is derived
automatically from \code{network} (non-zero entries become edges), so the
function accepts any weighted or binary network directly.

Each pass randomly permutes the degree sequence onto ring positions, then
greedily assigns edges in strictly increasing ring-distance order (proximity
construction). Any residual degree deficit is resolved by a swap-repair
phase. Passes that yield a disconnected graph or an unsatisfied degree
sequence are discarded; among valid passes the one with the highest average
clustering coefficient is returned. \code{shuffles} independent passes are
attempted in total.
}
\details{
## Algorithm

\strong{Pair precomputation.} A ring distance matrix is computed as
\eqn{d_{ij} = \min(|i - j|,\, n - |i - j|)}, giving true circular
distances bounded by \eqn{\lfloor n/2 \rfloor}. All unique unordered pairs
at each distance \eqn{r = 1, \ldots, \lfloor n/2 \rfloor} are extracted
once (retaining only entries where \eqn{i < j}) and cached for reuse
across all passes.

\strong{Proximity construction.} The observed degree sequence is randomly
permuted onto ring positions. Starting from an empty graph, edges are added
in distance order. At each distance band the eligible pairs (both endpoints
have remaining degree budget and are not yet connected) are sorted by
descending \eqn{\min(\text{budget}_i, \text{budget}_j)} so that high-need
pairs receive short connections first. Pairs are then assigned sequentially
with per-pair budget re-checks. The pass exits early once all budgets reach
zero. This phase is implemented in compiled C via \code{proximity_pass_c}.

\strong{Swap repair.} If any degree deficit remains after the proximity
pass, the highest-deficit node \eqn{i} is connected to its nearest
available ring neighbour, scanning clockwise and counter-clockwise
positions in interleaved distance order. When no direct partner with
remaining budget exists, an edge swap is performed: a nearby unconnected
node \eqn{j} is found, one of \eqn{j}'s existing edges to \eqn{k} is
removed, and a new edge \eqn{(i, j)} is added. Node \eqn{k} recovers its
budget for resolution in a subsequent iteration. This repeats until all
deficits are resolved or the iteration cap (\eqn{2n^2}) is reached. This
phase is implemented in compiled C via \code{swapping_pass_c}.

\strong{Weight assignment.} When \code{weighted = TRUE}, edge weights are
reassigned after the binary topology is finalised. The observed weights are
sorted by \emph{absolute value} in descending order (largest magnitude first)
and mapped onto lattice edges ranked by ascending ring distance (i.e.,
\eqn{d_{ij} = \min(|i - j|,\, n - |i - j|)}), so that shorter (more local)
connections receive the largest-magnitude weights. Original signs are
preserved: the sorted weight vector — not its absolute values — is assigned
to the lattice edges. This directly implements the distance-weight principle
of Muldoon, Bridgeford, & Bassett (2016), using the ring's structural
distances rather than any network-derived proxy. Ties in ring distance are
broken at random.

\strong{Pass selection.} A pass is valid only if the resulting graph is
connected and has zero residual degree error. Among valid passes, the one
with the highest average clustering coefficient is returned.

\strong{Empirical fallback.} If no valid pass is found, or if the best
lattice clustering coefficient is lower than that of the original network,
the empirical adjacency (or weighted matrix, if \code{weighted = TRUE}) is
returned with a warning.
}
\examples{
# Get network
network <- network_estimation(basic_smallworld)

# Construct binary ring lattice
L <- proxswap_lattice(network)

# Retrieve the attached clustering coefficient
attr(L, "CC")

# Degree sequences should match exactly
cbind(target = colSums(network != 0), achieved = colSums(L))

# Construct weighted ring lattice
L_weighted <- proxswap_lattice(network, weighted = TRUE)

# Retrieve the attached clustering coefficient
attr(L_weighted, "CC")

}
\references{
\strong{Logic for weight assignments} \cr
Muldoon, S. F., Bridgeford, E. W., & Bassett, D. S. (2016).
Small-world propensity and weighted brain connectivity.
\emph{Scientific Reports}, \emph{6}, 22057.
}
\author{
Alexander P. Christensen <alexpaulchristensen@gmail.com>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/polychoric_matrix.R
\name{polychoric_matrix}
\alias{polychoric_matrix}
\title{Computes Polychoric Correlations}
\usage{
polychoric_matrix(
  data,
  na_data = c("pairwise", "listwise"),
  empty_method = c("none", "zero", "all"),
  empty_value = c("none", "point_five", "one_over"),
  ...
)
}
\arguments{
\item{data}{Matrix or data frame.
A dataset with all ordinal values
(rows = cases, columns = variables).
Data are required to be between \code{0} and \code{11}.
Proper adjustments should be made prior to analysis (e.g.,
scales from -3 to 3 in increments of 1 should be shifted
by added 4 to all values)}

\item{na_data}{Character (length = 1).
How should missing data be handled?
Defaults to \code{"pairwise"}.
Available options:

\itemize{

\item \code{"pairwise"} --- Computes correlation for all available cases between
two variables

\item \code{"listwise"} --- Computes correlation for all complete cases in the dataset

}}

\item{empty_method}{Character (length = 1).
Method for empty cell correction.
Available options:

\itemize{

\item \code{"none"} --- Adds no value (\code{empty_value = "none"})
to the empirical joint frequency table between two variables

\item \code{"zero"} --- Adds \code{empty_value} to the cells with zero
in the joint frequency table between two variables

\item \code{"all"} --- Adds \code{empty_value} to all
in the joint frequency table between two variables

}}

\item{empty_value}{Character (length = 1).
Value to add to the joint frequency table cells.
Accepts numeric values between 0 and 1 or
specific methods:

\itemize{

\item \code{"none"} --- Adds no value (\code{0}) to the empirical joint
frequency table between two variables

\item \code{"point_five"} --- Adds \code{0.5} to the cells defined by \code{empty_method}

\item \code{"one_over"} --- Adds \code{1 / n} where \emph{n} equals the number of cells
based on \code{empty_method}. For \code{empty_method = "zero"},
\emph{n} equals the number of \emph{zero} cells

}}

\item{...}{Not used but made available for easier
argument passing}
}
\value{
A symmetric numeric matrix of dimension \eqn{p \times p}, where
\eqn{p} is the number of variables (columns) in \code{data}. Each
off-diagonal entry \eqn{[i,j]} is the polychoric correlation between
variables \eqn{i} and \eqn{j} — the estimated Pearson correlation between
the latent continuous variables assumed to underlie the observed ordinal
categories — with values in \eqn{[-1, 1]}. Diagonal entries are \eqn{1}.
Row and column names are inherited from \code{data}. If any variable has
zero variance, its corresponding row and column are set to \code{NA} and
a warning is issued.
}
\description{
A fast implementation of polychoric correlations in C.
Uses the Beasley-Springer-Moro algorithm (Boro & Springer, 1977; Moro, 1995)
to estimate the inverse univariate normal CDF, the Drezner-Wesolosky
approximation (Drezner & Wesolosky, 1990) to estimate the bivariate normal
CDF, and Brent's method (Brent, 2013) for optimization of rho
}
\examples{
# Categorize data
basic_categorized <- apply(
  basic_smallworld, 2, categorize, categories = 5
)

# Compute polychoric correlation matrix
correlations <- polychoric_matrix(basic_categorized)

# Randomly assign missing data
basic_categorized[sample(1:length(basic_categorized), 1000)] <- NA

# Compute polychoric correlation matrix with pairwise missing
pairwise_correlations <- polychoric_matrix(
  basic_categorized, na_data = "pairwise"
)

# Compute polychoric correlation matrix with listwise missing
pairwise_correlations <- polychoric_matrix(
  basic_categorized, na_data = "listwise"
)

}
\references{
\strong{Beasley-Moro-Springer algorithm} \cr
Beasley, J. D., & Springer, S. G. (1977).
Algorithm AS 111: The percentage points of the normal distribution.
\emph{Journal of the Royal Statistical Society. Series C (Applied Statistics)}, \emph{26}(1), 118-121.

Moro, B. (1995).
The full monte.
\emph{Risk 8 (February)}, 57-58.

\strong{Brent optimization} \cr
Brent, R. P. (2013).
Algorithms for minimization without derivatives.
Mineola, NY: Dover Publications, Inc.

\strong{Drezner-Wesolowsky bivariate normal approximation} \cr
Drezner, Z., & Wesolowsky, G. O. (1990).
On the computation of the bivariate normal integral.
\emph{Journal of Statistical Computation and Simulation}, \emph{35}(1-2), 101-107.
}
\author{
Alexander P. Christensen <alexpaulchristensen@gmail.com> with assistance from GPT-4
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate_smallworld.R
\name{simulate_smallworld}
\alias{simulate_smallworld}
\title{Simulates Small-World GGM Data}
\usage{
simulate_smallworld(
  nodes,
  density,
  rewire,
  snr = 1,
  negative_proportion,
  sample_size,
  skew = 0,
  skew_range = NULL,
  target_condition = 30,
  max_correlation = 0.8,
  max_iterations = 100
)
}
\arguments{
\item{nodes}{Numeric (length = 1).
Number of nodes in the network.
Minimum of three nodes. The total number of nodes should be between 8
and 54 to remain within the range of the empirical networks used to fit
the Weibull parameter model; values outside this range are accepted but
will trigger extrapolation and may produce a warning from
\code{\link{weibull_parameters}}.}

\item{density}{Numeric (length = 1).
Target edge density of the network after pruning the initial ring lattice.
Controls the number of nearest neighbors \code{k} each node is connected
to via \eqn{k = \mathrm{round}((\text{nodes} \times
\frac{\text{nodes}-1}{2} \times \text{density}) / \text{nodes})}, subject
to a minimum of 2 and a maximum of
\eqn{\lfloor (\text{nodes}-1)/2 \rfloor}. A ring lattice with
\code{neighbors + 1} connections per node is first generated and then
pruned to this density, introducing degree heterogeneity before rewiring.
Must be between 0 and 1. A minimum density sufficient to maintain a
connected graph is enforced; values below this threshold will produce an
informative error.}

\item{rewire}{Numeric (length = 1).
Probability of rewiring each edge. Unlike the standard Watts-Strogatz
model, rewired edges are placed using degree-weighted random selection:
node pairs with higher combined degree have a greater probability of
receiving a rewired edge (see Details). Values near 0 preserve the
pruned lattice structure; values near 1 produce approximately random
networks. The small-world regime typically occurs at intermediate values
(roughly 0.01 to 0.30). Must be between 0 and 1.}

\item{snr}{Numeric (length = 1).
Signal-to-noise ratio of the absolute partial correlation weights,
defined as \eqn{\bar{|w|} / \mathrm{SD}(|w|)}. Values less than 1
produce wider, more heterogeneous weight distributions; values greater
than 1 produce narrower, more homogeneous distributions. Empirically
observed SNR values ranged from 0.648 to 1.712; values outside this
range are accepted but will trigger a warning. Defaults to \code{1}.}

\item{negative_proportion}{Numeric (length = 1).
Proportion of edges that are negative (inhibitory).
Must be between 0 and 0.50. The upper bound of 0.50 is a mathematical
constraint of the sign-flipping procedure used to assign negative edges
(see Details). If not provided, a value is sampled from a truncated
normal distribution reflecting the empirical distribution of true
negative partial correlations across 194 psychometric networks:
mean = 0.34, SD = 0.086, bounded to \eqn{[0.083, 0.50]}.}

\item{sample_size}{Numeric (length = 1).
Number of observations to generate from the population multivariate
normal distribution. Also influences the predicted Weibull scale
parameter via \code{\link{weibull_parameters}}: larger samples are
associated with smaller, more precisely estimated edge weights.}

\item{skew}{Numeric (length = 1 or \code{nodes}).
Skew applied to each variable after generation from the multivariate
normal. Can be a single value (applied to all variables) or one value
per variable. Values are rounded to the nearest increment of 0.05 in
the range \eqn{[-2, 2]}. Defaults to \code{0} (no skew).}

\item{skew_range}{Numeric (length = 2).
If provided, a skew value is drawn uniformly from this range for each
variable, overriding \code{skew}. Both values must be between -2 and 2.}

\item{target_condition}{Numeric (length = 1).
Target condition number (using \code{\link{kappa}} with \code{exact =
TRUE}) applied when ridge regularization is needed to recover a positive
definite precision matrix. The smallest ridge penalty \eqn{\lambda} that
brings the condition number to this target is found via root-finding
(\code{\link{uniroot}}), subject to a maximum shrinkage of approximately
23\% following Peeters et al. (2020). Lower values produce
better-conditioned matrices. Defaults to \code{30}. Values up to
\code{100} are accepted but not recommended.}

\item{max_correlation}{Numeric (length = 1).
Maximum allowed absolute pairwise correlation in the population
correlation matrix \code{R}. Any draw where
\code{max(abs(R[lower.tri(R)])) > max_correlation} is rejected and a
new attempt is made. Must be between 0 and 1. Defaults to \code{0.80}.}

\item{max_iterations}{Numeric (length = 1).
Maximum number of attempts to find (1) a connected network structure,
(2) a network satisfying the small-world screening criterion, and (3) a
valid set of edge weights. The error message reports a frequency table
of rejection reasons to assist with diagnosing convergence failures.
Defaults to \code{100}.}
}
\value{
A named list with four elements:

\item{data}{Numeric matrix of dimension \code{sample_size x nodes}
containing the simulated observations drawn from the population GGM.
Rows are cases; columns are variables named \code{V01}, \code{V02},
etc. Values are continuous (or skewed continuous when \code{skew != 0}).
To produce ordinal data, pass the columns through
\code{\link{categorize}}.}

\item{parameters}{
A list of input, derived, and estimated parameters:
\itemize{
  \item \code{nodes} --- Number of nodes (as supplied)
  \item \code{density} --- Target edge density (as supplied)
  \item \code{neighbors} --- Number of nearest neighbors \code{k} derived
  from \code{nodes} and \code{density}; the initial ring lattice uses
  \code{neighbors + 1} connections before pruning
  \item \code{rewire} --- Edge rewiring probability (as supplied)
  \item \code{negative_proportion} --- Proportion of negative edges,
  either as supplied or as sampled from the empirical distribution
  \item \code{sample_size} --- Number of simulated observations
  \item \code{skew} --- Named numeric vector of per-variable skew values
  actually applied (after rounding and possible resampling)
  \item \code{weibull} --- Named numeric vector of length 2 giving the
  Weibull \code{shape} and \code{scale} parameters of the absolute edge
  weight distribution actually used. If ridge conditioning was applied,
  these are re-estimated from the conditioned network via MLE.
  \item \code{omega} --- Smallworldness omega statistic of the generated
  network (Telesford et al., 2011). Values near zero indicate small-world
  structure; negative values indicate lattice-like structure; positive
  values indicate random-like structure (see \code{\link[L0ggm]{smallworldness}})
  \item \code{Q} --- Newman-Girvan modularity of the population network
  (\code{Omega}) computed via \code{igraph::modularity} on absolute edge
  weights, using the community partition returned by
  \code{igraph::cluster_leiden} with \code{objective_function =
  "modularity"}. Because \code{cluster_leiden} is a heuristic algorithm,
  \code{Q} reflects a high-quality but not necessarily optimal partition.
  Values near zero indicate absence of community structure, consistent
  with the network theory of psychopathology.
}
}

\item{population}{
Population-level network parameters:
\itemize{
  \item \code{R} --- Population correlation matrix derived from the GGM
  via \code{pcor2cor}
  \item \code{Omega} --- Population partial correlation matrix (the GGM
  edge weight matrix), with zeros for absent edges
}
}

\item{convergence}{
Iteration and conditioning diagnostics:
\itemize{
  \item \code{iterations} --- Number of sampling attempts needed to find
  a valid network
  \item \code{rejections} --- Character vector recording the rejection
  reason for each failed attempt. Common reasons include disconnected
  graph structure, omega exceeding the small-world threshold, condition
  number exceeding \code{target_condition}, and maximum correlation
  exceeding \code{max_correlation}
  \item \code{lambda} --- Ridge regularization parameter \eqn{\lambda}
  added to the diagonal of the precision matrix to ensure positive
  definiteness; \code{NA} if no conditioning was required
  \item \code{condition} --- Condition number of the final population
  correlation matrix \code{R}
}
}
}
\description{
Simulates data from a Gaussian Graphical Model (GGM) with a small-world
network structure. The generative process proceeds in three stages. First,
a ring lattice is constructed with \code{neighbors + 1} nearest-neighbor
connections per node, where \code{neighbors} is derived from \code{nodes}
and \code{density}. Second, the lattice is randomly pruned to the target
\code{density}, introducing degree heterogeneity from the outset. Third,
edges are rewired with probability \code{rewire}, where rewired edges are
placed preferentially on node pairs with higher combined degree (degree-
weighted rewiring). This approach produces more realistic degree
distributions than standard Watts-Strogatz rewiring while preserving
the local clustering structure of the lattice. Edge weights are assigned
by structural priority: edges retained from the pruned lattice receive
larger partial correlation weights based on their original neighbor
distance, grounded in the empirical observation that shorter-distance
connections tend to carry stronger weights in psychometric networks.
The resulting network is used to generate multivariate normal (or skewed)
data. Parameters do not have default values (except
\code{negative_proportion}, \code{snr}, \code{target_condition},
\code{max_correlation}, and \code{max_iterations}) and must each be set.
See Details and Examples to get started.
}
\details{
\strong{Lattice generation and pruning}

The generative process begins by constructing a ring lattice with
\code{neighbors + 1} nearest-neighbor connections per node via
\code{\link[igraph]{sample_smallworld}} with \code{p = 0}. The
\code{+1} overshoot ensures the lattice always has more edges than the
target density, guaranteeing that pruning can proceed. The lattice is
then randomly pruned to the target \code{density} by removing edges
uniformly at random, subject to the constraint that the pruned graph
remains connected. Because removal is random, different nodes lose
different numbers of edges, producing a heterogeneous degree distribution
before any rewiring occurs. This heterogeneity provides a non-uniform
prior for the subsequent degree-weighted rewiring step.

\strong{Degree-weighted rewiring}

Each edge in the pruned lattice is independently selected for rewiring
with probability \code{rewire}. For each selected edge \eqn{(i, j)},
node \eqn{i} is kept fixed and the \eqn{j} endpoint is redirected to a
new target node. Valid targets are restricted to node pairs involving
node \eqn{i} that (1) are not currently connected, and (2) have never
been occupied during the current rewiring pass (i.e., were absent in
the original pruned lattice). Among valid targets, the new endpoint is
selected with probability proportional to \eqn{\sqrt{d_i + d_k}}, where
\eqn{d_i} and \eqn{d_k} are the current degrees of the two nodes in
the candidate pair. The square-root transformation moderates the
rich-get-richer tendency of linear preferential attachment, producing
degree heterogeneity consistent with the empirical range of psychometric
networks without generating extreme hubs. Node degrees are updated
incrementally after each rewire so that subsequent rewiring steps
reflect the current graph state.

\strong{Edge weight assignment}

Edge weights are assigned by ranking edges according to their distance
in the pruned lattice before rewiring. Edges retained from the pruned
lattice receive their original neighbor distance (1 = nearest neighbor,
2 = second nearest, etc.); rewired edges are assigned a distance of
\code{neighbors + 1}, placing them at the bottom of the priority order.
Absolute edge weights are drawn from a Weibull distribution whose
parameters are predicted from the network size, sample size, and signal-
to-noise ratio using a Seemingly Unrelated Regression (SUR) model fitted
to 194 empirical psychometric networks (Huth et al., 2025). The sorted
(descending) Weibull weights are then mapped onto the distance ranking
so that the largest weights go to the shortest-distance edges. This
assignment is grounded in the empirical observation that shorter-distance
local connections tend to carry larger partial correlation weights in
psychometric networks.

\strong{Negative edge assignment}

Negative edges are introduced by flipping the signs of a subset of
nodes — multiplying all edges incident to those nodes by -1. The number
of nodes to flip is derived from \code{negative_proportion} via the
inversion formula \eqn{(1 - \sqrt{1 - 2p}) / 2}, where \eqn{p} is the
target proportion of negative edges. This formula assumes that an edge
is negative if and only if exactly one of its endpoints is flipped,
giving an expected negative proportion of \eqn{2 \cdot (k/n) \cdot
(1 - k/n)} for \eqn{k} flipped nodes out of \eqn{n} total. The formula
requires \eqn{p \leq 0.50}, which is why \code{negative_proportion} is
bounded at 0.50.
}
\examples{
# Basic small-world network (moderate density, moderate rewiring)
result <- simulate_smallworld(
  nodes = 20,
  density = 0.30,
  rewire = 0.20,
  sample_size = 500
)

# Lattice-like structure (low rewiring preserves local connectivity)
result <- simulate_smallworld(
  nodes = 20,
  density = 0.30,
  rewire = 0.01,
  sample_size = 500
)

# Random-like structure (high rewiring destroys lattice regularity)
result <- simulate_smallworld(
  nodes = 20,
  density = 0.30,
  rewire = 0.80,
  sample_size = 500
)

# Fix the proportion of negative edges
result <- simulate_smallworld(
  nodes = 20,
  density = 0.30,
  rewire = 0.20,
  sample_size = 500,
  negative_proportion = 0.20
)

}
\references{
\strong{Seminal introduction to the Watts-Strogatz small-world model} \cr
Watts, D. J., & Strogatz, S. H. (1998).
Collective dynamics of 'small-world' networks.
\emph{Nature}, \emph{393}(6684), 440--442.

\strong{Logic for weight assignments} \cr
Muldoon, S. F., Bridgeford, E. W., & Bassett, D. S. (2016).
Small-world propensity and weighted brain networks.
\emph{Scientific Reports}, \emph{6}(1), 22057.

\strong{Analytical approximation of random-graph average path length} \cr
Newman, M. E. J., Strogatz, S. H., & Watts, D. J. (2001).
Random graphs with arbitrary degree distributions and their applications.
\emph{Physical Review E}, \emph{64}(2), 026118.

\strong{Empirical network data used to fit the Weibull SUR model} \cr
Huth, K. B. S., Haslbeck, J. M. B., Keetelaar, S., Van Holst, R. J., &
Marsman, M. (2025). Statistical evidence in psychological networks.
\emph{Nature Human Behaviour}.

\strong{Maximum ridge shrinkage bound} \cr
Peeters, C. F., van de Wiel, M. A., & van Wieringen, W. N. (2020).
The spectral condition number plot for regularization parameter evaluation.
\emph{Computational Statistics}, \emph{35}(2), 629--646.
}
\author{
Alexander P. Christensen <alexpaulchristensen@gmail.com>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate_sbm.R
\name{simulate_sbm}
\alias{simulate_sbm}
\title{Simulates Stochastic Block Model Data}
\usage{
simulate_sbm(
  nodes,
  blocks,
  density_matrix,
  snr = 1,
  diffusion = 0.3,
  diffusion_range = NULL,
  negative_proportion,
  sample_size,
  skew = 0,
  skew_range = NULL,
  target_condition = 30,
  max_correlation = 0.8,
  max_iterations = 100
)
}
\arguments{
\item{nodes}{Numeric (length = 1 or \code{blocks}).
Number of nodes per community block. Can be a single value applied to all
blocks, or a vector of length \code{blocks} specifying each block's size
individually. Minimum of three nodes per block. The total number of nodes
(\code{sum(nodes)}) should be between 8 and 54 to remain within the range
of the empirical networks used to fit the Weibull parameter model; values
outside this range are accepted but will trigger extrapolation and may
produce a warning from \code{\link{weibull_parameters}}.}

\item{blocks}{Numeric (length = 1).
Number of community blocks.}

\item{density_matrix}{Matrix (dim = \code{blocks x blocks}).
A symmetric numeric matrix specifying edge probabilities within and between
community blocks. Diagonal entry \eqn{[i,i]} gives the within-block edge
probability for community \eqn{i}; off-diagonal entry \eqn{[i,j]}
(\eqn{i \neq j}) gives the between-block edge probability for the pair of
communities \eqn{i} and \eqn{j}. All entries must be in \eqn{[0, 1]} and
the matrix must be symmetric (i.e., \code{density_matrix[i,j] ==
density_matrix[j,i]}). See Details for construction guidance.}

\item{snr}{Numeric (length = 1).
Signal-to-noise ratio of the absolute partial correlation weights,
defined as \eqn{\bar{|w|} / \mathrm{SD}(|w|)}. This ratio governs the
shape and scale of the Weibull distribution used to generate edge weights:
values below 1 produce wider, more heterogeneous weight distributions;
values above 1 produce narrower, more homogeneous distributions. Note that
the same SNR can arise from different combinations of mean and standard
deviation (e.g., mean = 0.05, SD = 0.05 and mean = 0.15, SD = 0.15 both
give SNR = 1), so SNR alone does not determine the absolute magnitude of
edge weights, which is additionally governed by \code{nodes} and
\code{sample_size} via \code{\link{weibull_parameters}}. Empirically
observed SNR values ranged from 0.648 to 1.712; values outside this range
are accepted but will trigger a warning. Defaults to \code{1}.}

\item{diffusion}{Numeric (length = 1).
Controls the minimum proportion of the top-ranked edges (by absolute
weight) that are guaranteed to be placed within communities rather than
distributed freely across the network. Specifically, \code{1 - diffusion}
of the within-community edges are reserved for the highest-weight draws;
the remaining edges (both within- and between-community) are filled from
a randomly shuffled pool. Because shuffled edges can land within communities
by chance, the actual proportion of top edges that end up within communities
will on average exceed \code{1 - diffusion}. Lower values of
\code{diffusion} (e.g., \code{0.10}) therefore produce stronger community
contrast than higher values (e.g., \code{0.90}), but \code{diffusion}
should be interpreted as a floor on within-community weight concentration,
not an exact control. Defaults to \code{0.30}. Must be between 0 and 1.}

\item{diffusion_range}{Numeric (length = 2).
If provided, overrides \code{diffusion} by drawing the diffusion proportion
uniformly from this interval on each call. Useful for introducing
replication-to-replication variability in community contrast. For example,
\code{diffusion_range = c(0.05, 0.20)} samples a value between 5\% and
20\% on each draw. The same floor interpretation applies as for
\code{diffusion}. Both values must be between 0 and 1.}

\item{negative_proportion}{Numeric (length = 1).
Proportion of between-community edges assigned a negative sign (inhibitory
connections). Each between-community edge is independently signed negative
with this probability (i.e., a Bernoulli draw per edge), so the realized
proportion will vary around the specified value. Must be between 0 and 1.
Applies only to between-community edges; all within-community edges are
positive. If not provided, a value is sampled from a truncated normal
distribution reflecting the empirical distribution of true negative partial
correlations across 194 psychometric networks: mean = 0.34, SD = 0.086,
bounded to \eqn{[0.083, 0.55]}.}

\item{sample_size}{Numeric (length = 1).
Number of observations to generate from the population multivariate
distribution. Also influences the predicted Weibull scale parameter via
\code{\link{weibull_parameters}}: larger samples are associated with
smaller, more precisely estimated edge weights.}

\item{skew}{Numeric (length = 1 or \code{sum(nodes)}).
Skew applied to each variable after generation from the multivariate
normal. Can be a single value applied to all variables, or one value per
variable. Values are rounded to the nearest 0.05 increment and must be
in \eqn{[-2, 2]}. Defaults to \code{0} (no skew).}

\item{skew_range}{Numeric (length = 2).
If provided, overrides \code{skew} by drawing a skew value independently
and uniformly from this interval for each variable. Both values must be
in \eqn{[-2, 2]}.}

\item{target_condition}{Numeric (length = 1).
Target condition number (computed via \code{\link{kappa}} with
\code{exact = TRUE}) used when ridge regularization is required to recover
a positive definite precision matrix. The smallest ridge penalty
\eqn{\lambda} that brings the condition number to this target is found via
root-finding (\code{\link{uniroot}}), subject to a maximum shrinkage of
approximately 23\% following Peeters et al. (2020). After conditioning,
the Weibull bounds are re-verified on the updated edge weights; draws that
fall outside the empirical bounds after conditioning are rejected. Lower
values produce better-conditioned (more stable) matrices. Defaults to
\code{30}. Values up to \code{100} are accepted but not recommended.}

\item{max_correlation}{Numeric (length = 1).
Maximum allowed absolute pairwise correlation in the population correlation
matrix \code{R}. Any draw where \code{max(abs(R[lower.tri(R)])) >
max_correlation} is rejected and a new attempt is made. Must be between
0 and 1. Defaults to \code{0.80}.}

\item{max_iterations}{Numeric (length = 1).
Maximum number of attempts to find a connected network with valid edge
weights before stopping with an error. The error message reports a
frequency table of rejection reasons to assist with diagnosing convergence
failures. Defaults to \code{100}.}
}
\value{
A named list with four elements:

\item{data}{Numeric matrix of dimension \code{sample_size x sum(nodes)}
containing the simulated observations drawn from the population GGM.
Rows are cases; columns are variables named \code{V01}, \code{V02},
etc. Values are continuous (or skewed continuous when \code{skew != 0}).
To produce ordinal data, pass the columns through
\code{\link{categorize}}.}

\item{parameters}{
A list of input, derived, and estimated parameters:
\itemize{
  \item \code{nodes} --- Integer vector of length \code{blocks} giving the
  number of nodes per block (scalar input is expanded to this length)
  \item \code{blocks} --- Number of community blocks
  \item \code{sample_size} --- Number of simulated observations
  \item \code{skew} --- Named numeric vector of per-variable skew values
  actually applied (after rounding and possible resampling)
  \item \code{density_matrix} --- The \code{blocks x blocks} density matrix
  as supplied
  \item \code{negative_proportion} --- The proportion of between-community
  edges assigned a negative sign, either as supplied or as sampled from the
  empirical distribution
  \item \code{weibull} --- Named numeric vector of length 2 giving the
  Weibull \code{shape} and \code{scale} parameters of the absolute edge
  weight distribution actually used. If ridge conditioning was applied,
  these are re-estimated from the conditioned network via MLE.
  \item \code{diffusion} --- Numeric vector of length 2 giving the
  within-block reservation range as a proportion of within-community edges,
  on the internal scale used by the sampler (i.e., \code{range(1 -
  diffusion)} or \code{range(1 - diffusion_range)}). Both values are equal
  when \code{diffusion} is scalar. Note this is the complement of the
  user-supplied \code{diffusion} value and represents the fraction of
  within-community slots filled from the top-ranked draws.
  \item \code{omega} --- Smallworldness omega statistic of the generated
  network (Telesford et al., 2011). Values near zero indicate small-world
  structure; negative values indicate lattice-like structure; positive
  values indicate random-like structure (see \code{\link[L0ggm]{smallworldness}})
  \item \code{Q} --- Newman-Girvan modularity of the population network
  (\code{Omega}) with respect to the block membership, computed via
  \code{igraph::modularity} on absolute edge weights. Provides a summary
  of the community contrast actually achieved after weight assignment and
  any ridge conditioning.
}
}

\item{population}{
Population-level network parameters:
\itemize{
  \item \code{R} --- Population correlation matrix derived from the GGM
  via \code{pcor2cor}
  \item \code{Omega} --- Population partial correlation matrix (the GGM
  edge weight matrix), with zeros for absent edges
  \item \code{membership} --- Named integer vector of length
  \code{sum(nodes)} giving the community block assignment (1 to
  \code{blocks}) for each node
}
}

\item{convergence}{
Iteration and conditioning diagnostics:
\itemize{
  \item \code{iterations} --- Number of sampling attempts needed to find a
  valid network (including graph structure and edge weight draws)
  \item \code{rejections} --- Character vector of length
  \code{max_iterations + 1} recording the rejection reason for each failed
  attempt; entries for successful or unused iterations are empty strings.
  Common reasons include disconnected graph structure, condition number
  exceeding \code{target_condition}, maximum correlation exceeding
  \code{max_correlation}, and Weibull parameters falling outside empirical
  bounds after ridge conditioning.
  \item \code{lambda} --- Ridge regularization parameter \eqn{\lambda}
  added to the diagonal of the precision matrix to ensure positive
  definiteness; \code{NA} if no conditioning was required
  \item \code{condition} --- Condition number of the final population
  correlation matrix \code{R}, computed via \code{kappa} with
  \code{exact = TRUE}
}
}
}
\description{
Simulates data from a Gaussian Graphical Model (GGM) with a stochastic
block model (SBM) structure. Nodes are partitioned into communities, with
edge density controlled separately within and between communities via a
\code{blocks x blocks} density matrix. Absolute edge weights are drawn from
a Weibull distribution whose parameters are predicted from the network size,
sample size, and signal-to-noise ratio using a Seemingly Unrelated
Regression (SUR) model fitted to 194 empirical psychometric networks
(Huth et al., 2025). The resulting population partial correlation matrix is
used to generate multivariate normal (or skewed) data.

A \code{diffusion} parameter controls the minimum proportion of the
strongest edges that are reserved for within-community positions. Because
the remaining edges are shuffled randomly, the effective within-community
advantage will generally exceed \code{1 - diffusion}; see Details.

Parameters do not have default values (except \code{negative_proportion},
\code{diffusion}, \code{target_condition}, \code{max_correlation}, and
\code{max_iterations}) and must each be set. See Details and Examples to
get started.
}
\details{
\strong{Constructing \code{density_matrix}}

The \code{density_matrix} is a \code{blocks x blocks} symmetric matrix
where entry \eqn{[i,i]} gives the within-block edge probability for
community \eqn{i}, and entry \eqn{[i,j]} (for \eqn{i \neq j}) gives the
between-block edge probability for communities \eqn{i} and \eqn{j}. The
simplest construction uses a uniform off-diagonal density with
block-specific diagonals:

\preformatted{
# Uniform within (0.90) and between (0.20) density for 3 blocks
dm <- matrix(0.20, nrow = 3, ncol = 3)
diag(dm) <- 0.90
}

For asymmetric community structure, each diagonal entry can differ:

\preformatted{
# Varying within-block density per community
dm <- matrix(0.20, nrow = 3, ncol = 3)
diag(dm) <- c(0.85, 0.90, 0.95)
}

For full pairwise control over between-block densities, specify the
complete symmetric matrix directly:

\preformatted{
dm <- matrix(c(
  0.90, 0.20, 0.10,
  0.20, 0.85, 0.30,
  0.10, 0.30, 0.95
), nrow = 3, ncol = 3)
}

\strong{Diffusion and within-community weight concentration}

The \code{diffusion} parameter does not exactly fix the proportion of
top-ranked edges placed within communities. Instead, \code{1 - diffusion}
of the within-community edge slots are filled deterministically from the
highest-weight draws. The remaining edge slots (within- and
between-community alike) are filled from a randomly shuffled pool of
lower-ranked weights, meaning some additional high-weight edges will land
within communities by chance. The realized within-community weight
advantage will therefore always be at least as large as implied by
\code{1 - diffusion}, and typically larger. The \code{Q} field in the
returned \code{parameters} list (Newman-Girvan modularity) provides a
post-hoc summary of the actual community contrast achieved.
}
\examples{
# Construct density matrix for 3 blocks with uniform densities
dm <- matrix(0.20, nrow = 3, ncol = 3)
diag(dm) <- 0.90

# Basic 3-block simulation with equal-sized communities
result <- simulate_sbm(
  nodes = 6, # 6 nodes per block = 18 total
  blocks = 3,
  sample_size = 500,
  density_matrix = dm
)

# Unequal block sizes
result <- simulate_sbm(
  nodes = c(4, 6, 8), # 18 total nodes
  blocks = 3,
  sample_size = 500,
  density_matrix = dm
)

# Varying within-block density per community
dm_varying <- matrix(0.20, nrow = 3, ncol = 3)
diag(dm_varying) <- c(0.85, 0.90, 0.95)

result <- simulate_sbm(
  nodes = 6,
  blocks = 3,
  sample_size = 500,
  density_matrix = dm_varying
)

# Full pairwise between-block density control
dm_pairwise <- matrix(c(
  0.90, 0.20, 0.10,
  0.20, 0.85, 0.30,
  0.10, 0.30, 0.95
), nrow = 3, ncol = 3)

result <- simulate_sbm(
  nodes = 6,
  blocks = 3,
  sample_size = 500,
  density_matrix = dm_pairwise
)

# Fix the proportion of negative between-community edges
result <- simulate_sbm(
  nodes = 6,
  blocks = 3,
  sample_size = 500,
  density_matrix = dm,
  negative_proportion = 0.20
)

# Introduce variability in diffusion across replications
result <- simulate_sbm(
  nodes = 6,
  blocks = 3,
  sample_size = 500,
  density_matrix = dm,
  diffusion_range = c(0.30, 0.70)
)

}
\references{
\strong{Seminal introduction to Stochastic Block Models} \cr
Holland, P. W., Laskey, K. B., & Leinhardt, S. (1983).
Stochastic blockmodels: First steps.
\emph{Social Networks}, \emph{5}(2), 109--137.

\strong{Empirical network data used to fit the Weibull SUR model} \cr
Huth, K. B. S., Haslbeck, J. M. B., Keetelaar, S., Van Holst, R. J., &
Marsman, M. (2025). Statistical evidence in psychological networks.
\emph{Nature Human Behaviour}.

\strong{Maximum ridge shrinkage bound} \cr
Peeters, C. F., van de Wiel, M. A., & van Wieringen, W. N. (2020).
The spectral condition number plot for regularization parameter evaluation.
\emph{Computational Statistics}, \emph{35}(2), 629--646.
}
\author{
Alexander P. Christensen <alexpaulchristensen@gmail.com>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smallworldness.R
\name{smallworldness}
\alias{smallworldness}
\title{Computes Various Small-Worldness Metrics}
\usage{
smallworldness(
  network,
  lattice,
  method = c("analytical", "omega", "S", "SWI", "SWP"),
  weighted = FALSE,
  iter = 100
)
}
\arguments{
\item{network}{Matrix or data frame.
A square, symmetric numeric matrix representing a network (e.g., partial
correlations). Absolute values are taken internally, so signed weights are
handled automatically. Non-zero off-diagonal entries are treated as edges.}

\item{lattice}{Matrix (optional).
A pre-computed lattice adjacency matrix to use as the regular-network
reference for the \code{"omega"} and \code{"SWP"} methods.
If not provided, a lattice is generated automatically via
\code{\link[L0ggm]{proxswap_lattice}}. Ignored by \code{"analytical"}
and \code{"S"}}

\item{method}{Character (length = 1).
The method used to compute small-worldness.
Defaults to \code{"SWP"}.
Available options:

\itemize{

\item \code{"analytical"} --- Computes the Humphries & Gurney (2008)
\eqn{S} metric using closed-form approximations for the Erdős–Rényi
random graph baseline:
\deqn{S = \frac{C / C_{\text{rand}}}{L / L_{\text{rand}}}}
where \eqn{C_{\text{rand}} \approx \langle k \rangle / n} and
\eqn{L_{\text{rand}} \approx \ln(n) / \ln(\langle k \rangle)}.
No random graphs are generated. The \code{weighted} argument is ignored.
Values greater than 1 indicate small-world structure

\item \code{"S"} --- Computes the Humphries & Gurney (2008) \eqn{S}
metric by simulation. Random graphs are generated under the
Erdős–Rényi model (\code{\link[igraph]{sample_gnm}}) with the same
number of nodes and edges as \code{network}. The empirical and random
clustering coefficients both use global transitivity (\eqn{C^\Delta}).
When \code{weighted = TRUE}, edge weights are reassigned to each random
graph topology via \code{assign_weights}. Values greater than 1
indicate small-world structure

\item \code{"omega"} --- Computes the \eqn{\omega} metric of
Telesford et al. (2011):
\deqn{\omega = \frac{L_{\text{rand}}}{L} - \frac{C}{C_{\text{latt}}}}
Random graphs preserve the degree sequence via edge-switching
(\code{\link[igraph]{sample_degseq}}, \code{method = "edge.switching.simple"}).
The clustering coefficient uses average local transitivity (\eqn{\bar{C}}).
Note: Telesford et al. (2011) defined and validated \eqn{\omega} exclusively
on binary, unweighted networks. When \code{weighted = TRUE}, edge weights
are reassigned to each random graph topology via \code{assign_weights} and
the lattice is constructed with \code{weighted = TRUE}, following the
approach of Muldoon et al. (2016); this is an extension beyond the original
formulation. Values near zero indicate small-world structure; negative
values indicate lattice-like structure; positive values indicate
random-like structure. Bounded in \eqn{[-1, 1]}

\item \code{"SWI"} --- Computes the Small-World Index of
Neal (2015):
\deqn{\text{SWI} = \frac{L - L_{\text{latt}}}{L_{\text{rand}} - L_{\text{latt}}} \times \frac{C - C_{\text{rand}}}{C_{\text{latt}} - C_{\text{rand}}}}
Each term captures where the observed network's path length and clustering
coefficient fall within the lattice-to-random range, so that SWI = 1
only when \eqn{L = L_{\text{rand}}} and \eqn{C = C_{\text{latt}}}
simultaneously. Because this ideal is mathematically unachievable in a
finite network, SWI = 1 is a conceptual upper bound rather than a
realisable value. Random graphs preserve the degree sequence via
edge-switching (\code{\link[igraph]{sample_degseq}},
\code{method = "edge.switching.simple"}). The clustering coefficient uses
average local transitivity (\eqn{\bar{C}}). When \code{weighted = TRUE},
edge weights are reassigned to each random graph topology and the lattice
is constructed with \code{weighted = TRUE}. Note: SWI and SWP were
developed independently and concurrently; both apply the same
double-normalisation framework but differ in how the two deviation terms
are combined (product vs. Euclidean distance). Values close to 1 indicate
strong small-world structure; values near 0 indicate lattice-like or
random-like structure. Bounded in \eqn{[0, 1]}

\item \code{"SWP"} --- Computes the Small-World Propensity of
Muldoon et al. (2016):
\deqn{\phi = 1 - \sqrt{\frac{\Delta_C^2 + \Delta_L^2}{2}}}
where \eqn{\Delta_C = (C_{\text{latt}} - C) / (C_{\text{latt}} - C_{\text{rand}})}
and \eqn{\Delta_L = (L - L_{\text{rand}}) / (L_{\text{latt}} - L_{\text{rand}})}.
Both \eqn{\Delta_C} and \eqn{\Delta_L} are bounded to \eqn{[0, 1]},
guaranteeing \eqn{\phi \in [0, 1]}. Random graphs preserve the degree
sequence via edge-switching. The clustering coefficient uses average
local transitivity (\eqn{\bar{C}}). When \code{weighted = TRUE}, edge
weights are reassigned to each random graph topology via
\code{assign_weights} and the lattice is constructed with
\code{weighted = TRUE}. Values close to 1 indicate strong small-world
structure; a pragmatic threshold of \eqn{\phi_T = 0.60} has been suggested

}}

\item{weighted}{Logical (length = 1).
Whether to compute small-worldness on the weighted network. When
\code{TRUE}, edge weights from \code{network} are preserved for the
empirical graph and reassigned to random and lattice graph topologies
via \code{assign_weights}, following Muldoon et al. (2016). When
\code{FALSE} (default), all graphs are treated as binary. Ignored when
\code{method = "analytical"}}

\item{iter}{Numeric (length = 1).
Number of random graphs to generate when estimating the null baseline.
Defaults to \code{100}.
Ignored when \code{method = "analytical"}.
Higher values produce more stable estimates at the cost of computation time}
}
\value{
Numeric (length = 1).
The small-worldness value computed using the chosen \code{method}:

\itemize{

\item \code{"analytical"} and \code{"S"} --- \eqn{S > 1} indicates small-world structure

\item \code{"omega"} --- Values near \eqn{|\omega| < 0.50} indicate small-world structure
(ranges between -1 and 1)

\item \code{"SWI"} --- Values close to 1 indicate strong small-world structure;
values near 0 indicate lattice-like or random-like structure
(ranges between 0 and 1)

\item \code{"SWP"} --- Values \eqn{\phi > 0.60} indicate strong small-world structure
(ranges between 0 and 1)

}
}
\description{
Computes the small-worldness of a network using one of five
methods. All simulation-based methods generate degree-preserving random
graphs as the null baseline. The lattice reference network (where required)
is constructed using \code{\link[L0ggm]{proxswap_lattice}}, which produces a
degree-preserving ring lattice that maximizes the clustering coefficient
}
\examples{
# Get network
network <- network_estimation(basic_smallworld)

# Compute SWP (default)
swp <- smallworldness(network)

# Compute omega
omega <- smallworldness(network, method = "omega")

# Compute analytical S
S <- smallworldness(network, method = "analytical")

# Compute simulated S
S_sim <- smallworldness(network, method = "S")

# Compute SWI
swi <- smallworldness(network, method = "SWI")

# Compute weighted SWP
swp_w <- smallworldness(network, weighted = TRUE)

# Compute weighted omega
omega_w <- smallworldness(network, method = "omega", weighted = TRUE)

}
\references{
\strong{omega} \cr
Telesford, Q. K., Joyce, K. E., Hayasaka, S., Burdette, J. H., & Laurienti, P. J. (2011).
The ubiquity of small-world networks.
\emph{Brain Connectivity}, \emph{1}(5), 367--375.

\strong{S and analytical} \cr
Humphries, M. D., & Gurney, K. (2008).
Network 'small-world-ness': A quantitative method for determining canonical network equivalence.
\emph{PLoS ONE}, \emph{3}(4), e0002051.

\strong{SWI} \cr
Neal, Z. P. (2015).
Making big communities small: Using network science to understand the ecological and behavioral requirements for community social capital.
\emph{American Journal of Community Psychology}, \emph{55}(3), 369--380.

\strong{SWP} \cr
Muldoon, S. F., Bridgeford, E. W., & Bassett, D. S. (2016).
Small-world propensity and weighted brain networks.
\emph{Scientific Reports}, \emph{6}(1), 22057.
}
\author{
Alexander P. Christensen <alexpaulchristensen@gmail.com>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/L0ggm.R
\docType{package}
\name{L0ggm-package}
\alias{L0ggm}
\alias{L0ggm-package}
\title{L0ggm-package}
\description{
Implements L0 norm regularization penalties for Gaussian graphical models
}
\seealso{
Useful links:
\itemize{
  \item Report bugs at \url{https://github.com/AlexChristensen/L0ggm/issues}
}

}
\author{
Alexander P. Christensen <alexpaulchristensen@gmail.com>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/auto_correlate.R
\name{auto_correlate}
\alias{auto_correlate}
\title{Automatic correlations}
\usage{
auto_correlate(
  data,
  corr = c("kendall", "pearson", "spearman"),
  ordinal_categories = 7,
  forcePD = TRUE,
  na_data = c("pairwise", "listwise"),
  empty_method = c("none", "zero", "all"),
  empty_value = c("none", "point_five", "one_over"),
  forceReturn = FALSE,
  verbose = FALSE,
  ...
)
}
\arguments{
\item{data}{Matrix or data frame.
Should consist only of variables to be used in the analysis}

\item{corr}{Character (length = 1).
The standard correlation method to be used.
Defaults to \code{"pearson"}.
Using \code{"pearson"} will compute polychoric, tetrachoric, polyserial,
and biserial correlations for categorical and categorical/continuous correlations
by default. To obtain \code{"pearson"} correlations regardless, use \code{\link{cor}}.
Other options of \code{"kendall"} and \code{"spearman"} are provided for
completeness and use \code{\link{cor}}}

\item{ordinal_categories}{Numeric (length = 1).
\emph{Up to} the number of categories \emph{before} a variable is considered continuous.
Defaults to \code{7} categories before \code{8} is considered continuous}

\item{forcePD}{Boolean (length = 1).
Whether positive definite matrix should be enforced.
Defaults to \code{TRUE}}

\item{na_data}{Character (length = 1).
How should missing data be handled?
Defaults to \code{"pairwise"}.
Available options:

\itemize{

\item \code{"pairwise"} --- Computes correlation for all available
cases between two variables

\item \code{"listwise"} --- Computes correlation for all complete
cases in the dataset

}}

\item{empty_method}{Character (length = 1).
Method for empty cell correction in \code{\link[L0ggm]{polychoric_matrix}}.
Defaults to \code{"none"}
Available options:

\itemize{

\item \code{"none"} --- Adds no value (\code{empty_value = "none"})
to the empirical joint frequency table between two variables

\item \code{"zero"} --- Adds \code{empty_value} to the cells with
zero in the joint frequency table between two variables

\item \code{"all"} --- Adds \code{empty_value} to all
in the joint frequency table between two variables

}}

\item{empty_value}{Character (length = 1).
Value to add to the joint frequency table cells in \code{\link[L0ggm]{polychoric_matrix}}.
Defaults to \code{"none"}.
Accepts numeric values between 0 and 1 or specific methods:

\itemize{

\item \code{"none"} --- Adds no value (\code{0}) to the empirical
joint  frequency table between two variables

\item \code{"point_five"} --- Adds \code{0.5} to the cells
defined by \code{empty_method}

\item \code{"one_over"} --- Adds \code{1 / n} where \code{n} equals the
number of cells based on \code{empty_method}. For
\code{empty_method = "zero"}, \code{n} equals the number of zero cells

}}

\item{forceReturn}{Boolean (length = 1).
Whether correlation matrix should be forced to return.
Defaults to \code{FALSE}.
Set to \code{TRUE} to receive the correlation matrix "as is"}

\item{verbose}{Boolean (length = 1).
Whether messages should be printed.
Defaults to \code{FALSE}}

\item{...}{Not actually used but makes it easier for general functionality
in the package}
}
\value{
A symmetric numeric matrix of dimension \eqn{p \times p}, where
\eqn{p} is the number of variables in \code{data}, with values in
\eqn{[-1, 1]} and ones on the diagonal. The correlation method used for
each pair of variables depends on their types when \code{corr = "pearson"}
(the default): polychoric for ordinal-ordinal pairs, polyserial for
ordinal-continuous pairs, and Pearson's for continuous-continuous pairs,
where \emph{ordinal} means having at most \code{ordinal_categories} unique
values. When \code{corr} is \code{"kendall"} or \code{"spearman"},
\code{\link[stats]{cor}} is used for all pairs regardless of variable
type. Row and column names are inherited from \code{data}. When
\code{forcePD = TRUE} (default) and the resulting matrix is not positive
definite, the nearest positive definite matrix is returned via
\code{\link[Matrix]{nearPD}}.
}
\description{
Automatically computes the proper correlations between
continuous and categorical variables. \code{NA} values are not treated
as categories
}
\examples{
# Obtain correlations
R <- auto_correlate(basic_smallworld)

}
\author{
Alexander P. Christensen <alexpaulchristensen@gmail.com>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/edge_confusion.R
\name{edge_confusion}
\alias{edge_confusion}
\title{Confusion Matrix Metrics for Edge Comparison and Recovery}
\usage{
edge_confusion(
  base,
  comparison,
  metric = c("all", "sen", "spec", "ppv", "npv", "fdr", "fom", "ba", "f1", "csi", "mcc"),
  full.names = FALSE
)
}
\arguments{
\item{base}{Matrix or data frame.
Network that will be treated as the "ground truth" such that
a false positive represents an edge that is present in \code{comparison}
but not in this network}

\item{comparison}{Matrix or data frame.
Network that will be treated as the estimator such that a
false positive represents an edge that is present in this network
but not in \code{base}}

\item{metric}{Character vector.
Defaults to \code{"all"} metrics.
Available options:

\itemize{

\item \code{"all"} --- All available metrics (default)

\item \code{"sen"} --- Sensitivity (True Positive Rate):
\deqn{\frac{TP}{TP + FN}}

\item \code{"spec"} --- Specificity (True Negative Rate):
\deqn{\frac{TN}{TN + FP}}

\item \code{"ppv"} --- Positive Predictive Value (Precision):
\deqn{\frac{TP}{TP + FP}}

\item \code{"npv"} --- Negative Predictive Value:
\deqn{\frac{TN}{TN + FN}}

\item \code{"fdr"} --- False Discovery Rate:
\deqn{1 - PPV = \frac{FP}{TP + FP}}

\item \code{"fom"} --- False Omission Rate:
\deqn{1 - NPV = \frac{FN}{TN + FN}}

\item \code{"ba"} --- Balanced Accuracy:
\deqn{\frac{Sensitivity + Specificity}{2}}

\item \code{"f1"} --- F1 Score (harmonic mean of PPV and Sensitivity):
\deqn{\frac{2TP}{2TP + FP + FN}}

\item \code{"csi"} --- Critical Success Index (Jaccard / Threat Score):
\deqn{\frac{TP}{TP + FP + FN}}

\item \code{"mcc"} --- Matthews Correlation Coefficient:
\deqn{\frac{TP \times TN - FP \times FN}{\sqrt{(TP+FP)(TP+FN)(TN+FP)(TN+FN)}}}

}}

\item{full.names}{Boolean (length = 1).
Whether full or abbreviated names should be used.
Defaults to \code{FALSE}.
Set to \code{TRUE} for full names}
}
\value{
A named numeric vector of the requested confusion matrix metrics.
Values are generally in \eqn{[0, 1]}, with the exception of \code{"mcc"}
(Matthews Correlation Coefficient), which ranges from \eqn{-1} to \eqn{1}
where 1 indicates perfect agreement, 0 indicates chance-level performance,
and -1 indicates perfect disagreement. The vector contains only the elements
specified by \code{metric} (all ten metrics when \code{metric = "all"},
which is the default). Names are abbreviated (e.g., \code{"sen"}) when
\code{full.names = FALSE} (default), or expanded
(e.g., \code{"Sensitivity"}) when \code{full.names = TRUE}. Any metric
whose denominator is zero (e.g., sensitivity when no edges exist in
\code{base}) is returned as \code{NA}.
}
\description{
Computes many commonly used confusion matrix metrics
}
\examples{
# Set split
split <- sample(
  1:nrow(basic_smallworld),
  round(nrow(basic_smallworld) / 2)
)

# Estimate networks
split1 <- network_estimation(basic_smallworld[split,])
split2 <- network_estimation(basic_smallworld[-split,])

# Estimate metrics
edge_confusion(split1, split2)

}
\author{
Alexander P. Christensen <alexpaulchristensen@gmail.com>
}
